# tests.status.test_aggregator
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the status events aggregator."""

import logging

from twisted.internet.task import Clock
from twisted.trial.unittest import TestCase

from ubuntuone.devtools.handlers import MementoHandler
from ubuntuone.status import aggregator
from ubuntuone.status.notification import AbstractNotification
from ubuntuone.syncdaemon import status_listener
from ubuntuone.syncdaemon.volume_manager import Share, UDF


class PatchedClock(Clock):
    """Patch the clock to fix twisted bug #4823."""

    def advance(self, amount):
        self.calls.sort(lambda a, b: cmp(a.getTime(), b.getTime()))
        Clock.advance(self, amount)


class TimerTestCase(TestCase):
    """Test the Timer class."""

    TIMEOUT = 3.0

    def setUp(self):
        """Initialize this test instance."""
        self.clock = PatchedClock()
        self.timer = aggregator.Timer(delay=3.0, clock=self.clock)

    def test_not_fired_initially(self):
        """The timer is not fired initially"""
        self.assertFalse(self.timer.called)

    def test_fired_after_delay(self):
        """The timer is fired after the initial delay."""
        self.clock.advance(self.timer.delay)
        self.assertTrue(self.timer.called)

    def test_cleanup_cancels_delay_call(self):
        """Calling cleanup cancels the delay call."""
        self.timer.cleanup()
        self.assertTrue(self.timer.delay_call.cancelled)

    def test_not_fired_immediately(self):
        """The timer is not fired immediately."""
        self.timer.reset()
        self.assertFalse(self.timer.called)

    def test_fired_after_initial_wait(self):
        """The timer is fired after an initial wait."""
        self.timer.reset()
        self.clock.advance(self.timer.delay)
        self.assertTrue(self.timer.called)

    def test_not_fired_if_reset_within_delay(self):
        """The timer is not fired if it is reset within the delay."""
        self.timer.reset()
        self.clock.advance(self.timer.delay/0.8)
        self.timer.reset()
        self.clock.advance(self.timer.delay/0.8)
        self.assertTrue(self.timer.called)


class DeadlineTimerTestCase(TimerTestCase):
    """Test the DeadlineTimer class."""

    DELAY = 0.5

    def setUp(self):
        """Initialize this test instance."""
        self.clock = PatchedClock()
        self.timer = aggregator.DeadlineTimer(delay=0.5, timeout=3.0,
                                            clock=self.clock)

    def test_fired_if_initial_timeout_exceeded(self):
        """Timer is fired if the initial timeout is exceeded."""
        small_delay = self.timer.delay * 0.8
        for n in range(int(self.timer.timeout / small_delay) + 1):
            self.timer.reset()
            self.clock.advance(small_delay)
        self.assertTrue(self.timer.called)

    def test_not_fired_twice_if_delay_exceeded(self):
        """Timer is not fired twice if the delay is exceeded."""
        large_delay = self.timer.delay * 1.2
        for n in range(int(self.timer.timeout / large_delay) + 1):
            self.timer.reset()
            self.clock.advance(large_delay)
        self.clock.advance(self.timer.delay)
        self.assertTrue(self.timer.called)

    def test_not_fired_twice_if_timeout_exceeded(self):
        """Timer is not fired twice if the timeout is exceeded."""
        small_delay = self.timer.delay * 0.8
        for n in range(int(self.timer.timeout / small_delay) + 1):
            self.timer.reset()
            self.clock.advance(small_delay)
        self.clock.advance(self.timer.delay)
        self.assertTrue(self.timer.called)

    def test_cleanup_cancels_timeout_call(self):
        """Calling cleanup cancels the delay call."""
        self.timer.cleanup()
        self.assertTrue(self.timer.timeout_call.cancelled)


class FakeNotification(AbstractNotification):
    """A fake notification class."""

    def __init__(self, application_name="fake app"):
        """Initialize this instance."""
        self.notifications_shown = []
        self.application_name = application_name

    def send_notification(self, title, message, icon=None, append=False):
        """Show a notification to the user."""
        self.notifications_shown.append((title, message, icon, append))
        return len(self.notifications_shown) - 1

    def update_notification(self, notification, new_title, new_message,
                            new_icon=None):
        """Update the notification with a new message body."""
        title, old_message, icon, append = self.notifications_shown[
            notification]
        if new_icon is not None:
            icon = new_icon
		# we store it as a new notification, to ease testing
        notification_params = (new_title, new_message, icon, append)
        self.notifications_shown.append(notification_params)


def FakeNotificationSingleton():
    """Builds a notification singleton, that logs all notifications shown."""
    instance = FakeNotification()

    def get_instance():
        """Returns the single instance."""
        return instance

    return get_instance


class FakeMessaging(object):
    """A fake messaging class."""

    def __init__(self, application_name="fake app"):
        self.messages_shown = {}
        self.application_name = application_name

    def show_message(self, sender, callback, message_time=None,
                     message_count=None, icon=None):
        """Show a message to the user."""
        self.messages_shown[sender] = (
            callback, message_time, message_count, icon)

    def update_count(self, sender, new_count):
        """Update the count for an existing indicator."""
        callback, message_time, _, icon = self.messages_shown[sender]
        self.messages_shown[sender] = (
            callback, message_time, new_count, icon)


class FakeStatusAggregator(object):
    """A fake status aggregator."""

    def __init__(self, clock):
        """Initialize this instance."""
        self.discovered = 0
        self.completed = 0
        self.restart_progress_bubble_called = False

    def get_discovery_message(self):
        """Return the file discovery message."""
        self.discovered += 1
        return self.build_discovery_message()

    def build_discovery_message(self):
        """Build the file discovery message."""
        return "a lot of files found (%d).""" % self.discovered

    def get_progress_message(self):
        """Return the progress message."""
        self.completed += 1
        return self.build_progress_message()

    def build_progress_message(self):
        """Build the progress message."""
        params = (self.discovered, self.completed)
        return "a lot of files transferring (%d/%d).""" % params

    def get_final_status_message(self):
        """Return the final status message."""
        return "a lot of files completed."""

    def restart_progress_bubble(self):
        """Reset the progress bubble."""
        self.restart_progress_bubble_called = True


class FileDiscoveryBubbleTestCase(TestCase):
    """Test the FileDiscoveryBubble class."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "Notification", FakeNotificationSingleton())
        self.clock = PatchedClock()
        self.aggregator = FakeStatusAggregator(clock=self.clock)
        self.bubble = aggregator.FileDiscoveryBubble(self.aggregator,
                                                     clock=self.clock)
        self.addCleanup(self.bubble.cleanup)
        fdis = aggregator.FileDiscoveryGatheringState
        self.initial_delay = fdis.initial_delay
        self.smaller_delay = self.initial_delay * 0.8
        self.initial_timeout = fdis.initial_timeout
        fdus = aggregator.FileDiscoveryUpdateState
        self.updates_delay = fdus.updates_delay
        self.updates_timeout = fdus.updates_timeout
        fdss = aggregator.FileDiscoverySleepState
        self.sleep_delay = fdss.sleep_delay

        self.handler = MementoHandler()
        self.handler.setLevel(logging.DEBUG)
        aggregator.logger.addHandler(self.handler)
        aggregator.logger.setLevel(logging.DEBUG)
        self.addCleanup(aggregator.logger.removeHandler, self.handler)


    def get_notifications_shown(self):
        """The list of notifications shown."""
        return self.bubble.notification.notifications_shown

    def test_popup_shows_notification(self):
        """The popup callback shows notifications."""
        self.bubble.new_file_found()
        self.bubble._popup()
        message = self.aggregator.build_discovery_message()
        notification = (aggregator.UBUNTUONE_TITLE, message, None, False)
        self.assertIn(notification, self.get_notifications_shown())

    def test_popup_resets_progress_bubble(self):
        """The popup callback resets the progress bubble."""
        self.bubble.new_file_found()
        self.bubble._popup()
        self.assertTrue(self.aggregator.restart_progress_bubble_called)

    def test_notification_is_logged_in_debug(self):
        """The notification is printed in the debug log."""
        self.bubble.new_file_found()
        self.bubble._popup()
        msg = "notification shown: %s" % self.get_notifications_shown()[0][1]
        self.assertTrue(self.handler.check_debug(msg))

    def test_bubble_is_not_shown_initially(self):
        """The bubble is not shown initially."""
        self.bubble.new_file_found()
        self.assertEqual(0, len(self.get_notifications_shown()))

    def test_bubble_is_shown_after_delay(self):
        """The bubble is shown after a delay."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.assertEqual(1, len(self.get_notifications_shown()))

    def test_bubble_not_shown_if_more_files_found(self):
        """The bubble is not shown if more files found within delay."""
        self.clock.advance(self.smaller_delay)
        self.bubble.new_file_found()
        self.clock.advance(self.smaller_delay)
        self.assertEqual(0, len(self.get_notifications_shown()))

    def test_bubble_shown_if_timeout_exceeded(self):
        """The bubble is shown if the timeout is exceeded."""
        self.bubble.new_file_found()
        count = int(self.initial_timeout / self.smaller_delay) + 1
        for n in range(count):
            self.clock.advance(self.smaller_delay)
            self.bubble.new_file_found()
        self.assertEqual(1, len(self.get_notifications_shown()))

    def test_idle_state(self):
        """The idle state is verified."""
        self.assertEqual(type(self.bubble.state),
                                   aggregator.FileDiscoveryIdleState)

    def test_gathering_state(self):
        """The gathering state is set after the first file is found."""
        self.bubble.new_file_found()
        self.assertEqual(type(self.bubble.state),
                                   aggregator.FileDiscoveryGatheringState)

    def test_update_state(self):
        """When the gathering state finishes, the update state is started."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.assertEqual(type(self.bubble.state),
                                   aggregator.FileDiscoveryUpdateState)

    def test_sleeping_state(self):
        """When the update state finishes, the sleeping state is started."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.clock.advance(self.updates_timeout)
        self.assertEqual(type(self.bubble.state),
                                   aggregator.FileDiscoverySleepState)

    def test_back_to_initial_state(self):
        """When the last state finishes, we return to the idle state."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.clock.advance(self.updates_timeout)
        self.clock.advance(self.sleep_delay)
        self.assertEqual(type(self.bubble.state),
                                   aggregator.FileDiscoveryIdleState)

    def test_new_files_found_while_updating_not_shown_immediately(self):
        """New files found in the updating state are not shown immediately."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.bubble.new_file_found()
        self.assertEqual(1, len(self.get_notifications_shown()))

    def test_new_files_found_while_updating_are_shown_after_a_delay(self):
        """New files found in the updating state are shown after a delay."""
        self.bubble.new_file_found()
        self.clock.advance(self.initial_delay)
        self.bubble.new_file_found()
        self.clock.advance(self.updates_delay)
        self.assertEqual(2, len(self.get_notifications_shown()))

    def test_update_resets_progress_bubble(self):
        """The update callback resets the progress bubble."""
        self.bubble.new_file_found()
        self.bubble._popup()
        self.bubble.new_file_found()
        self.bubble._update()
        self.assertTrue(self.aggregator.restart_progress_bubble_called)

    def test_update_modifies_notification(self):
        """The update callback updates notifications."""
        self.bubble.new_file_found()
        self.bubble._popup()
        self.bubble.new_file_found()
        self.bubble._update()
        message = self.aggregator.build_discovery_message()
        notification = (aggregator.UBUNTUONE_TITLE, message, None, False)
        self.assertIn(notification, self.get_notifications_shown())

    def test_update_is_logged_in_debug(self):
        """The notification is logged when _update is called."""
        self.bubble.new_file_found()
        self.bubble._popup()
        self.bubble.new_file_found()
        self.bubble._update()
        msg = "notification updated: %s" % self.get_notifications_shown()[1][1]
        self.assertTrue(self.handler.check_debug(msg))


class ProgressBubbleTestCase(TestCase):
    """Tests for the progress bubble."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "Notification", FakeNotificationSingleton())
        self.clock = PatchedClock()
        self.aggregator = FakeStatusAggregator(clock=self.clock)
        self.bubble = aggregator.ProgressBubble(self.aggregator,
                                                clock=self.clock)
        self.addCleanup(self.bubble.cleanup)
        self.smaller_delay = aggregator.ProgressBubble.sleep_delay * 0.8
        self.right_delay = aggregator.ProgressBubble.sleep_delay
        self.longer_delay = aggregator.ProgressBubble.sleep_delay * 1.2

    def test_is_created_successfully(self):
        """The progress bubble is created and idle."""
        self.assertEqual(None, self.bubble.timer)

    def test_restart_creates_timer(self):
        """Restarting creates a timer."""
        self.bubble.restart()
        self.assertNotEqual(None, self.bubble.timer)

    def test_not_shown_initially(self):
        """The bubble is not shown initially."""
        self.bubble.restart()
        self.assertEqual(None, self.bubble.notification)

    def test_not_shown_immediately(self):
        """The bubble is not shown immediately."""
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.assertEqual(None, self.bubble.notification)

    def test_shown_after_delay(self):
        """The bubble is shown after the standard delay."""
        self.bubble.restart()
        self.clock.advance(self.longer_delay)
        self.assertEqual(1, len(self.bubble.notification.notifications_shown))

    def test_not_shown_if_restarted(self):
        """The delay is adjusted when restarting."""
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.assertEqual(None, self.bubble.notification)

    def test_delay_reached_after_restart(self):
        """The timeout is still reached after a restart."""
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.bubble.restart()
        self.clock.advance(self.longer_delay)
        self.assertEqual(1, len(self.bubble.notification.notifications_shown))

    def test_can_be_stopped(self):
        """The timeout can be stopped."""
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.bubble.stop()
        self.clock.advance(self.longer_delay)
        self.assertEqual(None, self.bubble.notification)

    def test_delay_reached_after_stopping_then_restarting(self):
        """The timeout is still reached after a stop followed by a restart."""
        self.bubble.restart()
        self.clock.advance(self.smaller_delay)
        self.bubble.stop()
        self.clock.advance(self.longer_delay)
        self.bubble.restart()
        self.clock.advance(self.longer_delay)
        self.assertEqual(1, len(self.bubble.notification.notifications_shown))

    def test_restarts_automatically(self):
        """The timer is restarted automatically."""
        self.bubble.restart()
        self.clock.advance(self.right_delay)
        self.assertEqual(1, len(self.bubble.notification.notifications_shown))
        self.clock.advance(self.right_delay)
        self.assertEqual(2, len(self.bubble.notification.notifications_shown))


class FinalBubbleTestCase(TestCase):
    """Test for the final status notification bubble."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "Notification", FakeNotificationSingleton())
        self.clock = PatchedClock()
        self.aggregator = FakeStatusAggregator(clock=self.clock)
        self.bubble = aggregator.FinalStatusBubble(self.aggregator)
        self.addCleanup(self.bubble.cleanup)

    def test_notification_not_shown_initially(self):
        """The notification is not shown initially."""
        self.assertEqual(None, self.bubble.notification)

    def test_show_pops_bubble(self):
        """The show method pops the bubble immediately."""
        self.bubble.show()
        self.assertEqual(1, len(self.bubble.notification.notifications_shown))


class FakeDelayedBuffer(object):
    """Appends all status pushed into a list."""
    timer_reset = False
    processed = False

    def __init__(self, *args, **kwargs):
        """Initialize this instance."""
        self.events = []

    def push_event(self, event):
        """Push an event into this buffer."""
        self.events.append(event)

    def reset_threshold_timer(self):
        """The status has changed."""
        self.timer_reset = True

    def process_accumulated(self):
        """Process accumulated events."""
        self.processed = True


class FakeCommand(object):
    """A fake command."""


class FakeVolumeManager(object):
    """A fake vm."""

    def __init__(self):
        """Initialize this instance."""
        self.volumes = {}

    def get_volume(self, volume_id):
        """Return a volume given its id."""
        return self.volumes[volume_id]


class FakeAggregator(object):
    """A fake aggregator object."""

    def __init__(self, clock):
        """Initialize this fake instance."""
        self.interesting_events = []
        self.queued_commands = set()

    def interesting_event(self, event):
        """A new interesting event was found."""
        self.interesting_events.append(event)

    def queue_done(self):
        """The queue completed all operations."""
        self.queued_commands.clear()

    def misc_command_queued(self, command):
        """A new command was queued."""
        self.queued_commands.add(command)

    def misc_command_unqueued(self, command):
        """A new command was unqueued."""
        self.queued_commands.discard(command)

    download_started = misc_command_queued
    download_finished = misc_command_unqueued
    upload_started = misc_command_queued
    upload_finished = misc_command_unqueued


class StatusFrontendTestCase(TestCase):
    """Test the status frontend."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "StatusAggregator", FakeAggregator)
        self.fakefsm = None
        self.fakevm = FakeVolumeManager()
        self.status_frontend = aggregator.StatusFrontend()
        self.listener = status_listener.StatusListener(self.fakefsm,
                                                       self.fakevm,
                                                       self.status_frontend)
    def test_file_published(self):
        """A file published event is processed."""
        share_id = "fake share id"
        node_id = "fake node id"
        is_public = True
        public_url = "http://fake_public/url"

        self.listener.handle_AQ_CHANGE_PUBLIC_ACCESS_OK(share_id, node_id,
                                                        is_public, public_url)
        event = self.status_frontend.aggregator.interesting_events[0]
        self.assertIsInstance(event, aggregator.FilePublishingStatus)

    def test_file_unpublished(self):
        """A file unpublished event is processed."""
        share_id = "fake share id"
        node_id = "fake node id"
        is_public = False
        public_url = None  # SD sends None when unpublishing

        self.listener.handle_AQ_CHANGE_PUBLIC_ACCESS_OK(share_id, node_id,
                                                        is_public, public_url)
        event = self.status_frontend.aggregator.interesting_events[0]
        self.assertIsInstance(event, aggregator.FileUnpublishingStatus)

    def test_download_started(self):
        """A download was added to the queue."""
        self.patch(status_listener.action_queue, "Download", FakeCommand)
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertIn(fake_command, qc)

    def test_download_finished(self):
        """A download was removed from the queue."""
        self.patch(status_listener.action_queue, "Download", FakeCommand)
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        self.listener.handle_SYS_QUEUE_REMOVED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertNotIn(fake_command, qc)

    def test_upload_started(self):
        """A upload was added to the queue."""
        self.patch(status_listener.action_queue, "Upload", FakeCommand)
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertIn(fake_command, qc)

    def test_upload_finished(self):
        """A upload was removed from the queue."""
        self.patch(status_listener.action_queue, "Upload", FakeCommand)
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        self.listener.handle_SYS_QUEUE_REMOVED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertNotIn(fake_command, qc)

    def test_queue_added(self):
        """A command was added to the queue."""
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertIn(fake_command, qc)

    def test_queue_removed(self):
        """A command has finished and is removed from the queue."""
        fake_command = FakeCommand()
        self.listener.handle_SYS_QUEUE_ADDED(fake_command)
        self.listener.handle_SYS_QUEUE_REMOVED(fake_command)
        qc = self.status_frontend.aggregator.queued_commands
        self.assertNotIn(fake_command, qc)

    def test_queue_done(self):
        """The queue is empty."""
        fake_command = FakeCommand()
        qc = self.status_frontend.aggregator.queued_commands
        qc.add(fake_command)
        self.listener.handle_SYS_QUEUE_DONE()
        self.assertEqual(0, len(qc))

    def test_new_share_available(self):
        """A new share is available for subscription."""
        SHARE_ID = "fake share id"
        share = Share(volume_id=SHARE_ID)
        self.fakevm.volumes[SHARE_ID] = share
        self.listener.handle_VM_SHARE_CREATED(SHARE_ID)
        event = self.status_frontend.aggregator.interesting_events[0]
        self.assertIsInstance(event, aggregator.ShareAvailableStatus)

    def test_new_udf_available(self):
        """A new udf is available for subscription."""
        udf = UDF()
        self.listener.handle_VM_UDF_CREATED(udf)
        event = self.status_frontend.aggregator.interesting_events[0]
        self.assertIsInstance(event, aggregator.UDFAvailableStatus)


class StatusEventTestCase(TestCase):

    """Test the status event class and children."""

    CLASS = aggregator.StatusEvent
    CLASS_ARGS = ()
    status = None

    def setUp(self):
        """Initialize this test instance."""
        if type(self) == StatusEventTestCase:
            self.assertRaises(AssertionError, self.CLASS, *self.CLASS_ARGS)
        else:
            self.status = self.CLASS(*self.CLASS_ARGS)

    def test_one_message_built_correctly(self):
        """The message returned by one() is returned ok."""
        if self.status:
            self.assertEqual(self.status.one(), self.CLASS.MESSAGE_ONE)

    def test_many_message_built_correctly(self):
        """The message returned by many() is returned ok."""
        if self.status:
            test_count = 99
            expected = self.CLASS.MESSAGE_MANY % test_count
            self.assertEqual(self.status.many(test_count), expected)


class FilePublishingStatusTestCase(StatusEventTestCase):
    """Test the file publishing status class."""

    CLASS = aggregator.FilePublishingStatus
    CLASS_ARGS = ("http://fake_public/url",)

    def test_one_message_built_correctly(self):
        """The message returned by one() should include the url."""
        expected = self.CLASS.MESSAGE_ONE % self.status.args
        self.assertEqual(self.status.one(), expected)


class FileUnpublishingStatusTestCase(StatusEventTestCase):
    """Test the file unpublishing status class."""

    CLASS = aggregator.FileUnpublishingStatus
    CLASS_ARGS = (None,)


class ShareAvailableEventTestCase(StatusEventTestCase):
    """Test the folder available status class with a Share."""

    FOLDER_NAME = "folder name"
    OTHER_USER_NAME = "person name"
    SAMPLE_SHARE = Share(accepted=False, name=FOLDER_NAME,
                         other_visible_name=OTHER_USER_NAME)
    CLASS = aggregator.ShareAvailableStatus
    CLASS_ARGS = (SAMPLE_SHARE,)

    def test_one_message_built_correctly(self):
        """one() must return the folder name and user name."""
        params = (self.FOLDER_NAME, self.OTHER_USER_NAME)
        expected = self.CLASS.MESSAGE_ONE % params
        self.assertEqual(self.status.one(), expected)


class UDFAvailableEventTestCase(StatusEventTestCase):
    """Test the folder available status class with a UDF."""

    FOLDER_NAME = "folder name"
    SAMPLE_UDF = UDF(subscribed=False, suggested_path=FOLDER_NAME)
    CLASS = aggregator.UDFAvailableStatus
    CLASS_ARGS = (SAMPLE_UDF,)

    def test_one_message_built_correctly(self):
        """one() must return the folder name."""
        params = (self.FOLDER_NAME,)
        expected = self.CLASS.MESSAGE_ONE % params
        self.assertEqual(self.status.one(), expected)


class FakeStatus(aggregator.StatusEvent):
    """A fake status to test weight comparisons."""

    def __init__(self, weight):
        """Initialize with the fake weight."""
        super(FakeStatus, self).__init__()
        self.WEIGHT = weight


class FakeFileDiscoveryBubble(object):
    """A fake FileDiscoveryBubble."""

    count = 0

    def __init__(self, status_aggregator, clock=None):
        """Initialize this instance."""
        self.status_aggregator = status_aggregator

    def new_file_found(self):
        """New files were found."""
        self.count += 1

    def cleanup(self):
        """Cleanup this instance."""


class FakeProgressBubble(object):
    """A fake ProgressBubble object."""

    started = False
    stopped = False

    def __init__(self, status_aggregator, clock=None):
        """Initialize this instance."""
        self.status_aggregator = status_aggregator

    def cleanup(self):
        """Cleanup this instance."""

    def restart(self):
        """Start this bubble waiting."""
        self.started = True

    def stop(self):
        """Make this bubble stop."""


class FakeFinalBubble(object):
    """A fake FinalStatusBubble."""

    shown = False

    def __init__(self, status_aggregator):
        """Initialize this fake instance."""
        self.status_aggregator = status_aggregator

    def cleanup(self):
        """Cleanup this instance."""

    def show(self):
        """Show this bubble."""
        self.shown = True


class StatusAggregatorTestCase(TestCase):
    """Test the backend of the status aggregator."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "FileDiscoveryBubble",
                   FakeFileDiscoveryBubble)
        self.patch(aggregator, "ProgressBubble",
                   FakeProgressBubble)
        self.patch(aggregator, "FinalStatusBubble",
                   FakeFinalBubble)
        self.status_frontend = aggregator.StatusFrontend()
        self.aggregator = self.status_frontend.aggregator
        self.fake_bubble = self.aggregator.file_discovery_bubble

        self.handler = MementoHandler()
        self.handler.setLevel(logging.DEBUG)
        aggregator.logger.addHandler(self.handler)
        aggregator.logger.setLevel(logging.DEBUG)
        self.addCleanup(aggregator.logger.removeHandler, self.handler)

    def assertStatusReset(self):
        """Assert that the status is at zero."""
        self.assertEqual(self.aggregator.download_total, 0)
        self.assertEqual(self.aggregator.download_done, 0)
        self.assertEqual(self.aggregator.upload_total, 0)
        self.assertEqual(self.aggregator.upload_done, 0)
        self.assertEqual(self.aggregator.total_counter, 0)
        self.assertEqual(self.aggregator.done_counter, 0)
        self.assertEqual(len(self.aggregator.files_uploading), 0)
        self.assertEqual(len(self.aggregator.files_downloading), 0)
        self.assertEqual(len(self.aggregator.interesting_events), 0)

    def assertMiscCommandQueued(self, fc):
        """Assert that some command was queued."""
        self.assertEqual(self.aggregator.total_counter, 1)
        message = "queueing command (0/1): %s" % fc.__class__.__name__
        self.assertTrue(self.handler.check_debug(message))

    def assertMiscCommandUnqueued(self, fc):
        """Assert that some command was unqueued."""
        self.assertEqual(self.aggregator.done_counter, 1)
        message = "unqueueing command (1/1): %s" % fc.__class__.__name__
        self.assertTrue(self.handler.check_debug(message))

    def test_counters_start_at_zero(self):
        """Test that the counters start at zero."""
        self.assertStatusReset()

    def test_misc_command_queue(self):
        """Test that a misc command was queued."""
        fc = FakeCommand()
        self.status_frontend.queue_added(fc)
        self.assertMiscCommandQueued(fc)

    def test_misc_command_unqueue(self):
        """Test that a misc command was unqueued."""
        fc = FakeCommand()
        self.status_frontend.queue_added(fc)
        self.status_frontend.queue_removed(fc)
        self.assertMiscCommandUnqueued(fc)

    def test_file_download_started(self):
        """Test that a file has started download."""
        fc = FakeCommand()
        self.status_frontend.download_started(fc)
        self.assertEqual(self.aggregator.download_total, 1)
        self.assertMiscCommandQueued(fc)
        self.assertEqual(1, self.fake_bubble.count)

    def test_file_download_finished(self):
        """Test that a file has finished downloading."""
        fc = FakeCommand()
        self.status_frontend.download_started(fc)
        self.status_frontend.download_finished(fc)
        self.assertEqual(self.aggregator.download_done, 1)
        self.assertMiscCommandUnqueued(fc)

    def test_file_upload_started(self):
        """Test that a file has started upload."""
        fc = FakeCommand()
        self.status_frontend.upload_started(fc)
        self.assertEqual(self.aggregator.upload_total, 1)
        self.assertMiscCommandQueued(fc)
        self.assertEqual(1, self.fake_bubble.count)

    def test_file_upload_finished(self):
        """Test that a file has finished uploading."""
        fc = FakeCommand()
        self.status_frontend.upload_started(fc)
        self.status_frontend.upload_finished(fc)
        self.assertEqual(self.aggregator.upload_done, 1)
        self.assertMiscCommandUnqueued(fc)

    def test_get_discovery_message(self):
        """Test the message that's shown on the discovery bubble."""
        uploading = 10
        downloading = 8
        self.aggregator.files_uploading.extend(range(uploading))
        self.aggregator.files_downloading.extend(range(downloading))
        expected = (aggregator.FILES_UPLOADING % uploading + "\n" +
                    aggregator.FILES_DOWNLOADING % downloading)
        result = self.aggregator.get_discovery_message()
        self.assertEqual(expected, result)

    def test_get_progress_message(self):
        """Test the message that's shown on the progress bubble."""
        self.aggregator.upload_done = 5
        self.aggregator.upload_total = 10
        self.aggregator.download_done = 3
        self.aggregator.download_total = 8
        self.aggregator.done_counter = 9
        self.aggregator.total_counter = 20
        expected = (aggregator.PROGRESS_UPLOADED % (5, 10) + " " +
                    aggregator.PROGRESS_DOWNLOADED % (3, 8) + " " +
                    aggregator.PROGRESS_COMPLETED % int(100.0 * 9 / 20))
        result = self.aggregator.get_progress_message()
        self.assertEqual(expected, result)

    def test_get_progress_message_no_uploads(self):
        """The progress message when no uploads are going on."""
        self.aggregator.upload_done = 0
        self.aggregator.upload_total = 0
        self.aggregator.download_done = 3
        self.aggregator.download_total = 8
        self.aggregator.done_counter = 9
        self.aggregator.total_counter = 20
        expected = (aggregator.PROGRESS_DOWNLOADED % (3, 8) + " " +
                    aggregator.PROGRESS_COMPLETED % int(100.0 * 9 / 20))
        result = self.aggregator.get_progress_message()
        self.assertEqual(expected, result)

    def test_get_progress_message_no_downloads(self):
        """The progress message when no downloads are going on."""
        self.aggregator.upload_done = 5
        self.aggregator.upload_total = 10
        self.aggregator.download_done = 0
        self.aggregator.download_total = 0
        self.aggregator.done_counter = 9
        self.aggregator.total_counter = 20
        expected = (aggregator.PROGRESS_UPLOADED % (5, 10) + " " +
                    aggregator.PROGRESS_COMPLETED % int(100.0 * 9 / 20))
        result = self.aggregator.get_progress_message()
        self.assertEqual(expected, result)

    def test_get_progress_message_no_total(self):
        """No progress message possible if total counter is zero."""
        self.aggregator.total_counter = 0
        self.assertRaises(AssertionError, self.aggregator.get_progress_message)

    def test_get_final_status_message(self):
        """The final status message."""
        done = (5, 10)
        self.aggregator.upload_done, self.aggregator.download_done = done
        expected = (aggregator.FINAL_COMPLETED + "\n" +
                    aggregator.FINAL_UPLOADED % done[0] + "\n" +
                    aggregator.FINAL_DOWNLOADED % done[1])
        result = self.aggregator.get_final_status_message()
        self.assertEqual(expected, result)

    def test_get_final_status_message_no_uploads(self):
        """The final status message when there were no uploads."""
        done = (0, 12)
        self.aggregator.upload_done, self.aggregator.download_done = done
        expected = (aggregator.FINAL_COMPLETED + "\n" +
                    aggregator.FINAL_DOWNLOADED % done[1])
        result = self.aggregator.get_final_status_message()
        self.assertEqual(expected, result)

    def test_get_final_status_message_no_downloads(self):
        """The final status message when there were no downloads."""
        done = (8, 0)
        self.aggregator.upload_done, self.aggregator.download_done = done
        expected = (aggregator.FINAL_COMPLETED + "\n" +
                    aggregator.FINAL_UPLOADED % done[0])
        result = self.aggregator.get_final_status_message()
        self.assertEqual(expected, result)

    def test_started_progress_bubble(self):
        """The progress bubble is started."""
        self.aggregator.restart_progress_bubble()
        self.assertTrue(self.aggregator.progress_bubble.started)

    def test_queue_done(self):
        """On queue done, show final bubble and reset counters."""
        fc = FakeCommand()
        self.status_frontend.upload_started(fc)
        old_final_bubble = self.aggregator.final_status_bubble
        self.aggregator.queue_done()
        self.assertTrue(old_final_bubble.shown)
        self.assertStatusReset()


class StatusGrouperTestCase(TestCase):
    """Tests for the group_statuses function."""

    def test_group_status(self):
        """The status grouper sorts and groups by weight."""
        status99 = FakeStatus(99)
        statuses = [
            status99,
            status99,
            FakeStatus(12),
            FakeStatus(1),
        ]

        result = [ list(k) for j, k in aggregator.group_statuses(statuses) ]
        expected = [
            [statuses[3]],
            [statuses[2]],
            [status99, status99],
        ]

        self.assertEqual(result, expected)


class HundredFeetTestCase(TestCase):
    """Try to make all parts work together."""

    def test_all_together_now(self):
        """Make all parts work together."""
        self.patch(aggregator, "Notification", FakeNotificationSingleton())
        clock = PatchedClock()
        upload = FakeCommand()
        sf = aggregator.StatusFrontend(clock=clock)
        clock.advance(aggregator.ProgressBubble.sleep_delay)
        sf.upload_started(upload)
        notifications_shown = (sf.aggregator.file_discovery_bubble.
                                             notification.notifications_shown)
        # no notifications shown yet
        self.assertEqual(0, len(notifications_shown))
        clock.advance(aggregator.FileDiscoveryGatheringState.initial_delay)
        # files found notification
        self.assertEqual(1, len(notifications_shown))
        download = FakeCommand()
        sf.download_started(download)
        self.assertEqual(1, len(notifications_shown))
        clock.advance(aggregator.FileDiscoveryUpdateState.updates_delay)
        # files count update
        self.assertEqual(2, len(notifications_shown))
        clock.advance(aggregator.FileDiscoveryUpdateState.updates_timeout -
                      aggregator.FileDiscoveryUpdateState.updates_delay)
        sf.upload_finished(upload)
        # the progress bubble has no notifications yet
        self.assertEqual(None, sf.aggregator.progress_bubble.notification)
        clock.advance(aggregator.ProgressBubble.sleep_delay)
        # the progress bubble is shown
        self.assertEqual(3, len(notifications_shown))
        sf.upload_finished(download)
        # progress, but the progress bubble is not updated immediately
        self.assertEqual(3, len(notifications_shown))
        clock.advance(aggregator.ProgressBubble.sleep_delay)
        # the progress bubble is updated after a delay
        self.assertEqual(4, len(notifications_shown))
        sf.queue_done()
        # the final bubble is shown immediately
        self.assertEqual(5, len(notifications_shown))
        clock.advance(aggregator.ProgressBubble.sleep_delay * 2)
        # no more notifications are show
        self.assertEqual(5, len(notifications_shown))
