# -*- coding: utf-8 -*-
#
# Author: John R. Lenton <john.lenton@canonical.com>
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
# Author: Facundo Batista <facundo@canonical.com>
#
# Copyright 2009, 2010, 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the action queue module."""

from __future__ import with_statement

import base64
import inspect
import logging
import operator
import os
import shutil
import unittest
import urllib2
import uuid

from functools import wraps

import dbus
from mocker import Mocker
from StringIO import StringIO
from oauth import oauth
from twisted.internet import defer, threads, reactor
from twisted.internet import error as twisted_error
from twisted.python.failure import DefaultException, Failure
from twisted.web import server
from twisted.trial.unittest import TestCase as TwistedTestCase
from zope.interface.verify import verifyObject, verifyClass

from contrib.testing.testcase import (
    BaseTwistedTestCase, DummyClass, FakeMain, FakeActionQueue,
)
from ubuntuone.devtools import handlers
from ubuntuone import logger
from ubuntuone.storageprotocol import (
    client, errors, protocol_pb2, request
)
from ubuntuone.syncdaemon import states, interfaces
from ubuntuone.syncdaemon.action_queue import (
    ActionQueue, ActionQueueCommand, ChangePublicAccess, CreateUDF,
    DeleteVolume, Download, ListVolumes, ActionQueueProtocol,
    RequestQueue, UploadProgressWrapper, Upload,
    CreateShare, DeleteShare, GetPublicFiles, GetDelta, GetDeltaFromScratch,
    TRANSFER_PROGRESS_THRESHOLD, Unlink, Move, MakeFile, MakeDir, DeltaList,
    ZipQueue, DeferredMap, ThrottlingStorageClient,
    UploadCompressionCancelled, ClientNoLongerThere, PathLockingTree,
)
from ubuntuone.syncdaemon.event_queue import EventQueue, EVENTS
from ubuntuone.syncdaemon.marker import MDMarker


PATH = u'~/Documents/pdfs/moño/'
NAME = u'UDF-me'
VOLUME = uuid.UUID('12345678-1234-1234-1234-123456789abc')
NODE = uuid.UUID('FEDCBA98-7654-3211-2345-6789ABCDEF12')
USER = u'Dude'
SHARE = uuid.uuid4()

NO_OP = lambda *args, **kwargs: None


def fire_and_check(f, deferred, check):
    """Callback a deferred."""
    @wraps(f)
    def inner(*args, **kwargs):
        """Execute f and fire the deferred."""
        result = f(*args, **kwargs)
        error = check()
        if not error:
            deferred.callback(True)
        else:
            deferred.errback(error)
        return result
    return inner


class MementoHandler(handlers.MementoHandler):
    """Wrapper to handle custom logger levels."""

    def check_note(self, *msgs):
        """Shortcut for checking in ERROR."""
        return self.check(logger.NOTE, *msgs)


class FakeCommand(object):
    """Yet another fake action queue command."""

    is_runnable = True
    cleaned = False
    resumed = False
    conditions_checked = False

    def __init__(self, share_id=None, node_id=None):
        self.share_id = share_id
        self.node_id = node_id
        self.cancelled = False
        self.log = logging.getLogger('ubuntuone.SyncDaemon')

    run = locked_run = lambda self: defer.succeed(True)

    def cleanup(self):
        """Mark as cleaned."""
        self.cleaned = True

    def resume(self):
        """Mark as resumed."""
        self.resumed = True

    @property
    def uniqueness(self):
        """Fake uniqueness."""
        if self.share_id is None and self.node_id is None:
            return self
        else:
            return (self.__class__.__name__, self.share_id, self.node_id)

    def cancel(self):
        """Cancel!"""
        self.cancelled = True

    def check_conditions(self):
        """Mark as checked."""
        self.conditions_checked = True


class FakedEventQueue(EventQueue):
    """Faked event queue."""

    def __init__(self, fs=None):
        """Initialize a faked event queue."""
        super(FakedEventQueue, self).__init__(fs=fs)
        self.events = []

    def push(self, event_name, **kwargs):
        """Faked event pushing."""
        self.events.append((event_name, kwargs))
        super(FakedEventQueue, self).push(event_name, **kwargs)


class FakedVolume(object):
    """Faked volume."""
    volume_id = None
    generation = None
    free_bytes = None


class FakeSemaphore(object):
    """Fake semaphore."""
    release = acquire = lambda self: None


class TestingProtocol(ActionQueue.protocol):
    """Protocol for testing."""

    def connectionMade(self):
        """connectionMade."""
        ActionQueue.protocol.connectionMade(self)

        # assure we're connected
        events = [x[0] for x in self.factory.event_queue.events]
        assert 'SYS_CONNECTION_MADE' in events

        self.factory.event_queue.events = [] # reset events
        if hasattr(self, 'testing_deferred'):
            self.testing_deferred.callback(True)

class TestActionQueue(ActionQueue):
    """AQ class that uses the testing protocol."""
    protocol = TestingProtocol


class BasicTestCase(BaseTwistedTestCase):
    """Basic test case for ActionQueue."""

    timeout = 5

    def setUp(self):
        """Init."""
        BaseTwistedTestCase.setUp(self)

        self.root = self.mktemp('root')
        self.home = self.mktemp('home')
        self.data = self.mktemp('data')
        self.shares = self.mktemp('shares')
        self.partials = self.mktemp('partials')

        # set up Fake Main to use the testing AQ, the port will
        # be decided later
        self._orig_fakemain_aq = (FakeMain._fake_AQ_class,
                                  FakeMain._fake_AQ_params)
        FakeMain._fake_AQ_class = TestActionQueue
        FakeMain._fake_AQ_params = ('127.0.0.1', 0, False)

        self.main = FakeMain(root_dir=self.root, shares_dir=self.shares,
                             data_dir=self.data, partials_dir=self.partials)

        self.action_queue = self.main.action_q
        self.action_queue.connection_timeout = 3
        self.action_queue.event_queue.events = []

        def keep_a_copy(f):
            """Keep a copy of the pushed events."""
            @wraps(f)
            def recording(event_name, **kwargs):
                """Keep a copy of the pushed events."""
                value = (event_name, kwargs)
                if event_name != 'SYS_STATE_CHANGED' and \
                   not event_name.startswith('VM_'):
                    self.action_queue.event_queue.events.append(value)
                return f(event_name, **kwargs)
            return recording

        self.main.event_q.push = keep_a_copy(self.main.event_q.push)

        self.handler = MementoHandler()
        self.handler.setLevel(logging.DEBUG)
        self._logger = logging.getLogger('ubuntuone.SyncDaemon')
        self._logger.addHandler(self.handler)

        dbus.service.BusName.__del__ = lambda _: None

    def tearDown(self):
        """Cleanup."""
        self._logger.removeHandler(self.handler)
        self.main.shutdown()

        # restore Fake Main settings
        cls, params = self._orig_fakemain_aq
        FakeMain._fake_AQ_class = cls
        FakeMain._fake_AQ_params = params

        shutil.rmtree(self.root)
        shutil.rmtree(self.shares)
        shutil.rmtree(self.data)
        shutil.rmtree(self.partials)

        for record in self.handler.records:
            exc_info = getattr(record, 'exc_info', None)
            if exc_info is not None:
                raise exc_info[0], exc_info[1], exc_info[2]

        BaseTwistedTestCase.tearDown(self)

    def user_connect(self):
        """User requested to connect to server."""
        token = {'token': 'bla', 'token_secret': 'ble',
                 'consumer_key': 'foo', 'consumer_secret': 'bar'}
        self.action_queue.event_queue.push('SYS_USER_CONNECT',
                                           access_token=token)


class BasicTests(BasicTestCase):
    """Basic tests to check ActionQueue."""

    def test_implements_interface(self):
        """Verify ActionQueue and FakeActionQueue interface."""
        verifyObject(interfaces.IActionQueue, self.action_queue)
        verifyClass(interfaces.IActionQueue, FakeActionQueue)

    @defer.inlineCallbacks
    def test_get_root_and_demark(self):
        """Get the received Root and demark its mdid."""
        # get the marker
        d = self.action_queue.uuid_map.get('mdid')

        # we received a root!
        self.main.event_q.push('SYS_ROOT_RECEIVED',
                               root_id='node_id', mdid='mdid')

        # it should be demarked with the root node_id
        root_node_id = yield d
        self.assertEqual(root_node_id, 'node_id')

    def test_cancelupload_calls_cancelop(self):
        """cancel_upload passes the correct args to the generic method."""
        called = []
        self.action_queue._cancel_op = lambda *a: called.append(a)
        self.action_queue.cancel_upload('share', 'node')
        self.assertEqual(called, [('share', 'node', Upload)])

    def test_canceldownload_calls_cancelop(self):
        """cancel_download passes the correct args to the generic method."""
        called = []
        self.action_queue._cancel_op = lambda *a: called.append(a)
        self.action_queue.cancel_download('share', 'node')
        self.assertEqual(called, [('share', 'node', Download)])

    def test_cancelop_nothing(self):
        """It does not cancel anything, because all empty."""
        assert not self.action_queue.queue.waiting
        self.action_queue._cancel_op('shr', 'node', Upload)
        self.assertTrue(self.handler.check_debug('starting', 'cancel_upload',
                                                 'shr', 'node'))
        self.assertFalse(self.handler.check_debug('cancelled', 'shr', 'node'))
        self.assertTrue(self.handler.check_debug('finished', 'shr', 'node'))

    def _set_queue(self, *waiting):
        """Set the content queue content."""
        cq = self.action_queue.queue
        for cmd in waiting:
            cq.waiting.append(cmd)
            cq.hashed_waiting[cmd.uniqueness] = cmd

    def test_cancelop_different_sharenode(self):
        """It does not cancel anything, because queue with different stuff."""
        cmd1 = FakeCommand('sh', 'nd1')
        cmd2 = FakeCommand('sh', 'nd2')
        self._set_queue(cmd1, cmd2)
        self.action_queue._cancel_op('sh', 'nd3', FakeCommand)
        self.assertFalse(self.handler.check_debug('cancelled', 'sh', 'nd3'))
        self.assertTrue(self.handler.check_debug('finished without cancel',
                                                 'sh', 'nd3'))
        self.assertFalse(cmd1.cancelled)
        self.assertFalse(cmd2.cancelled)

    def test_cancelop_different_operation(self):
        """It does not cancel anything, because queue with different stuff."""
        cmd1 = FakeCommand('sh', 'nd')
        cmd2 = FakeCommand('sh', 'nd')
        self._set_queue(cmd1, cmd2)
        self.action_queue._cancel_op('sh', 'nd', Upload)
        self.assertFalse(self.handler.check_debug('cancelled', 'sh', 'nd'))
        self.assertTrue(self.handler.check_debug('finished without cancel',
                                                 'sh', 'nd'))
        self.assertFalse(cmd1.cancelled)
        self.assertFalse(cmd2.cancelled)

    def test_cancelop_inwaiting(self):
        """Cancel something that is in the waiting queue."""
        cmd = FakeCommand('sh', 'nd')
        self._set_queue(cmd)
        self.action_queue._cancel_op('sh', 'nd', FakeCommand)
        self.assertTrue(self.handler.check_debug('cancelled', 'sh', 'nd'))
        self.assertFalse(self.handler.check_debug('finished without cancel',
                                                  'sh', 'nd'))
        self.assertTrue(cmd.cancelled)

    def test_node_is_queued_move_api(self):
        """Test that it calls the queue method."""
        called = []
        aq = self.action_queue
        aq.queue.node_is_queued = lambda *a: called.append(a)
        aq.node_is_with_queued_move('share', 'node')
        self.assertEqual(called, [(Move, 'share', 'node')])

    def test_node_is_queued_move_integration(self):
        """Kind of integration test for this method."""
        aq = self.action_queue
        cmd = Move(aq.queue, VOLUME, 'node', 'o_p', 'n_p', 'n_name',
                   "pathfrom", "pathto")
        self.assertFalse(aq.node_is_with_queued_move(VOLUME, 'node'))
        aq.queue.waiting.append(cmd)
        aq.queue.hashed_waiting[cmd.uniqueness] = cmd
        self.assertTrue(aq.node_is_with_queued_move(VOLUME, 'node'))

    def test_event_listener(self):
        """All event listeners should define methods with correct signature."""
        for evtname, evtargs in EVENTS.iteritems():
            meth = getattr(ActionQueue, 'handle_' + evtname, None)
            if meth is not None:
                defined_args = inspect.getargspec(meth)[0]
                self.assertEqual(defined_args[0], 'self')
                self.assertEqual(set(defined_args[1:]), set(evtargs))


class TestRequestQueue(TwistedTestCase):
    """Tests for the RequestQueue."""

    def setUp(self):
        """Set up."""

        class FakeAQ(object):
            """Fake AQ."""
            event_queue = self.eq = FakedEventQueue()

        self.rq = RequestQueue(action_queue=FakeAQ())

        # add a Memento handler to the logger
        self.log_handler = MementoHandler()
        self.log_handler.setLevel(logging.DEBUG)
        logging.getLogger('ubuntuone.SyncDaemon').addHandler(self.log_handler)

    def tearDown(self):
        """Tear down."""
        self.eq.shutdown()

    def _add_to_rq(self, *cmds):
        """Add the commands to rq.waiting and hashed_waiting."""
        for cmd in cmds:
            self.rq.waiting.append(cmd)
            self.rq.hashed_waiting[cmd.uniqueness] = cmd

    def test_len_nothing(self):
        """Len with nothing queued."""
        self.assertEqual(len(self.rq), 0)

    def test_len_waiting(self):
        """Len with something in the queue."""
        self.rq.waiting.append(1)
        self.assertEqual(len(self.rq), 1)

    def test_len_bigger(self):
        """Len with several commands."""
        self.rq.waiting.append(1)
        self.rq.waiting.append(1)
        self.assertEqual(len(self.rq), 2)

    def test_queue_adds_to_waiting(self):
        """Command queues is appended to waiting."""
        # set up
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.append(cmd1)

        # queue and test
        self.rq.queue(cmd2)
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])

    def test_queue_sends_changed_event(self):
        """Alert something changed."""
        cmd = FakeCommand()
        self.rq.queue(cmd)
        evt = ('SYS_QUEUE_ADDED', dict(command=cmd))
        self.assertIn(evt, self.eq.events)

    def test_queue_waiting_if_first(self):
        """It should send the WAITING signal ."""
        # set up
        cmd = FakeCommand()

        # queue and test
        self.rq.queue(cmd)
        self.assertEqual(list(self.rq.waiting), [cmd])
        self.assertIn(('SYS_QUEUE_WAITING', {}), self.eq.events)

    def test_queue_nowaiting_if_not_first(self):
        """It should not send the WAITING signal if no first cmd."""
        # set up
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.append(cmd1)

        # queue and test
        self.rq.queue(cmd2)
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])
        self.assertNotIn(('SYS_QUEUE_WAITING', {}), self.eq.events)

    def test_with_one_run(self):
        """Run will execute the command."""
        cmd = FakeCommand()
        self.rq.queue(cmd)
        self.assertIn(('SYS_QUEUE_WAITING', {}), self.eq.events)
        self.assertNotIn(('SYS_QUEUE_DONE', {}), self.eq.events)
        self.rq.unqueue(cmd)
        self.assertIn(('SYS_QUEUE_DONE', {}), self.eq.events)

    def test_with_two_run(self):
        """Run will execute both commands."""
        # first queuing, get the event
        cmd1 = FakeCommand()
        self.rq.queue(cmd1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_WAITING', {})), 1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_DONE', {})), 0)

        # second queuing, don't get the event
        cmd2 = FakeCommand()
        self.rq.queue(cmd2)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_WAITING', {})), 1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_DONE', {})), 0)

        # first run, no new events
        self.rq.unqueue(cmd1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_WAITING', {})), 1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_DONE', {})), 0)

        # second run, now we're done
        self.rq.unqueue(cmd2)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_WAITING', {})), 1)
        self.assertEqual(self.eq.events.count(('SYS_QUEUE_DONE', {})), 1)

    def test_init_notactive(self):
        """RQ borns not active."""
        self.assertFalse(self.rq.active)

    def test_run_goes_active(self):
        """Activate on run."""
        self.rq.run()
        self.assertTrue(self.rq.active)

    def test_run_resume_commands(self):
        """Resume all queued command on run."""
        # set up
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.extend((cmd1, cmd2))
        assert not cmd1.resumed and not cmd2.resumed

        # run and check
        self.rq.run()
        self.assertTrue(cmd1.resumed)
        self.assertTrue(cmd2.resumed)

    def test_stop_goes_inactive(self):
        """Desactivate on stop."""
        self.rq.active = True
        self.rq.stop()
        self.assertFalse(self.rq.active)

    def test_stop_clean_commands(self):
        """Clean up all queued commands on stop."""
        # set up
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.extend((cmd1, cmd2))
        assert not cmd1.cleaned and not cmd2.cleaned

        # stop and test
        self.rq.stop()
        self.assertTrue(cmd1.cleaned)
        self.assertTrue(cmd2.cleaned)

    def test_check_conditions(self):
        """Check all conditions on the commands."""
        # set up
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.extend((cmd1, cmd2))
        assert not cmd1.conditions_checked and not cmd2.conditions_checked

        # check conditions and test
        self.rq.check_conditions()
        self.assertTrue(cmd1.conditions_checked)
        self.assertTrue(cmd2.conditions_checked)

    def test_unqueue_remove(self):
        """Remove the command from queue on unqueue."""
        # set up a couple of commands
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.extend((cmd1, cmd2))
        self.rq.hashed_waiting[cmd1] = cmd1
        self.rq.hashed_waiting[cmd2] = cmd2

        # unqueue and check that 1 was removed from both structures and
        # that cmd2 is still there untouched
        self.rq.unqueue(cmd1)
        self.assertNotIn(cmd1, self.rq.waiting)
        self.assertIn(cmd2, self.rq.waiting)
        self.assertNotIn(cmd1, self.rq.hashed_waiting)
        self.assertIn(cmd2, self.rq.hashed_waiting)

    def test_unqueue_sysqueuedone_if_empty(self):
        """Send SYS_QUEUE_DONE if empty after unqueue."""
        # set up one command
        cmd = FakeCommand()
        self.rq.waiting.append(cmd)
        self.rq.hashed_waiting[cmd] = cmd

        # unqueue it and check
        self.rq.unqueue(cmd)
        self.assertIn(('SYS_QUEUE_DONE', {}), self.eq.events)

    def test_unqueue_sysqueuedone_if_not_empty(self):
        """Do not send SYS_QUEUE_DONE if not empty after unqueue."""
        # set up a couple of commands
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.waiting.extend((cmd1, cmd2))
        self.rq.hashed_waiting[cmd1] = cmd1
        self.rq.hashed_waiting[cmd2] = cmd2

        # unqueue only one and check
        self.rq.unqueue(cmd1)
        self.assertNotIn(('SYS_QUEUE_DONE', {}), self.eq.events)

    def test_unqueue_sends_changed_event(self):
        """Alert something changed."""
        cmd = FakeCommand()
        self.rq.waiting.append(cmd)
        self.rq.unqueue(cmd)
        evt = ('SYS_QUEUE_REMOVED', dict(command=cmd))
        self.assertIn(evt, self.eq.events)

    def test_remove_empty(self):
        """Don't remove if waiting is empty."""
        assert not self.rq.waiting, "test badly set up"
        cmd = FakeCommand()
        self.rq.remove(cmd)
        self.assertFalse(self.rq.waiting)
        self.assertFalse(self.rq.hashed_waiting)

    def test_remove_other(self):
        """Don't remove if waiting has other command."""
        cmd1 = FakeCommand(1, 2)
        self._add_to_rq(cmd1)
        cmd2 = FakeCommand(2, 3)
        self.rq.remove(cmd2)
        self.assertEqual(list(self.rq.waiting), [cmd1])
        self.assertEqual(self.rq.hashed_waiting.values(), [cmd1])

    def test_remove_command(self):
        """Remove for the command."""
        cmd = FakeCommand()
        self._add_to_rq(cmd)
        self.rq.remove(cmd)
        self.assertFalse(self.rq.waiting)
        self.assertFalse(self.rq.hashed_waiting)

    def test_remove_mixed(self):
        """Remove ok in a mixed situation."""
        cmd1 = FakeCommand(1, 2)
        cmd2 = FakeCommand(2, 3)
        cmd3 = FakeCommand(3, 4)
        self._add_to_rq(cmd1, cmd2, cmd3)
        self.rq.remove(cmd2)
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd3])
        self.assertEqual(set(self.rq.hashed_waiting.values()),
                         set([cmd1, cmd3]))

    def test_hashedwaiting_queue(self):
        """Queue a command and it will be added to hashed waiting."""
        cmd = FakeCommand()
        self.rq.queue(cmd)
        self.assertTrue(self.rq.hashed_waiting.values(), [cmd])

    def test_node_is_queued_nothing(self):
        """Test with empty queues."""
        self.assertFalse(self.rq.node_is_queued(Move, 'share', 'node'))

    def test_node_is_queued_waiting(self):
        """Test with a command in waiting."""
        cmd = FakeCommand('share', 'node')
        self._add_to_rq(cmd)
        self.assertTrue(self.rq.node_is_queued(FakeCommand, 'share', 'node'))

    def test_node_is_queued_different_command(self):
        """The node is queued, but other command on it."""
        cmd = FakeCommand('share', 'node')
        self._add_to_rq(cmd)
        self.assertFalse(self.rq.node_is_queued(Move, 'share', 'node'))

    def test_node_is_queued_different_node(self):
        """The command is queued, but on other node."""
        cmd = FakeCommand('share', 'node')
        self._add_to_rq(cmd)
        self.assertFalse(self.rq.node_is_queued(FakeCommand, 'share', 'other'))

    def test_len_empty(self):
        """Counter return that it's empty."""
        self.assertEqual(len(self.rq), 0)

    def test_len_with_one(self):
        """Counter return that it has one."""
        cmd = FakeCommand()
        self.rq.queue(cmd)
        self.assertEqual(len(self.rq), 1)

    def test_len_with_two(self):
        """Counter return that it has two."""
        cmd = FakeCommand()
        self.rq.queue(cmd)
        self.rq.queue(cmd)
        self.assertEqual(len(self.rq), 2)

    def test_len_run_decreases(self):
        """Counter behaviour when adding/running."""
        cmd1 = FakeCommand()
        cmd2 = FakeCommand()
        self.rq.queue(cmd1)
        self.assertEqual(len(self.rq), 1)
        self.rq.queue(cmd2)
        self.assertEqual(len(self.rq), 2)
        self.rq.unqueue(cmd1)
        self.assertEqual(len(self.rq), 1)
        self.rq.unqueue(cmd2)
        self.assertEqual(len(self.rq), 0)


class TestDeferredMap(TwistedTestCase):
    """Test the deferred map."""

    def setUp(self):
        """Set up."""
        self.dm = DeferredMap()

    def test_one_get_returns_stored_deferred(self):
        """Get will return the stored deferred."""
        d = self.dm.get('foo')
        self.assertEqual(self.dm.waiting, {'foo': [d]})

    def test_two_gets_returns_second_deferred_other_key(self):
        """A second get for other key will return other deferred."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('bar')
        self.assertEqual(self.dm.waiting, {'foo': [d1], 'bar': [d2]})

    def test_two_gets_returns_second_deferred_same_key(self):
        """A second get for the same key will return other deferred."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('foo')
        self.assertEqual(self.dm.waiting, {'foo': [d1, d2]})

    def test_mixed_gets(self):
        """Several gets with different keys."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('bar')
        d3 = self.dm.get('foo')
        d4 = self.dm.get('baz')
        self.assertEqual(self.dm.waiting,
                         {'foo': [d1, d3], 'bar': [d2], 'baz': [d4]})

    def test_set_to_nothing(self):
        """It's ok to set a key that is not being waited."""
        self.dm.set('not there', 'value')

    @defer.inlineCallbacks
    def test_set_fires_deferred_single(self):
        """The set fires the unique waiting deferred with the value."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('bar')
        d3 = self.dm.get('foo')
        self.assertEqual(self.dm.waiting, {'foo': [d1, d3], 'bar': [d2]})

        self.dm.set('bar', 'value')
        res = yield d2
        self.assertEqual(res, 'value')
        self.assertEqual(self.dm.waiting, {'foo': [d1, d3]})

    @defer.inlineCallbacks
    def test_set_fires_deferred_multiple(self):
        """The set fires the multiple waiting deferreds with the value."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('bar')
        d3 = self.dm.get('foo')
        self.assertEqual(self.dm.waiting, {'foo': [d1, d3], 'bar': [d2]})

        self.dm.set('foo', 'value')
        res1 = yield d1
        res2 = yield d3
        self.assertEqual(res1, 'value')
        self.assertEqual(res2, 'value')
        self.assertEqual(self.dm.waiting, {'bar': [d2]})

    def test_err_to_nothing(self):
        """It's ok to err a key that is not being waited."""
        self.dm.err('not there', 'failure')

    @defer.inlineCallbacks
    def test_err_fires_deferred_single(self):
        """The set fires the unique waiting deferred with the failure."""
        d1 = self.dm.get('foo')
        d2 = self.dm.get('bar')
        d3 = self.dm.get('foo')
        self.assertEqual(self.dm.waiting, {'foo': [d1, d3], 'bar': [d2]})

        exc = Exception('problem!')
        self.dm.err('bar', Failure(exc))
        try:
            yield d2
        except Exception, e:
            self.assertEqual(e, exc)
        else:
            self.fail("It didn't fired the deferred with a failure!")
        self.assertEqual(self.dm.waiting, {'foo': [d1, d3]})


class TestZipQueue(TwistedTestCase):
    """Test the zipping queue."""

    def setUp(self):
        """Set up."""
        self.zq = ZipQueue()

    @defer.inlineCallbacks
    def test_zip_calls_compress_in_thread(self):
        """Test that self._compress is called in another thread."""
        upload = FakeCommand()
        upload.fileobj_factory = lambda: "foo"

        def fake_compress(deferred, _upload, fileobj):
            """Fake the _compress method."""
            self.assertEqual(upload, _upload)
            self.assertEqual(fileobj, "foo")
            deferred.callback(True)

        self.zq._compress = fake_compress
        yield self.zq.zip(upload)

    @defer.inlineCallbacks
    def test_zip_calls_compress_with_file_object(self):
        """Test that _compress is called with the result of fileobj factory."""
        upload = FakeCommand()
        upload.fileobj_factory = lambda: "foo"

        def fake_compress(deferred, upload, fileobj):
            """Fake the _compress method."""
            deferred.callback(True)

        self.zq._compress = fake_compress
        yield self.zq.zip(upload)

    @defer.inlineCallbacks
    def test_fileobj_factory_error_is_logged(self):
        """Log the error when fileobj_factory fails."""
        def crash():
            """Crash!"""
            raise ValueError("foo")

        upload = FakeCommand()
        upload.fileobj_factory = crash

        # set up the logger
        self.handler = MementoHandler()
        self.handler.setLevel(logging.DEBUG)
        upload.log.addHandler(self.handler)

        try:
            yield self.zq.zip(upload)
        except UploadCompressionCancelled:
            # consume the error
            pass

        self.assertTrue(self.handler.check_warning("Unable to build fileobj",
                                                   "ValueError", "foo"))

    @defer.inlineCallbacks
    def test_fileobj_factory_error_cancels_upload(self):
        """Cancel the upload when fileobj_factory fails."""
        upload = FakeCommand()

        try:
            yield self.zq.zip(upload)
        except UploadCompressionCancelled:
            # consume the error
            pass

        self.assertTrue(upload.cancelled)

    @defer.inlineCallbacks
    def test_fileobj_factory_error_raises(self):
        """Raise the exception when fileobj_factory fails."""
        upload = FakeCommand()
        try:
            yield self.zq.zip(upload)
        except UploadCompressionCancelled:
            pass
        else:
            self.fail("Test should have raised UploadCompressionCancelled")

    @defer.inlineCallbacks
    def test_zip_acquire_lock(self):
        """Test that it acquires the lock."""
        called = []
        self.zq._compress = lambda deferred, upl, fobj: deferred.callback(True)

        def fake_acquire():
            """Fake the acquire method."""
            self.zq.tokens = 1
            called.append(True)
            return defer.succeed(True)

        self.zq.acquire = fake_acquire
        upload = FakeCommand()
        upload.fileobj_factory = lambda: "foo"
        yield self.zq.zip(upload)
        self.assertTrue(called)

    @defer.inlineCallbacks
    def test_zip_release_lock_ok(self):
        """Test that it releases the lock when all ok."""
        called = []
        self.zq._compress = lambda deferred, upl, fobj: deferred.callback(True)
        self.zq.release = lambda: called.append(True)

        upload = FakeCommand()
        upload.fileobj_factory = lambda: "foo"
        yield self.zq.zip(upload)
        self.assertTrue(called)

    @defer.inlineCallbacks
    def test_zip_release_lock_error(self):
        """Test that it releases the lock even on error."""
        called = []
        exc = Exception('bad')
        self.zq._compress = lambda deferred, upl, fobj: deferred.errback(exc)
        self.zq.release = lambda: called.append(True)
        upload = FakeCommand()
        upload.fileobj_factory = lambda: "foo"

        try:
            yield self.zq.zip(upload)
        except Exception, e:
            # need to silent the exception we're generating in the test
            self.assertEqual(e, exc)
        self.assertTrue(called)


class FactoryBaseTestCase(BasicTestCase):
    """Helper for by-pass Twisted."""

    def _start_sample_webserver(self):
        """Start a web server serving content at its root"""
        # start listening on `decide yourself` port, and fix AQ with it
        website = server.Site(None)
        webport = reactor.listenTCP(0, website)
        self.action_queue.port = webport.getHost().port

        transport_class = webport.transport
        def save_an_instance(skt, protocol, addr, sself, s, sreactor):
            self.server_transport = transport_class(skt, protocol, addr, sself,
                                                    s, sreactor)
            return self.server_transport
        webport.transport = save_an_instance

        self.addCleanup(webport.stopListening)
        return webport

    def _connect_factory(self):
        """Connect the instance factory."""
        self.server = self._start_sample_webserver()
        orig = self.action_queue.buildProtocol

        d = defer.Deferred()
        def faked_buildProtocol(*args, **kwargs):
            """Override buildProtocol to hook a deferred."""
            protocol = orig(*args, **kwargs)
            protocol.testing_deferred = d
            return protocol

        self.action_queue.buildProtocol = faked_buildProtocol
        self.action_queue.connect()
        return d

    def _disconnect_factory(self):
        """Disconnect the instance factory."""
        if self.action_queue.client is not None:
            orig = self.action_queue.client.connectionLost

            d = defer.Deferred()
            def faked_connectionLost(reason):
                """Receive connection lost and fire tearDown."""
                orig(reason)
                d.callback(True)

            self.action_queue.client.connectionLost = faked_connectionLost
        else:
            d = defer.succeed(True)

        if self.action_queue.connect_in_progress:
            self.action_queue.disconnect()

        return d


class ConnectionTestCase(FactoryBaseTestCase):
    """Test TCP/SSL connection mechanism for ActionQueue."""

    def assert_connection_state_reset(self):
        """Test connection state is properly reset."""
        self.assertTrue(self.action_queue.client is None)
        self.assertTrue(self.action_queue.connector is None)
        self.assertEqual(False, self.action_queue.connect_in_progress)

    def test_init(self):
        """Test connection init state."""
        self.assert_connection_state_reset()

    @defer.inlineCallbacks
    def test_connect_if_already_connected(self):
        """Test that double connections are avoided."""
        yield self._connect_factory()

        assert self.action_queue.connector is not None
        assert self.action_queue.connect_in_progress == True
        # double connect, it returns None instead of a Deferred
        result = self.action_queue.connect()
        self.assertTrue(result is None, 'not connecting again')

        yield self._disconnect_factory()

    @defer.inlineCallbacks
    def test_disconnect_if_connected(self):
        """self.action_queue.connector.disconnect was called."""
        yield self._connect_factory()

        self.action_queue.event_queue.events = [] # cleanup events
        assert self.action_queue.connector.state == 'connected'
        self.action_queue.disconnect()

        self.assert_connection_state_reset()
        self.assertEqual([], self.action_queue.event_queue.events)

        yield self._disconnect_factory()

    @defer.inlineCallbacks
    def test_clientConnectionFailed(self):
        """Test clientConnectionFailed.

        The connection will not be completed since the server will be down. So,
        self.action_queue.connector will never leave the 'connecting' state.
        When interrupting the connection attempt, twisted automatically calls
        self.action_queue.clientConnectionFailed.

        """
        self.action_queue.event_queue.events = []
        orig = self.action_queue.clientConnectionFailed

        d = defer.Deferred()
        def faked_clientConnectionFailed(connector, reason):
            """Receive connection failed and check."""
            orig(connector, reason)
            self.assert_connection_state_reset()
            self.assertEqual([('SYS_CONNECTION_FAILED', {})],
                             self.action_queue.event_queue.events)
            self.action_queue.clientConnectionFailed = orig
            d.callback(True)

        self.action_queue.clientConnectionFailed = faked_clientConnectionFailed
        # factory will never finish the connection, server was never started
        self.action_queue.connect()
        # stopConnecting() will be called since the connection is in progress
        assert self.action_queue.connector.state == 'connecting'
        self.action_queue.connector.disconnect()

        yield d

    @defer.inlineCallbacks
    def test_clientConnectionLost(self):
        """Test clientConnectionLost

        The connection will be completed successfully.
        So, self.action_queue.connector will be in the 'connected' state.
        When disconnecting the connector, twisted automatically calls
        self.action_queue.clientConnectionLost.

        """
        yield self._connect_factory()

        self.action_queue.event_queue.events = []
        orig = self.action_queue.clientConnectionLost

        d = defer.Deferred()
        def faked_clientConnectionLost(connector, reason):
            """Receive connection lost and check."""
            orig(connector, reason)
            self.assert_connection_state_reset()
            self.assertEqual([('SYS_CONNECTION_LOST', {})],
                             self.action_queue.event_queue.events)
            self.action_queue.clientConnectionLost = orig
            d.callback(True)

        self.action_queue.clientConnectionLost = faked_clientConnectionLost
        # loseConnection() will be called since the connection was completed
        assert self.action_queue.connector.state == 'connected'
        self.action_queue.connector.disconnect()
        yield d

        yield self._disconnect_factory()

    @defer.inlineCallbacks
    def test_server_disconnect(self):
        """Test factory's connection when the server goes down."""

        yield self._connect_factory()

        self.action_queue.event_queue.events = []
        orig = self.action_queue.clientConnectionLost

        d = defer.Deferred()
        def faked_connectionLost(*args, **kwargs):
            """Receive connection lost and check."""
            orig(*args, **kwargs)
            self.assert_connection_state_reset()
            self.assertEqual([('SYS_CONNECTION_LOST', {})],
                             self.action_queue.event_queue.events)
            self.action_queue.clientConnectionLost = orig
            d.callback(True)

        self.action_queue.clientConnectionLost = faked_connectionLost
        # simulate a server failure!
        yield self.server_transport.loseConnection()
        yield d
        yield self._disconnect_factory()

    def test_buildProtocol(self):
        """Test buildProtocol."""
        protocol = self.action_queue.buildProtocol(addr=None)
        self.assertTrue(protocol is self.action_queue.client)
        self.assertTrue(self.action_queue is self.action_queue.client.factory)

        # callbacks are connected
        # pylint: disable-msg=W0212
        aq = self.action_queue
        self.assertEqual(aq.client._share_change_callback,
                         aq._share_change_callback)
        self.assertEqual(aq.client._share_answer_callback,
                         aq._share_answer_callback)
        self.assertEqual(aq.client._free_space_callback,
                         aq._free_space_callback)
        self.assertEqual(aq.client._account_info_callback,
                         aq._account_info_callback)
        self.assertEqual(aq.client._volume_created_callback,
                         aq._volume_created_callback)
        self.assertEqual(aq.client._volume_deleted_callback,
                         aq._volume_deleted_callback)
        self.assertEqual(aq.client._volume_new_generation_callback,
                         aq._volume_new_generation_callback)

    @defer.inlineCallbacks
    def test_connector_gets_assigned_on_connect(self):
        """Test factory's connector gets assigned on connect."""
        yield self._connect_factory()

        self.assertTrue(self.action_queue.connector is not None)

        yield self._disconnect_factory()


class NetworkmanagerTestCase(FactoryBaseTestCase):
    """Base test case generating a connected factory."""

    timeout = 15

    def setUp(self):
        """Init."""
        super(NetworkmanagerTestCase, self).setUp()
        self.action_queue.event_queue.push('SYS_NET_CONNECTED')
        self.main.start()

    @defer.inlineCallbacks
    def test_wrong_disconnect(self):
        """Test factory's connection when SYS_NET_DISCONNECTED."""

        d1 = self.main.wait_for('SYS_CONNECTION_MADE')
        d2 = self.main.wait_for('SYS_CONNECTION_LOST')

        self.server = self._start_sample_webserver()
        self.user_connect()
        yield d1

        self.action_queue.event_queue.push('SYS_NET_DISCONNECTED')
        yield d2

    @defer.inlineCallbacks
    def test_disconnect_twice(self):
        """Test connection when SYS_NET_DISCONNECTED is received twice."""

        d1 = self.main.wait_for('SYS_CONNECTION_MADE')
        d2 = self.main.wait_for('SYS_CONNECTION_LOST')

        self.server = self._start_sample_webserver()
        self.user_connect()
        yield d1

        self.action_queue.event_queue.push('SYS_NET_DISCONNECTED')
        yield d2

        self.action_queue.event_queue.events = []
        self.action_queue.event_queue.push('SYS_NET_DISCONNECTED')
        self.assertEqual([('SYS_NET_DISCONNECTED', {})],
                         self.action_queue.event_queue.events,
                        'No new events after a misplaced SYS_NET_DISCONNECTED')


    @defer.inlineCallbacks
    def test_net_connected_if_already_connected(self):
        """Test connection when SYS_NET_CONNECTED is received twice."""

        d1 = self.main.wait_for('SYS_CONNECTION_MADE')

        self.server = self._start_sample_webserver()
        self.user_connect()
        yield d1

        self.action_queue.event_queue.events = []
        self.action_queue.event_queue.push('SYS_NET_CONNECTED')
        self.assertEqual([('SYS_NET_CONNECTED', {})],
                         self.action_queue.event_queue.events,
                         'No new events after a misplaced SYS_NET_CONNECTED')

    @defer.inlineCallbacks
    def test_messy_mix(self):
        """Test connection when a messy mix of events is received."""
        orig_waiting = states.MAX_WAITING
        states.MAX_WAITING = 1

        self.action_queue.event_queue.events = []
        self.server = self._start_sample_webserver()

        conn_made = self.main.wait_for('SYS_CONNECTION_MADE')
        self.user_connect()
        yield conn_made

        events = ['SYS_NET_CONNECTED', 'SYS_NET_DISCONNECTED',
                  'SYS_NET_CONNECTED', 'SYS_NET_CONNECTED',
                  'SYS_NET_DISCONNECTED', 'SYS_NET_DISCONNECTED',
                  'SYS_NET_CONNECTED']

        for i in events:
            self.action_queue.event_queue.push(i)

        yield self.main.wait_for_nirvana()

        expected = ['SYS_NET_CONNECTED', # from the DBus fake NetworkManager
                    'SYS_NET_CONNECTED', 'SYS_NET_DISCONNECTED',
                    'SYS_CONNECTION_LOST', 'SYS_CONNECTION_RETRY',
                    'SYS_NET_CONNECTED', 'SYS_NET_CONNECTED',
                    'SYS_CONNECTION_MADE', 'SYS_NET_DISCONNECTED',
                    'SYS_NET_DISCONNECTED']

        avoid = ('SYS_STATE_CHANGED', 'SYS_LOCAL_RESCAN_DONE',
                 'SYS_PROTOCOL_VERSION_OK', 'SYS_SET_CAPABILITIES_OK',
                 'SYS_AUTH_OK', 'SYS_SERVER_RESCAN_DONE')
        actual = [event for (event, kwargs) in
                  self.action_queue.event_queue.events
                  if event not in avoid]
        self.assertEqual(sorted(expected), sorted(actual))

        states.MAX_WAITING = orig_waiting


class ConnectedBaseTestCase(FactoryBaseTestCase):
    """Base test case generating a connected factory."""

    @defer.inlineCallbacks
    def setUp(self):
        """Init."""
        FactoryBaseTestCase.setUp(self)
        yield self._connect_factory()
        assert self.action_queue.connector.state == 'connected'

    @defer.inlineCallbacks
    def tearDown(self):
        """Clean up."""
        yield self._disconnect_factory()
        FactoryBaseTestCase.tearDown(self)

    def silent_connection_lost(self, failure):
        """Some tests will generate connection lost, support it."""
        if not failure.check(twisted_error.ConnectionDone):
            return failure


class VolumeManagementTestCase(ConnectedBaseTestCase):
    """Test Volume managemenr for ActionQueue."""

    def setUp(self):
        """Init."""
        res = super(VolumeManagementTestCase, self).setUp()

        self.patch(self.action_queue.main.vm,
                   'handle_SV_VOLUME_DELETED', NO_OP)

        return res

    def test_volume_created_push_event(self):
        """Volume created callback push proper event."""
        volume = FakedVolume()
        self.action_queue._volume_created_callback(volume)
        self.assertEqual([('SV_VOLUME_CREATED', {'volume': volume})],
                          self.action_queue.event_queue.events)

    def test_volume_deleted_push_event(self):
        """Volume deleted callback push proper event."""
        volume_id = VOLUME
        self.action_queue._volume_deleted_callback(volume_id)
        self.assertEqual([('SV_VOLUME_DELETED', {'volume_id': volume_id})],
                          self.action_queue.event_queue.events)

    def test_volume_new_generation_push_event_root(self):
        """Volume New Generation callback push proper event with root."""
        volume = request.ROOT
        self.action_queue._volume_new_generation_callback(volume, 77)
        event = ('SV_VOLUME_NEW_GENERATION',
                 {'volume_id': volume, 'generation': 77})
        self.assertTrue(event in self.action_queue.event_queue.events)

    def test_volume_new_generation_push_event_uuid(self):
        """Volume New Generation callback push proper event with uuid."""
        volume = uuid.uuid4()
        self.action_queue._volume_new_generation_callback(volume, 77)
        event = ('SV_VOLUME_NEW_GENERATION',
                 {'volume_id': volume, 'generation': 77})
        self.assertTrue(event in self.action_queue.event_queue.events)

    def test_valid_events(self):
        """Volume events are valid in EventQueue."""
        new_events = ('SV_VOLUME_CREATED', 'SV_VOLUME_DELETED',
                      'AQ_CREATE_UDF_OK', 'AQ_CREATE_UDF_ERROR',
                      'AQ_LIST_VOLUMES', 'AQ_LIST_VOLUMES_ERROR',
                      'AQ_DELETE_VOLUME_OK', 'AQ_DELETE_VOLUME_ERROR',
                      'SV_VOLUME_NEW_GENERATION')
        for event in new_events:
            self.assertTrue(event in EVENTS)

        self.assertEqual(('volume',), EVENTS['SV_VOLUME_CREATED'])
        self.assertEqual(('volume_id',), EVENTS['SV_VOLUME_DELETED'])
        self.assertEqual(('volume_id', 'node_id', 'marker'),
                          EVENTS['AQ_CREATE_UDF_OK'])
        self.assertEqual(('error', 'marker'), EVENTS['AQ_CREATE_UDF_ERROR'])
        self.assertEqual(('volumes',), EVENTS['AQ_LIST_VOLUMES'])
        self.assertEqual(('error',), EVENTS['AQ_LIST_VOLUMES_ERROR'])
        self.assertEqual(('volume_id',), EVENTS['AQ_DELETE_VOLUME_OK'])
        self.assertEqual(('volume_id', 'error',),
                         EVENTS['AQ_DELETE_VOLUME_ERROR'])
        self.assertEqual(('volume_id', 'generation',),
                         EVENTS['SV_VOLUME_NEW_GENERATION'])

    def test_create_udf(self):
        """Test volume creation."""
        path = PATH
        name = NAME
        res = self.action_queue.create_udf(path, name, marker=None)
        self.assertTrue(res is None) # this is what start returns

    def test_list_volumes(self):
        """Test volume listing."""
        res = self.action_queue.list_volumes()
        self.assertTrue(res is None) # this is what start returns

    def test_delete_volume(self):
        """Test volume deletion."""
        volume_id = VOLUME
        res = self.action_queue.delete_volume(volume_id, 'path')
        self.assertTrue(res is None) # this is what start returns


class ActionQueueCommandTestCase(ConnectedBaseTestCase):
    """Test for the generic functionality of ActionQueueCommand."""

    def setUp(self):
        """Set up."""
        res = super(ActionQueueCommandTestCase, self).setUp()

        class MyCommand(ActionQueueCommand):
            logged_attrs = ('a', 'b', 'c', 'd')
            a = 3
            b = 'foo'
            c = u'año'

            def _run(self):
                return defer.succeed(True)

            @property
            def uniqueness(self):
                return (self.a, self.b, self.c)

        self.rq = RequestQueue(action_queue=self.action_queue)
        self.rq.active = True
        self.cmd = MyCommand(self.rq)
        self.cmd.log = self.cmd.make_logger()

        return res

    def test_runnable(self):
        """All commands are runnable by default."""
        self.assertTrue(self.cmd.is_runnable)

    def test_cancelled(self):
        """All commands are not cancelled by default."""
        self.assertFalse(self.cmd.cancelled)

    def test_dump_to_dict(self):
        """Test to dict dumping."""
        d = self.cmd.to_dict()
        self.assertEqual(d, dict(a=3, b='foo', c=u'año', d=None))

    @defer.inlineCallbacks
    def test_demark_not_marker(self):
        """Test demark with not a marker."""
        self.cmd.possible_markers = 'foo',
        self.cmd.foo = 'not a marker'
        yield self.cmd.demark()
        self.assertEqual(self.cmd.foo, 'not a marker')

    @defer.inlineCallbacks
    def test_demark_with_marker_future(self):
        """Test demark with a marker not ready.

        Here, on purpose, set up everything and trigger later.
        """
        marker = MDMarker('foo')
        self.cmd.possible_markers = 'foo',
        self.cmd.foo = marker
        d = self.cmd.demark()

        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d
        self.assertEqual(self.cmd.foo, 'node_id')
        self.assertTrue(self.handler.check_debug(
                        "waiting for the real value of marker:foo"))

    @defer.inlineCallbacks
    def test_demark_with_marker_ready(self):
        """Test demark with a marker that had data."""
        marker = MDMarker('foo')
        self.cmd.possible_markers = 'foo',
        self.cmd.foo = marker
        d = self.cmd.demark()
        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d
        self.assertEqual(self.cmd.foo, 'node_id')
        self.assertTrue(self.handler.check_debug(
                        "waiting for the real value of marker:foo"))

    @defer.inlineCallbacks
    def test_demark_mixed_markers(self):
        """Test demark with both a marker and not."""
        # call demark with both
        marker = MDMarker('foo')
        self.cmd.possible_markers = 'foo', 'bar'
        self.cmd.foo = 'notamarker'
        self.cmd.bar = marker
        d = self.cmd.demark()
        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d

        # check
        self.assertEqual(self.cmd.foo, 'notamarker')
        self.assertEqual(self.cmd.bar, 'node_id')
        self.assertTrue(self.handler.check_debug(
                        "waiting for the real value of marker:foo"))
        self.assertFalse(self.handler.check_debug(
                         "waiting for the real value of 'notamarker'"))

    @defer.inlineCallbacks
    def test_demark_marker_future_got_ok(self):
        """Test demark getting a marker triggered ok later."""
        # don't have the info now
        marker = MDMarker('foo')
        self.cmd.possible_markers = 'foo',
        self.cmd.foo = marker
        d = self.cmd.demark()
        self.assertFalse(self.handler.check_debug("for marker:foo"))

        # set and check
        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d
        self.assertEqual(self.cmd.foo, 'node_id')
        self.assertTrue(self.handler.check_debug(
                        "for marker:foo got value 'node_id'"))

    @defer.inlineCallbacks
    def test_demark_marker_future_got_failure(self):
        """Test demark getting a marker triggered with failure later."""
        # don't have the info now
        marker = MDMarker('foo')
        self.cmd.possible_markers = 'foo',
        self.cmd.foo = marker
        d = self.cmd.demark()
        self.assertFalse(self.handler.check_error("failed marker:foo"))

        # set the marker and check
        self.action_queue.uuid_map.err(marker, Failure(Exception('bad')))
        yield d
        self.assertTrue(self.handler.check_error("failed marker:foo"))
        try:
            yield self.cmd.markers_resolved_deferred
        except Exception, e:
            self.assertEqual(str(e), 'bad')
        else:
            self.fail("An exception should have been raised!")

    @defer.inlineCallbacks
    def test_demark_two_markers_ok(self):
        """Test demark with two markers that finish ok."""
        # call demark with both
        marker1 = MDMarker('foo')
        marker2 = MDMarker('bar')
        self.cmd.possible_markers = 'foo', 'bar'
        self.cmd.foo = marker1
        self.cmd.bar = marker2
        d = self.cmd.demark()
        self.action_queue.uuid_map.set(marker1, 'data1')
        self.action_queue.uuid_map.set(marker2, 'data2')
        yield d

        # check
        self.assertEqual(self.cmd.foo, 'data1')
        self.assertEqual(self.cmd.bar, 'data2')
        yield self.cmd.markers_resolved_deferred

    @defer.inlineCallbacks
    def test_demark_two_markers_one_fail(self):
        """Test demark with two markers that one ends in failure."""
        # call demark with both
        marker1 = MDMarker('foo')
        marker2 = MDMarker('bar')
        self.cmd.possible_markers = 'foo', 'bar'
        self.cmd.foo = marker1
        self.cmd.bar = marker2
        d = self.cmd.demark()
        self.action_queue.uuid_map.set(marker1, 'data ok')
        self.action_queue.uuid_map.err(marker2, Failure(Exception('data bad')))
        yield d

        # check
        try:
            yield self.cmd.markers_resolved_deferred
        except Exception, e:
            self.assertEqual(str(e), 'data bad')
        else:
            self.fail("An exception should have been raised!")

    @defer.inlineCallbacks
    def test_demark_fixes_hashedwaiting_active(self):
        """The attribute changes: it also need to change the hashed_waiting."""
        marker = MDMarker('b')
        self.cmd.possible_markers = 'b',
        self.cmd.b = marker
        queue = self.cmd._queue
        queue.hashed_waiting[self.cmd.uniqueness] = self.cmd
        d = self.cmd.demark()

        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d
        self.assertTrue(queue.hashed_waiting[self.cmd.uniqueness], self.cmd)

    @defer.inlineCallbacks
    def test_demark_fixes_hashedwaiting_cancelled(self):
        """The attribute changes: no change because cancelled."""
        marker = MDMarker('b')
        self.cmd.possible_markers = 'b',
        self.cmd.b = marker
        self.cmd.cancelled = True
        queue = self.cmd._queue
        queue.hashed_waiting[self.cmd.uniqueness] = self.cmd
        d = self.cmd.demark()

        self.action_queue.uuid_map.set(marker, 'node_id')
        yield d
        self.assertFalse(self.cmd.uniqueness in queue.hashed_waiting)

    @defer.inlineCallbacks
    def test_run_call_start_if_not_started(self):
        """Call ._start if it's not started."""
        called = []
        self.cmd._start = lambda: called.append(True) or defer.succeed(True)
        self.cmd.start_done = False
        self.cmd.markers_resolved_deferred.callback(True)
        yield self.cmd.run()
        self.assertTrue(called)
        self.assertTrue(self.cmd.start_done)

    def test_run_dont_call_start_if_started(self):
        """Do not call ._start if it's started."""
        called = []
        self.cmd._start = lambda: called.append(True) or defer.Deferred()
        self.cmd.start_done = True
        self.cmd.run()
        self.assertFalse(called)

    def test_run_dont_call_start_if_not_runnable(self):
        """Do not call ._start if not runnable."""
        called = []
        self.cmd._start = lambda: called.append(True)
        self.cmd.start_done = False
        self.cmd.running = True
        self.cmd.is_runnable = False

        # run and check
        self.cmd.run()
        self.assertFalse(called)
        self.assertFalse(self.cmd.running)
        self.assertTrue(self.handler.check_debug(
                        'not running because of conditions'))

    def test_run_waits_markers_dereferencing(self):
        """Don't call _run_command until have the markers."""
        called = []
        self.cmd._run_command = lambda: called.append(True)
        self.cmd.run()
        self.assertFalse(called)
        self.cmd.markers_resolved_deferred.callback(True)
        self.assertTrue(called)

    def test_run_not_enderrback_if_ok(self):
        """If _start is ok, do not call end_errback."""
        called = []
        self.cmd.end_errback = lambda _: called.append(True)
        # default _start will always succeed
        self.cmd.run()
        self.assertFalse(called)

    def test_run_enderrback_if_problem(self):
        """If _start fails, call end_errback."""
        called = []
        self.cmd.end_errback = lambda _: called.append(True)
        self.cmd._start = lambda: defer.fail(Exception('foo'))
        self.cmd.run()
        self.assertTrue(called)

    @defer.inlineCallbacks
    def test_run_done_running_if_ok(self):
        """If _start is ok, done running."""
        # default _start will always succeed
        self.cmd.markers_resolved_deferred.callback(True)
        yield self.cmd.run()
        self.assertFalse(self.cmd.running)

    def test_run_done_running_if_problem_in_start(self):
        """If _start fails, done running."""
        called = []
        self.cmd.end_errback = lambda _: called.append(True)
        self.cmd._start = lambda: defer.fail(Exception('foo'))
        self.cmd.run()
        self.assertFalse(self.cmd.running)
        self.assertTrue(called)

    def test_run_done_running_if_problem_in_markers(self):
        """If deferring markers fails, done running."""
        called = []
        self.cmd.end_errback = lambda _: called.append(True)
        self.cmd.markers_resolved_deferred.errback(Failure(Exception('foo')))
        self.cmd.run()
        self.assertFalse(self.cmd.running)
        self.assertTrue(called)

    def test_queue_calls_prequeued(self):
        """Do the pre-queue when queued."""
        called = []
        self.cmd.pre_queue_setup = lambda: called.append(True)
        self.cmd.queue()
        self.assertTrue(called)

    def test_prequeued_calls_demark(self):
        """Pre-queue implies dereferencing markers."""
        called = []
        self.cmd.demark = lambda: called.append(True)
        self.cmd.pre_queue_setup()
        self.assertTrue(called)

    @defer.inlineCallbacks
    def test_start_default(self):
        """Default _start just returns a triggered deferred and sets done."""
        yield self.cmd._start()

    def test_possible_markers_default(self):
        """Default value for possible markers."""
        self.assertEqual(self.cmd.possible_markers, ())

    @defer.inlineCallbacks
    def test_marker_storing(self):
        """Pre-queueing deferres the marker storing."""
        self.cmd.pre_queue_setup()
        yield self.cmd.markers_resolved_deferred

    @defer.inlineCallbacks
    def test_path_locking(self):
        """Test it has a generic _acquire_pathlock."""
        r = yield self.cmd._acquire_pathlock()
        self.assertIdentical(r, None)

    def test_locked_run_locks_and_run_if_active(self):
        """Test that acquires the lock and runs the command."""
        called = []
        d = defer.Deferred()
        self.cmd._acquire_pathlock = lambda: called.append(1) or d
        self.cmd.run = lambda: called.append(2)

        # call and check that the lock was acquired
        self.cmd.locked_run()
        self.assertEqual(called, [1])

        # release the lock, it's running now, and check run() is called
        d.callback(3)
        self.assertTrue(self.cmd.running)
        self.assertEqual(called, [1, 2])

        # check that it stored the release function
        self.assertEqual(self.cmd.pathlock_release, 3)

    def test_locked_run_locks_and_run_if_notactive(self):
        """Test that acquires the lock and dont' run the command."""
        called = []
        d = defer.Deferred()
        self.cmd._acquire_pathlock = lambda: called.append(1) or d
        self.cmd.run = lambda: called.append(2)
        self.rq.active = False

        # call and check that the lock was acquired
        self.cmd.locked_run()
        self.assertEqual(called, [1])

        # release the lock, it's running now, and check run() is not called
        d.callback(3)
        self.assertTrue(self.cmd.running)
        self.assertEqual(called, [1])

        # check that it stored the release function
        self.assertEqual(self.cmd.pathlock_release, 3)

    def test_locked_run_locks_if_cancelled_exit_release(self):
        """Test that acquires the lock and dont' run the command."""
        called = []
        d = defer.Deferred()
        self.cmd._acquire_pathlock = lambda: called.append(1) or d
        self.cmd.run = lambda: called.append('should not')

        # call and check that the lock was acquired
        self.cmd.locked_run()
        self.assertEqual(called, [1])

        # cancel the command!
        self.cmd.cancelled = True

        # release the lock, it's running now, and check run() is not called
        d.callback(lambda: called.append(2))
        self.assertEqual(called, [1, 2])

    def test_endcallback_calls_finishing_stuff_if_running(self):
        """Call finish on end_callback."""
        called = []
        self.cmd.finish = lambda: called.append(1)
        self.cmd.handle_success = lambda a: called.append(a)
        self.cmd.running = True
        self.cmd.end_callback('foo')

        # check that handle_success was called *before* finish
        self.assertEqual(called, ['foo', 1])
        self.assertTrue(self.handler.check_debug('success'))

    def test_endcallback_dont_call_finishing_stuff_if_not_running(self):
        """Call finish on end_callback."""
        called = []
        self.cmd.finish = lambda: called.append(True)
        self.cmd.handle_success = lambda a: called.append(True)
        assert not self.cmd.running
        self.cmd.end_callback('foo')
        self.assertFalse(called)
        self.assertTrue(self.handler.check_debug('not running, so no success'))

    def test_enderrback_calls_finish(self):
        """Call finish on end_errback."""
        called = []
        self.cmd.finish = lambda: called.append(1)
        self.cmd.handle_failure = lambda f: called.append(f)
        failure = Failure(ValueError)
        self.cmd.end_errback(failure)

        # check that handle_failure was called *before* finish
        self.assertEqual(called, [failure, 1])

    def test_enderrback_retry_active(self):
        """Command retried with active queue, don't call finish and retry."""
        called = []
        self.cmd.finish = lambda: called.append(True)
        self.cmd.handle_failure = lambda: called.append(True)
        assert self.rq.active

        # run is called on retry, need to wait for it
        d = defer.Deferred()
        self.cmd.run = lambda: d.callback(True)

        # run and check finish was not called
        self.cmd.end_errback(Failure(ClientNoLongerThere()))
        self.assertFalse(called)
        return d

    def test_enderrback_retry_not_active(self):
        """Command not retried with no active queue, don't call nothing."""
        called = []
        self.cmd.finish = lambda: called.append(True)
        self.cmd.handle_failure = lambda: called.append(True)
        self.cmd.run = lambda: called.append(True)
        self.rq.active = False
        self.cmd.end_errback(Failure(ClientNoLongerThere()))

        # check nothing was called, but logged
        self.assertFalse(called)
        self.assertTrue(self.handler.check_debug(
                        'not retrying because queue not active'))

    def test_finish_releases_pathlock(self):
        """Have a release method, release the pathlock."""
        called = []
        self.cmd.pathlock_release = lambda: called.append(True)
        self.rq.unqueue = lambda c: None # don't do anything
        self.cmd.finish()
        self.assertTrue(called)

    def test_finish_dont_release_pathlock_nofunction(self):
        """Don't have a release method, don't release the pathlock."""
        assert self.cmd.pathlock_release is None
        self.rq.unqueue = lambda c: None # don't do anything
        self.cmd.finish()

    def test_finish_running(self):
        """Set running to False when finish."""
        self.cmd.running = True
        self.rq.unqueue = lambda c: None # don't do anything
        self.cmd.finish()
        self.assertFalse(self.cmd.running)

    def test_finish_unqueue(self):
        """Unqueue the command when finish."""
        called = []
        self.rq.unqueue = lambda c: called.append(c)
        self.cmd.finish()
        self.assertEqual(called, [self.cmd])

    def test_resume_run_in_certain_state(self):
        """Call run only if running and queue active."""
        self.cmd.run = lambda: called.append(True)

        for run in [True, False]:
            for act in [True, False]:
                called = []
                self.cmd.running = run
                self.rq.active = act
                self.cmd.resume()

                if run and act:
                    self.assertTrue(called)
                else:
                    self.assertFalse(called)

    def test_check_conditions_run_in_certain_state(self):
        """Call run only if not running and queue active."""
        self.cmd.run = lambda: called.append(True)

        for run in [True, False]:
            for act in [True, False]:
                called = []
                self.cmd.running = run
                self.rq.active = act
                self.cmd.check_conditions()

                if not run and act:
                    self.assertTrue(called)
                    self.assertTrue(self.cmd.running) # also set to running
                else:
                    self.assertFalse(called)

    def test_cancel_works(self):
        """Do default cleaning."""
        called = []
        self.cmd.cleanup = lambda: called.append(1)
        self.cmd.finish = lambda: called.append(2)
        assert not self.cmd.cancelled
        self.cmd.cancel()
        self.assertEqual(called, [1, 2])
        self.assertTrue(self.cmd.cancelled)


class CreateUDFTestCase(ConnectedBaseTestCase):
    """Test for CreateUDF ActionQueueCommand."""

    def setUp(self):
        """Init."""
        res = super(CreateUDFTestCase, self).setUp()

        request_queue = RequestQueue(action_queue=self.action_queue)
        self.marker = VOLUME
        self.command = CreateUDF(request_queue, PATH, NAME, marker=self.marker)

        vm = self.action_queue.main.vm
        self.patch(vm, 'handle_AQ_CREATE_UDF_OK', NO_OP)
        self.patch(vm, 'handle_AQ_CREATE_UDF_ERROR', NO_OP)

        return res

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_init(self):
        """Test creation."""
        self.assertEqual(PATH, self.command.path)
        self.assertEqual(NAME, self.command.name)
        self.assertEqual(self.marker, self.command.marker)

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's create_udf is called."""
        original = self.command.action_queue.client.create_udf
        self.called = False

        def check(path, name):
            """Take control over client's feature."""
            self.called = True
            self.assertEqual(PATH, path)
            self.assertEqual(NAME, name)

        self.command.action_queue.client.create_udf = check

        self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.create_udf = original

    def test_handle_success_push_event(self):
        """Test AQ_CREATE_UDF_OK is pushed on success."""
        request = client.CreateUDF(self.action_queue.client, PATH, NAME)
        request.volume_id = VOLUME
        request.node_id = NODE
        self.command.handle_success(success=request)
        events = [('AQ_CREATE_UDF_OK', {'volume_id': VOLUME,
                                            'node_id': NODE,
                                            'marker': self.marker})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_CREATE_UDF_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        events = [('AQ_CREATE_UDF_ERROR',
                    {'error': msg, 'marker': self.marker})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        self.command._acquire_pathlock()
        self.assertEqual(t, [tuple(PATH.split("/")), {'logger': None}])


class ActionQueueCommandErrorsTestCase(ConnectedBaseTestCase):
    """Test the error handling in ActionQueueCommand."""

    def setUp(self):
        res = super(ActionQueueCommandErrorsTestCase, self).setUp()

        self.deferred = defer.Deferred()

        class MyLogger(object):
            """Fake logger that just stores error and warning calls."""
            def __init__(self):
                self.logged = None

            def error(self, *a):
                """Mark that this method was called."""
                self.logged = "error"

            def warn(self, *a):
                """Mark that this method was called."""
                self.logged = "warn"

            def debug(self, *a):
                """Nothing."""

        class MyCommand(ActionQueueCommand):
            """Inherit ACQ to provide a retry signaller and a custom log."""
            # class-closure, cannot use self, pylint: disable-msg=E0213
            def __init__(innerself, request_queue):
                super(MyCommand, innerself).__init__(request_queue)
                innerself.log = MyLogger()

            def run(innerself):
                """Signal the retry."""
                self.deferred.callback(True)

        self.rq = RequestQueue(action_queue=self.action_queue)
        self.rq.unqueue = lambda c: None
        self.rq.active = True
        self.command = MyCommand(self.rq)
        return res

    def test_suppressed_yes_knownerrors(self):
        """Check that the log is in warning for the known errors."""
        # desactive the queue to never call retry
        self.rq.active = False

        def send_failure_and_check(errnum, exception_class):
            """Send the failure."""
            # prepare what to send
            protocol_msg = protocol_pb2.Message()
            protocol_msg.type = protocol_pb2.Message.ERROR
            protocol_msg.error.type = errnum
            err = exception_class("request", protocol_msg)

            # set up and test
            self.command.log.logged = None
            self.command.end_errback(failure=Failure(err))
            self.assertEqual(self.command.log.logged, "warn",
                             "Bad log in exception %s" % (exception_class,))

        known_errors = [x for x in errors._error_mapping.items()
                        if x[1] != errors.InternalError]
        for errnum, exception_class in known_errors:
            send_failure_and_check(errnum, exception_class)

    def test_suppressed_no_internalerror(self):
        """Check that the log is in error for InternalError."""
        # prepare what to send
        protocol_msg = protocol_pb2.Message()
        protocol_msg.type = protocol_pb2.Message.ERROR
        protocol_msg.error.type = protocol_pb2.Error.INTERNAL_ERROR
        err = errors.InternalError("request", protocol_msg)

        # set up and test
        self.command.end_errback(failure=Failure(err))
        self.assertEqual(self.command.log.logged, "error")

    def test_suppressed_yes_cancelled(self):
        """Check that the log is in warning for Cancelled."""
        err = errors.RequestCancelledError("CANCELLED")
        self.command.end_errback(failure=Failure(err))
        self.assertEqual(self.command.log.logged, "warn")

    def test_suppressed_yes_and_retry_when_connectiondone(self):
        """Check that the log is in warning and retries for ConnectionDone."""
        self.command.running = True
        err = twisted_error.ConnectionDone()
        self.command.end_errback(failure=Failure(err))
        self.assertEqual(self.command.log.logged, "warn")
        return self.deferred

    def test_retry_connectionlost(self):
        """Check that it retries when ConnectionLost."""
        self.command.running = True
        err = twisted_error.ConnectionLost()
        self.command.end_errback(failure=Failure(err))
        return self.deferred

    def test_retry_tryagain(self):
        """Check that it retries when TryAgain."""
        # prepare what to send
        self.command.running = True
        protocol_msg = protocol_pb2.Message()
        protocol_msg.type = protocol_pb2.Message.ERROR
        protocol_msg.error.type = protocol_pb2.Error.TRY_AGAIN
        err = errors.TryAgainError("request", protocol_msg)

        # set up and test
        self.command.end_errback(failure=Failure(err))
        return self.deferred

    def test_retry_clientnolongerthere(self):
        """Check that it retries when ConnectionLost."""
        self.command.running = True
        self.command.end_errback(failure=Failure(ClientNoLongerThere()))
        return self.deferred


class ListVolumesTestCase(ConnectedBaseTestCase):
    """Test for ListVolumes ActionQueueCommand."""

    def setUp(self):
        """Init."""
        res = super(ListVolumesTestCase, self).setUp()

        request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = ListVolumes(request_queue)

        return res

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's list_volumes is called."""
        original = self.command.action_queue.client.list_volumes
        self.called = False

        def check():
            """Take control over client's feature."""
            self.called = True

        self.command.action_queue.client.list_volumes = check

        self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.list_volumes = original

    def test_handle_success_push_event(self):
        """Test AQ_LIST_VOLUMES is pushed on success."""
        request = client.ListVolumes(self.action_queue.client)
        request.volumes = [FakedVolume(), FakedVolume()]
        self.command.handle_success(success=request)
        event = ('AQ_LIST_VOLUMES', {'volumes': request.volumes})
        self.assertIn(event, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_LIST_VOLUMES_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        event = ('AQ_LIST_VOLUMES_ERROR', {'error': msg})
        self.assertIn(event, self.command.action_queue.event_queue.events)


class DeleteVolumeTestCase(ConnectedBaseTestCase):
    """Test for DeleteVolume ActionQueueCommand."""

    def setUp(self):
        """Init."""
        res = super(DeleteVolumeTestCase, self).setUp()

        request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = DeleteVolume(request_queue, VOLUME, PATH)

        vm = self.action_queue.main.vm
        self.patch(vm, 'handle_AQ_DELETE_VOLUME_OK', NO_OP)
        self.patch(vm, 'handle_AQ_DELETE_VOLUME_ERROR', NO_OP)

        return res

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_init(self):
        """Test creation."""
        self.assertEqual(VOLUME, self.command.volume_id)

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's delete_volume is called."""
        original = self.command.action_queue.client.delete_volume
        self.called = False

        def check(volume_id):
            """Take control over client's feature."""
            self.called = True
            self.assertEqual(VOLUME, volume_id)

        self.command.action_queue.client.delete_volume = check

        self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.delete_volume = original

    def test_handle_success_push_event(self):
        """Test AQ_DELETE_VOLUME_OK is pushed on success."""
        request = client.DeleteVolume(self.action_queue.client,
                                      volume_id=VOLUME)
        self.command.handle_success(success=request)
        events = [('AQ_DELETE_VOLUME_OK', {'volume_id': VOLUME})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_DELETE_VOLUME_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        events = [('AQ_DELETE_VOLUME_ERROR',
                  {'volume_id': VOLUME, 'error': msg})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        self.command._acquire_pathlock()
        self.assertEqual(t, [tuple(PATH.split("/")), {'logger': None}])


class FilterEventsTestCase(BasicTestCase):
    """Tests for event filtering when a volume is not of our interest."""

    def setUp(self):
        """Init."""
        BasicTestCase.setUp(self)
        self.vm = self.main.vm
        self.old_home = os.environ.get('HOME', None)
        os.environ['HOME'] = self.home

    def tearDown(self):
        """Clean up."""
        if self.old_home is None:
            os.environ.pop('HOME')
        else:
            os.environ['HOME'] = self.old_home

        BasicTestCase.tearDown(self)


class ChangePublicAccessTests(ConnectedBaseTestCase):
    """Tests for the ChangePublicAccess ActionQueueCommand."""

    def setUp(self):
        super(ChangePublicAccessTests, self).setUp()
        self.user_connect()
        request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = ChangePublicAccess(request_queue, VOLUME, NODE, True)

    def test_change_public_access(self):
        """Test the change_public_access method.."""
        res = self.action_queue.change_public_access(VOLUME, NODE, True)
        self.assertTrue(res is None) # this is what start returns

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_init(self):
        """Test creation."""
        self.assertEqual(VOLUME, self.command.share_id)
        self.assertEqual(NODE, self.command.node_id)
        self.assertEqual(True, self.command.is_public)

    def test_run_defers_work_to_thread(self):
        """Test that work is deferred to a thread."""
        original = threads.deferToThread
        self.called = False

        def check(function):
            self.called = True
            self.assertEqual(
                self.command._change_public_access_http, function)
            return defer.Deferred()

        threads.deferToThread = check
        try:
            res = self.command._run()
        finally:
            threads.deferToThread = original

        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        self.assertTrue(self.called, "deferToThread was called")

    def test_change_public_access_http(self):
        """Test the blocking portion of the command."""
        self.called = False
        def check(request):
            self.called = True
            url = 'https://one.ubuntu.com/files/api/set_public/%s:%s' % (
                base64.urlsafe_b64encode(VOLUME.bytes).strip("="),
                base64.urlsafe_b64encode(NODE.bytes).strip("="))
            self.assertEqual(url, request.get_full_url())
            self.assertEqual("is_public=True", request.get_data())
            return StringIO(
                '{"is_public": true, "public_url": "http://example.com"}')

        from ubuntuone.syncdaemon import action_queue
        action_queue.urlopen = check
        try:
            res = self.command._change_public_access_http()
        finally:
            action_queue.urlopen = urllib2.urlopen

        self.assertEqual(
            {'is_public': True, 'public_url': 'http://example.com'}, res)

    def test_handle_success_push_event(self):
        """Test AQ_CHANGE_PUBLIC_ACCESS_OK is pushed on success."""
        response = {'is_public': True, 'public_url': 'http://example.com'}
        self.command.handle_success(success=response)
        event = ('AQ_CHANGE_PUBLIC_ACCESS_OK',
                 {'share_id': VOLUME, 'node_id': NODE, 'is_public': True,
                  'public_url': 'http://example.com'})
        self.assertIn(event, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_CHANGE_PUBLIC_ACCESS_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(urllib2.HTTPError(
                "http://example.com", 500, "Error", [], StringIO(msg)))
        self.command.handle_failure(failure=failure)
        event = ('AQ_CHANGE_PUBLIC_ACCESS_ERROR',
                 {'share_id': VOLUME, 'node_id': NODE, 'error': msg})
        self.assertIn(event, self.command.action_queue.event_queue.events)

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        res = [getattr(self.command, x) for x in self.command.possible_markers]
        self.assertEqual(res, [NODE])


class GetPublicFilesTestCase(ConnectedBaseTestCase):
    """Tests for GetPublicFiles ActionQueueCommand."""

    def setUp(self):
        super(GetPublicFilesTestCase, self).setUp()
        self.user_connect()
        request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = GetPublicFiles(request_queue)

    def test_init(self):
        """Test __init__ method."""
        default_url = 'https://one.ubuntu.com/files/api/public_files'
        request_queue = RequestQueue(action_queue=self.action_queue)
        command = GetPublicFiles(request_queue)
        self.assertEqual(command._url, default_url)
        custom_url = 'http://example.com:1234/files/api/public_files'
        command_2 = GetPublicFiles(request_queue,
                                   base_url='http://example.com:1234')
        self.assertEqual(command_2._url, custom_url)

    def test_change_public_access(self):
        """Test the get_public_files method.."""
        res = self.action_queue.get_public_files()
        self.assertTrue(res is None) # this is what start returns

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_run_defers_work_to_thread(self):
        """Test that work is deferred to a thread."""
        original = threads.deferToThread
        self.called = False

        def check(function):
            self.called = True
            self.assertEqual(
                self.command._get_public_files_http, function)
            return defer.Deferred()

        threads.deferToThread = check
        try:
            res = self.command._run()
        finally:
            threads.deferToThread = original

        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        self.assertTrue(self.called, "deferToThread was called")

    def test_get_public_files_http(self):
        """Test the blocking portion of the command."""
        self.called = False
        node_id = uuid.uuid4()
        nodekey = '%s' % (base64.urlsafe_b64encode(node_id.bytes).strip("="))
        node_id_2 = uuid.uuid4()
        nodekey_2 = '%s' % (base64.urlsafe_b64encode(
                                                node_id_2.bytes).strip("="))
        volume_id = uuid.uuid4()
        def check(request):
            self.called = True
            url = 'https://one.ubuntu.com/files/api/public_files'
            self.assertEqual(url, request.get_full_url())
            return StringIO(
                '[{"nodekey": "%s", "volume_id": null,"public_url": '
                '"http://example.com"}, '
                '{"nodekey": "%s", "volume_id": "%s", "public_url": '
                '"http://example.com"}]' % (nodekey, nodekey_2, volume_id))

        from ubuntuone.syncdaemon import action_queue
        action_queue.urlopen = check
        try:
            res = self.command._get_public_files_http()
        finally:
            action_queue.urlopen = urllib2.urlopen
        self.assertEqual([{'node_id': str(node_id), 'volume_id': '',
                          'public_url': 'http://example.com'},
                          {'node_id': str(node_id_2),
                           'volume_id': str(volume_id),
                           'public_url': 'http://example.com'}], res)

    def test_handle_success_push_event(self):
        """Test AQ_PUBLIC_FILES_LIST_OK is pushed on success."""
        response = [{'node_id': uuid.uuid4(), 'volume_id':None,
                    'public_url': 'http://example.com'}]
        self.command.handle_success(success=response)
        event = ('AQ_PUBLIC_FILES_LIST_OK', {'public_files': response,})
        self.assertIn(event, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_PUBLIC_FILES_LIST_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(urllib2.HTTPError(
                "http://example.com", 500, "Error", [], StringIO(msg)))
        self.command.handle_failure(failure=failure)
        event = ('AQ_PUBLIC_FILES_LIST_ERROR', {'error': msg})
        self.assertIn(event, self.command.action_queue.event_queue.events)


class DownloadUnconnectedTestCase(FactoryBaseTestCase):
    """Test for Download ActionQueueCommand, no connection"""

    def setUp(self):
        """Init."""
        super(DownloadUnconnectedTestCase, self).setUp()

        self.rq = request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = Download(request_queue, share_id='a_share_id',
                               node_id='a_node_id', server_hash='server_hash',
                               path='path', fileobj_factory=lambda: None)
        self.command.pre_queue_setup() # create the logger

    def test_progress_information_setup(self):
        """Test the setting up of the progress information in ._run()."""
        calls = []
        d = defer.Deferred()
        class FakedRequest():
            """Fake Request."""
            pass
        class FakedClient(object):
            """Fake Client."""
            def get_content_request(innerself, *args, **kwds):
                """Fake a get content request with its deferred."""
                calls.append(kwds.get('offset'))
                req = FakedRequest()
                req.deferred = d
                return req

        self.command.action_queue.connect_in_progress = False
        self.command.action_queue.client = FakedClient()
        self.command._run()

        self.assertEqual(self.command.n_bytes_read_last, 0)
        self.assertEqual(calls, [0])

        calls = []
        dloading = self.command.action_queue.downloading['a_share_id',
                                                         'a_node_id']
        dloading['n_bytes_read'] = 20
        self.command._run()

        self.assertEqual(self.command.n_bytes_read_last, 20)
        self.assertEqual(calls, [20])


class DownloadTestCase(ConnectedBaseTestCase):
    """Test for Download ActionQueueCommand."""

    def setUp(self):
        """Init."""
        res = super(DownloadTestCase, self).setUp()

        self.rq = RequestQueue(action_queue=self.action_queue)
        self.rq.transfers_semaphore = FakeSemaphore()

        class MyDownload(Download):
            """Just to allow monkeypatching."""
            sync = lambda s: None

        self.command = MyDownload(self.rq, share_id='a_share_id',
                                  node_id='a_node_id',
                                  server_hash='server_hash', path="/foo/bar",
                                  fileobj_factory=lambda: None)
        self.command.pre_queue_setup() # create the logger
        self.rq.waiting.append(self.command)

        return res

    def test_AQ_DOWNLOAD_FILE_PROGRESS_is_valid_event(self):
        """AQ_DOWNLOAD_FILE_PROGRESS is a valid event."""
        event = 'AQ_DOWNLOAD_FILE_PROGRESS'
        self.assertTrue(event in EVENTS)
        self.assertEqual(('share_id', 'node_id', 'n_bytes_read',
                          'deflated_size'), EVENTS[event])

    def test_progress_hook(self):
        # would first get the node attribute including this
        self.command.action_queue.downloading['a_share_id', 'a_node_id'] = {}
        self.command.nacb(deflated_size = 2*TRANSFER_PROGRESS_THRESHOLD)
        self.command.progress_start(0)

        self.command.progress_hook(5)
        self.assertEqual([], self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_read_last, 5)

        self.command.progress_hook(TRANSFER_PROGRESS_THRESHOLD)
        self.assertEqual([], self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_read_last,
                         TRANSFER_PROGRESS_THRESHOLD)

        self.command.progress_start(5)
        self.command.progress_hook(TRANSFER_PROGRESS_THRESHOLD+5)
        kwargs = {'share_id': 'a_share_id', 'node_id': 'a_node_id',
                  'deflated_size': 2*TRANSFER_PROGRESS_THRESHOLD,
                  'n_bytes_read': 5+TRANSFER_PROGRESS_THRESHOLD}
        events = [('AQ_DOWNLOAD_FILE_PROGRESS', kwargs)]
        self.assertEqual(events, self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_read_last,
                         5+TRANSFER_PROGRESS_THRESHOLD)

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        res = [getattr(self.command, x) for x in self.command.possible_markers]
        self.assertEqual(res, ['a_node_id'])

    def test_cancel_set_cancelled(self):
        """Set the command to cancelled."""
        assert not self.command.cancelled, "test badly set up"
        self.command.cancel()
        self.assertTrue(self.command.cancelled)

    def test_cancel_download_req_is_none(self):
        """It's ok to have download_req in None when cancelling."""
        assert self.command.download_req is None, "test badly set up"
        self.command.cancel()

    def test_cancel_download_req_is_something(self):
        """download_req is also cancelled."""
        # set up the mocker
        mocker = Mocker()
        obj = mocker.mock()
        obj.cancel()

        # test
        with mocker:
            self.command.download_req = obj
            self.command.cancel()

    def test_cancel_remove(self):
        """Remove the command from the queue."""
        # set up the mocker
        mocker = Mocker()
        obj = mocker.mock()
        obj.unqueue(self.command)

        # test
        with mocker:
            self.command._queue = obj
            self.command.cancel()

    def test_cancel_clean_up(self):
        """Clean up."""
        called = []
        self.command.cleanup = lambda: called.append(True)
        self.command.cancel()
        self.assertTrue(called)

    def test_uniqueness(self):
        """Info used for uniqueness."""
        u = self.command.uniqueness
        self.assertEqual(u, ('MyDownload', 'a_share_id', 'a_node_id'))

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        self.command._acquire_pathlock()
        self.assertEqual(t, [("", "foo", "bar"), {'logger': self.command.log}])

    def test_upload_download_uniqueness(self):
        """There should be only one upload/download for a specific node."""
        # totally fake, we don't care: the messages are only validated on run
        self.action_queue.download('foo', 'bar', 0, 'path', 0)
        self.action_queue.upload('foo', 'bar', 0, 0, 0, 0, 'path', 0)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    def test_uniqueness_upload(self):
        """There should be only one upload/download for a specific node."""
        # totally fake, we don't care: the messages are only validated on run
        self.action_queue.upload('foo', 'bar', 0, 0, 0, 0, 'path', 0)
        self.action_queue.download('foo', 'bar', 0, 'path', 0)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    def test_uniqueness_download(self):
        """There should be only one upload/download for a specific node."""
        # totally fake, we don't care: the messages are only validated on run
        self.action_queue.download('foo', 'bar', 0, 'path0', 0)
        self.action_queue.download('foo', 'bar', 1, 'path1', 1)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    @defer.inlineCallbacks
    def test_uniqueness_even_with_markers(self):
        """Only one upload/download per node, even using markers."""
        m = MDMarker('bar')
        yield self.action_queue.download('share', m, 0, 'path', 0)
        self.action_queue.uuid_map.set('bar', 'bah')
        self.action_queue.download('share', 'bah', 0, 'path', 0)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    def test_start_locks_on_semaphore(self):
        """_start acquire the semaphore and locks."""
        lock = defer.Deferred()
        self.rq.transfers_semaphore.acquire = lambda: lock
        self.action_queue.zip_queue.zip = lambda u: defer.succeed(True)

        # _start and check it locked
        started = self.command._start()
        self.assertFalse(started.called)

        # release the lock and check it finished
        lock.callback(True)
        self.assertTrue(started.called)
        self.assertTrue(self.handler.check_debug('semaphore acquired'))

    def test_handle_success_releases_semaphore(self):
        """Test semaphore is released on success."""
        called = []
        self.rq.transfers_semaphore.release = lambda: called.append(True)

        # success and check
        self.command.handle_success('foo')
        self.assertTrue(called)
        self.assertTrue(self.handler.check_debug('semaphore released'))

    def test_handle_failure_releases_semaphore(self):
        """Test semaphore is released on failure."""
        called = []
        self.rq.transfers_semaphore.release = lambda: called.append(True)

        # error and check
        self.command.handle_failure(Failure(DefaultException('oops')))
        self.assertTrue(called)
        self.assertTrue(self.handler.check_debug('semaphore released'))


class UploadUnconnectedTestCase(FactoryBaseTestCase):
    """Test for Upload ActionQueueCommand, no connection"""

    def setUp(self):
        """Init."""
        super(UploadUnconnectedTestCase, self).setUp()

        self.rq = request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = Upload(request_queue, share_id='a_share_id',
                              node_id='a_node_id', previous_hash='prev_hash',
                              hash='yadda', crc32=0, size=0, path='path',
                              fileobj_factory=lambda: None,
                              tempfile_factory=lambda: None)
        self.command.pre_queue_setup() # create the logger

    def test_upload_progress_wrapper_setup(self):
        """Test the setting up of the progress wrapper in ._run()."""
        calls = []
        d = defer.Deferred()
        class FakedRequest():
            """Fake Request."""
            pass
        class FakedClient(object):
            """Fake Client."""
            def put_content_request(innerself, *args):
                """Fake a put content request with its deferred."""
                calls.append(args)
                req = FakedRequest()
                req.deferred = d
                return req

        self.command.action_queue.connect_in_progress = False
        self.command.action_queue.client = FakedClient()
        self.command._run()

        self.assertEqual(len(calls), 1)
        upload_wrapper = calls[0][-1]
        uploading = self.action_queue.uploading['a_share_id', 'a_node_id']
        self.assertTrue(upload_wrapper.data_dict is uploading)
        self.assertEqual(upload_wrapper.progress_hook,
                         self.command.progress_hook)
        self.assertEqual(self.command.n_bytes_written_last, 0)

        self.command.n_bytes_written_last = 20
        self.command._run()
        self.assertEqual(self.command.n_bytes_written_last, 0)


class UploadProgressWrapperTestCase(BaseTwistedTestCase):
    """Test for the UploadProgressWrapper helper class."""

    def test_read(self):
        """Test the read method."""
        progress = []

        def progress_hook(n_bytes_written):
            progress.append(n_bytes_written)

        info = {'n_bytes_written': 0}
        f = StringIO("x"*10+"y"*5)
        upw = UploadProgressWrapper(f, info, progress_hook)

        res = upw.read(10)
        self.assertEqual(res, "x"*10)
        self.assertEqual(progress, [0])

        res = upw.read(5)
        self.assertEqual(res, "y"*5)
        self.assertEqual(progress, [0, 10])


class UploadTestCase(ConnectedBaseTestCase):
    """Test for Upload ActionQueueCommand."""

    def setUp(self):
        """Init."""
        super(UploadTestCase, self).setUp()

        self.rq = RequestQueue(action_queue=self.action_queue)
        self.rq.transfers_semaphore = FakeSemaphore()
        self.rq.unqueue = lambda c: None
        self.rq.active = True

        class MyUpload(Upload):
            """Just to allow monkeypatching."""

        self.share_id = str(uuid.uuid4())
        self.command = MyUpload(self.rq, share_id=self.share_id,
                                node_id='a_node_id', previous_hash='prev_hash',
                                hash='yadda', crc32=0, size=0, path='/foo/bar',
                                fileobj_factory=lambda: None,
                                tempfile_factory=lambda: None)
        self.command.pre_queue_setup() # create the logger

    def test_upload_in_progress(self):
        """Test Upload retries on UploadInProgress."""
        d = defer.Deferred()
        self.command.run = lambda: d.callback(True)
        self.command.running = True

        # send the failure
        protocol_msg = protocol_pb2.Message()
        protocol_msg.type = protocol_pb2.Message.ERROR
        protocol_msg.error.type = protocol_pb2.Error.UPLOAD_IN_PROGRESS
        err = errors.UploadInProgressError("request", protocol_msg)
        self.command.end_errback(failure=Failure(err))
        return d

    def test_handle_success_push_event(self):
        """Test AQ_UPLOAD_FINISHED is pushed on success."""
        # create a request and fill it with succesful information
        request = client.PutContent(self.action_queue.client, VOLUME, 'node',
                                    'prvhash', 'newhash', 'crc32', 'size',
                                    'deflated', 'fd')
        request.new_generation = 13

        # trigger success in the command
        self.command.handle_success(request)

        # check for successful event
        kwargs = dict(share_id=self.command.share_id, node_id='a_node_id',
                      hash='yadda', new_generation=13)
        events = [('AQ_UPLOAD_FINISHED', kwargs)]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_UPLOAD_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        kwargs = dict(share_id=self.command.share_id, node_id='a_node_id',
                      hash='yadda', error=msg)
        events = [('AQ_UPLOAD_ERROR', kwargs)]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_handle_failure_removes_temp_file(self):
        """Test temp file is removed on failure."""
        class TempFile(object): pass
        self.command.tempfile = TempFile()
        self.command.tempfile.name = os.path.join(self.tmpdir, 'remove-me.zip')
        open(self.command.tempfile.name, 'w').close()
        assert os.path.exists(self.command.tempfile.name)

        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        self.assertFalse(os.path.exists(self.command.tempfile.name))

    def test_handle_failure_push_quota_exceeded_if_quota_exceeded_error(self):
        """Test SYS_QUOTA_EXCEEDED is pushed on QuotaExceededError."""
        protocol_msg = protocol_pb2.Message()
        protocol_msg.type = protocol_pb2.Message.ERROR
        protocol_msg.error.type = protocol_pb2.Error.QUOTA_EXCEEDED
        protocol_msg.free_space_info.share_id = self.command.share_id
        protocol_msg.free_space_info.free_bytes = 1331564676
        error = errors.QuotaExceededError("request", protocol_msg)
        failure = Failure(error)

        self.command.handle_failure(failure=failure)
        volume_id = self.command.share_id
        kwargs = dict(share_id=volume_id, node_id='a_node_id',
                      hash='yadda', error=str(error))
        event1 = ('SYS_QUOTA_EXCEEDED',
                  {'volume_id': volume_id, 'free_bytes': 1331564676})
        event2 = ('AQ_UPLOAD_ERROR', kwargs)
        self.assertTrue(event1 in self.command.action_queue.event_queue.events)
        self.assertTrue(event2 in self.command.action_queue.event_queue.events)

    @defer.inlineCallbacks
    def test_compress_failed_pushes_upload_error(self):
        """Test that upload error is pushed when compress fails."""
        msg = 'Zip can not be accomplished.'
        error = DefaultException(msg)
        self.action_queue.zip_queue.zip = lambda upload: defer.fail(error)
        yield self.command.run()
        kwargs = dict(share_id=self.command.share_id, node_id='a_node_id',
                      hash='yadda', error=msg)
        event = ('AQ_UPLOAD_ERROR', kwargs)
        self.assertIn(event, self.command.action_queue.event_queue.events)

    def test_AQ_UPLOAD_FILE_PROGRESS_is_valid_event(self):
        """AQ_UPLOAD_FILE_PROGRESS is a valid event."""
        event = 'AQ_UPLOAD_FILE_PROGRESS'
        self.assertTrue(event in EVENTS)
        self.assertEqual(('share_id', 'node_id', 'n_bytes_written',
                          'deflated_size'), EVENTS[event])

    def test_progress_hook(self):
        """Test the progress hook."""
        self.command.deflated_size = 2*TRANSFER_PROGRESS_THRESHOLD
        self.command.n_bytes_written_last = 0

        self.command.progress_hook(5)
        self.assertEqual([], self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_written_last, 5)

        self.command.progress_hook(TRANSFER_PROGRESS_THRESHOLD)
        self.assertEqual([], self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_written_last,
                         TRANSFER_PROGRESS_THRESHOLD)

        self.command.n_bytes_written_last = 5
        self.command.progress_hook(TRANSFER_PROGRESS_THRESHOLD+14)

        kwargs = {'share_id': self.command.share_id, 'node_id': 'a_node_id',
                  'deflated_size': 2*TRANSFER_PROGRESS_THRESHOLD,
                  'n_bytes_written': 14+TRANSFER_PROGRESS_THRESHOLD }
        events = [('AQ_UPLOAD_FILE_PROGRESS', kwargs)]
        self.assertEqual(events, self.command.action_queue.event_queue.events)
        self.assertEqual(self.command.n_bytes_written_last,
                         14+TRANSFER_PROGRESS_THRESHOLD)

    def test_runnable_space_ok(self):
        """The upload is runnable if space ok."""
        self.action_queue.have_sufficient_space_for_upload = lambda *a: True
        self.assertTrue(self.command.is_runnable)

    def test_runnable_space_bad(self):
        """The upload is not runnable without free space."""
        self.action_queue.have_sufficient_space_for_upload = lambda *a: False
        self.assertFalse(self.command.is_runnable)

    def test_runnable_space_bad_cancelled(self):
        """The upload is runnable if cancelled even with no free space."""
        self.action_queue.have_sufficient_space_for_upload = lambda *a: False
        self.command.cancelled = True
        self.assertTrue(self.command.is_runnable)

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        res = [getattr(self.command, x) for x in self.command.possible_markers]
        self.assertEqual(res, ['a_node_id'])

    def test_cancel_set_cancelled(self):
        """Set the command to cancelled."""
        assert not self.command.cancelled, "test badly set up"
        self.command.cancel()
        self.assertTrue(self.command.cancelled)

    def test_cancel_upload_req_is_none(self):
        """It's ok to have upload_req in None when cancelling."""
        assert self.command.upload_req is None, "test badly set up"
        self.command.cancel()

    def test_cancel_upload_req_is_something(self):
        """upload_req is also cancelled."""
        # set up the mocker
        mocker = Mocker()
        obj = mocker.mock()
        obj.cancel()
        obj.producer

        # test
        with mocker:
            self.command.upload_req = obj
            self.command.cancel()

    def test_cancel_remove(self):
        """Remove the command from the queue."""
        # set up the mocker
        mocker = Mocker()
        obj = mocker.mock()
        obj.unqueue(self.command)

        # test
        with mocker:
            self.command._queue = obj
            self.command.cancel()

    def test_cancel_clean_up(self):
        """Clean up."""
        called = []
        self.command.cleanup = lambda: called.append(True)
        self.command.cancel()
        self.assertTrue(called)

    def test_uniqueness(self):
        """Info used for uniqueness."""
        u = self.command.uniqueness
        self.assertEqual(u, ('MyUpload', self.share_id, 'a_node_id'))

    @defer.inlineCallbacks
    def test_loses_client_while_in_start(self):
        """Handle losing the client while in _start."""
        self.action_queue.client = None
        try:
            yield self.command._run()
        except ClientNoLongerThere, e:
            self.assertEqual(str(e), 'Detected in _run')
        else:
            self.fail("Test should have raised ClientNoLongerThere")

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        self.command._acquire_pathlock()
        self.assertEqual(t, [("", "foo", "bar"), {'logger': self.command.log}])

    def test_uniqueness_upload(self):
        """There should be only one upload/download for a specific node."""
        # totally fake, we don't care: the messages are only validated on run
        self.action_queue.upload('foo', 'bar', 0, 0, 0, 0, 'path0', 0)
        self.action_queue.upload('foo', 'bar', 1, 1, 1, 1, 'path1', 1)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    def test_uniqueness_download(self):
        """There should be only one upload/download for a specific node."""
        # totally fake, we don't care: the messages are only validated on run
        self.action_queue.download('foo', 'bar', 0, 'path', 0)
        self.action_queue.upload('foo', 'bar', 0, 0, 0, 0, 'path', 0)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    @defer.inlineCallbacks
    def test_uniqueness_even_with_markers(self):
        """Only one upload/download per node, even using markers."""
        m = MDMarker('bar')
        yield self.action_queue.download('share', m, 0, 'path', 0)
        self.action_queue.uuid_map.set('bar', 'bah')
        self.action_queue.upload('share', 'bah', 0, 0, 0, 0, 'path', 0)
        self.assertEqual(len(self.action_queue.queue.waiting), 1)

    def test_start_locks_on_semaphore(self):
        """_start acquire the semaphore and locks."""
        lock = defer.Deferred()
        self.rq.transfers_semaphore.acquire = lambda: lock
        self.action_queue.zip_queue.zip = lambda u: defer.succeed(True)

        # _start and check it locked
        started = self.command._start()
        self.assertFalse(started.called)

        # release the lock and check it finished
        lock.callback(True)
        self.assertTrue(started.called)
        self.assertTrue(self.handler.check_debug('semaphore acquired'))

    def test_handle_success_releases_semaphore(self):
        """Test semaphore is released on success."""
        called = []
        self.rq.transfers_semaphore.release = lambda: called.append(True)

        # create a request and fill it with succesful information
        request = client.PutContent(self.action_queue.client, VOLUME, 'node',
                                    'prvhash', 'newhash', 'crc32', 'size',
                                    'deflated', 'fd')
        request.new_generation = 13

        # success and check
        self.command.handle_success(request)
        self.assertTrue(called)
        self.assertTrue(self.handler.check_debug('semaphore released'))

    def test_handle_failure_releases_semaphore(self):
        """Test semaphore is released on failure."""
        called = []
        self.rq.transfers_semaphore.release = lambda: called.append(True)

        # error and check
        self.command.handle_failure(Failure(DefaultException('oops')))
        self.assertTrue(called)
        self.assertTrue(self.handler.check_debug('semaphore released'))


class CreateShareTestCase(ConnectedBaseTestCase):
    """Test for CreateShare ActionQueueCommand."""

    @defer.inlineCallbacks
    def setUp(self):
        """Init."""
        yield super(CreateShareTestCase, self).setUp()
        self.request_queue = RequestQueue(action_queue=self.action_queue)
        self.orig_create_share_http = CreateShare._create_share_http

    @defer.inlineCallbacks
    def tearDown(self):
        yield super(CreateShareTestCase, self).tearDown()
        CreateShare._create_share_http = self.orig_create_share_http

    @defer.inlineCallbacks
    def test_access_level_modify_http(self):
        """Test proper handling of the access level in the http case."""
        # replace _create_share_http with a fake, just to check the args
        d = defer.Deferred()
        def check_create_http(self, node_id, user, name, read_only, deferred):
            """Fire the deferred with the args."""
            d.callback((node_id, user, name, read_only))
            deferred.callback(None)
        CreateShare._create_share_http = check_create_http
        command = CreateShare(self.request_queue, 'node_id',
                                   'share_to@example.com', 'share_name',
                                   'Modify', 'marker', 'path')
        self.assertTrue(command.use_http, 'CreateShare should be in http mode')

        command._run()
        node_id, user, name, read_only = yield d
        self.assertEqual('node_id', node_id)
        self.assertEqual('share_to@example.com', user)
        self.assertEqual('share_name', name)
        self.assertFalse(read_only)

    @defer.inlineCallbacks
    def test_access_level_view_http(self):
        """Test proper handling of the access level in the http case."""
        # replace _create_share_http with a fake, just to check the args
        d = defer.Deferred()
        def check_create_http(self, node_id, user, name, read_only, deferred):
            """Fire the deferred with the args."""
            d.callback((node_id, user, name, read_only))
            deferred.callback(None)
        CreateShare._create_share_http = check_create_http
        command = CreateShare(self.request_queue, 'node_id',
                                   'share_to@example.com', 'share_name',
                                   'View', 'marker', 'path')
        self.assertTrue(command.use_http, 'CreateShare should be in http mode')
        command._run()
        node_id, user, name, read_only = yield d
        self.assertEqual('node_id', node_id)
        self.assertEqual('share_to@example.com', user)
        self.assertEqual('share_name', name)
        self.assertTrue(read_only)

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        cmd = CreateShare(self.request_queue, 'node_id', 'shareto@example.com',
                          'share_name', 'View', 'marker', 'path')
        res = [getattr(cmd, x) for x in cmd.possible_markers]
        self.assertEqual(res, ['node_id'])

    def test_handle_success_sends_invitation(self):
        """The success handler sends the AQ_SHARE_INVITATION_SENT event."""
        marker_id = "marker"
        share_name = "share name"
        share_to = "share_to@example.com"

        class MockResult(object):
            """A mock result."""
            share_id = SHARE

        mock_success = MockResult()
        cmd = CreateShare(self.request_queue, NODE, share_to,
                          share_name, 'View', marker_id, 'path')
        cmd.log = cmd.make_logger()
        cmd.use_http = True
        cmd.handle_success(mock_success)

        event_params = { 'marker': marker_id }
        events = [('AQ_SHARE_INVITATION_SENT', event_params)]
        self.assertEqual(events, cmd.action_queue.event_queue.events)

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        cmd = CreateShare(self.request_queue, NODE, 'share_to',
                          'share_name', 'View', 'marker_id', 'foo/bar')
        cmd._acquire_pathlock()
        self.assertEqual(t, [('foo', 'bar'), {'logger': None}])


class DeleteShareTestCase(ConnectedBaseTestCase):
    """Test for DeleteShare ActionQueueCommand."""

    @defer.inlineCallbacks
    def setUp(self):
        """Init."""
        yield super(DeleteShareTestCase, self).setUp()
        request_queue = RequestQueue(action_queue=self.action_queue)
        self.command = DeleteShare(request_queue, SHARE)
        vm = self.action_queue.main.vm
        self.patch(vm, 'handle_AQ_DELETE_SHARE_OK', NO_OP)
        self.patch(vm, 'handle_AQ_DELETE_SHARE_ERROR', NO_OP)

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's delete_volume is called."""
        self.called = False
        def check(share_id):
            """Take control over client's feature."""
            self.called = True
            self.assertEqual(SHARE, share_id)
        self.patch(self.command.action_queue.client, 'delete_share', check)
        self.command._run()
        self.assertTrue(self.called, "command wasn't called")

    def test_handle_success_push_event(self):
        """Test AQ_DELETE_SHARE_OK is pushed on success."""
        request = client.DeleteShare(self.action_queue.client, SHARE)
        self.command.handle_success(success=request)
        events = [('AQ_DELETE_SHARE_OK', {'share_id': SHARE})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)

    def test_handle_failure_push_event(self):
        """Test AQ_DELETE_SHARE_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        self.command.handle_failure(failure=failure)
        events = [('AQ_DELETE_SHARE_ERROR',
                  {'share_id': SHARE, 'error': msg})]
        self.assertEqual(events, self.command.action_queue.event_queue.events)


class SimpleAQTestCase(BasicTestCase):
    """Simple tests for AQ API."""

    def test_aq_query_volumes(self):
        """Check the API of AQ.query_volumes."""
        self.main.start()
        d = defer.Deferred()
        def list_volumes():
            """Fake list_volumes."""
            result = DummyClass()
            result.volumes = ['foo', 'bar']
            return defer.succeed(result)

        self.action_queue.client = DummyClass()
        self.action_queue.client.list_volumes = list_volumes
        d = self.action_queue.query_volumes()
        def check(result):
            self.assertIn('foo', result)
            self.assertIn('bar', result)
            return result
        d.addCallback(check)
        return d

    def test_have_sufficient_space_for_upload_if_free_space_is_none(self):
        """Check have_sufficient_space_for_upload.

        If free_space is None, SYS_QUOTA_EXCEEDED is not pushed.

        """
        self.patch(self.action_queue.main.vm, 'get_free_space',
                   lambda share_id: None)  # free space is None
        volume_id = 'test share'
        res = self.action_queue.have_sufficient_space_for_upload(volume_id,
                                                                 upload_size=1)
        self.assertTrue(res, "Must have enough space to upload.")
        events = map(operator.itemgetter(0),
                     self.action_queue.event_queue.events)
        self.assertNotIn('SYS_QUOTA_EXCEEDED', events)

    def test_have_sufficient_space_for_upload_if_no_free_space(self):
        """Check have_sufficient_space_for_upload pushes SYS_QUOTA_EXCEEDED."""
        self.patch(self.action_queue.main.vm, 'get_free_space',
                   lambda share_id: 0) # no free space, always
        volume_id = 'test share'
        res = self.action_queue.have_sufficient_space_for_upload(volume_id,
                                                                 upload_size=1)
        self.assertEqual(res, False, "Must not have enough space to upload.")
        msg = 'SYS_QUOTA_EXCEEDED must have been pushed to event queue.'
        expected = ('SYS_QUOTA_EXCEEDED',
                    {'volume_id': volume_id, 'free_bytes': 0})
        self.assertTrue(expected in self.action_queue.event_queue.events, msg)

    def test_have_sufficient_space_for_upload_if_free_space(self):
        """Check have_sufficient_space_for_upload doesn't push any event."""
        self.patch(self.action_queue.main.vm, 'get_free_space',
                   lambda share_id: 1) # free space, always
        res = self.action_queue.have_sufficient_space_for_upload(share_id=None,
                                                                 upload_size=0)
        self.assertEqual(res, True, "Must have enough space to upload.")
        msg = 'No event must have been pushed to event queue.'
        self.assertEqual(self.action_queue.event_queue.events, [], msg)

    def test_SYS_QUOTA_EXCEEDED_is_valid_event(self):
        """SYS_QUOTA_EXCEEDED is a valid event."""
        event = 'SYS_QUOTA_EXCEEDED'
        self.assertTrue(event in EVENTS)
        self.assertEqual(('volume_id', 'free_bytes'), EVENTS[event])

    def test_SYS_USER_CONNECT_is_valid_event(self):
        """SYS_USER_CONNECT is a valid event."""
        event = 'SYS_USER_CONNECT'
        self.assertIn(event, EVENTS)
        self.assertEqual(('access_token',), EVENTS[event])

    def test_handle_SYS_USER_CONNECT(self):
        """handle_SYS_USER_CONNECT stores credentials."""
        self.assertEqual(self.action_queue.token, None)
        self.assertEqual(self.action_queue.consumer, None)

        self.user_connect()

        expected = oauth.OAuthToken('bla', 'ble')
        self.assertEqual(self.action_queue.token.key, expected.key)
        self.assertEqual(self.action_queue.token.secret, expected.secret)

        expected = oauth.OAuthConsumer('foo', 'bar')
        self.assertEqual(self.action_queue.consumer.key, expected.key)
        self.assertEqual(self.action_queue.consumer.secret, expected.secret)


class SpecificException(Exception):
    """The specific exception."""


class SillyClass(object):
    """Silly class that accepts the set of any attribute.

    We can't use object() directly, since its raises AttributeError.

    """


class ErrorHandlingTestCase(BasicTestCase):
    """Error handling tests for ActionQueue."""

    def setUp(self):
        """Init."""
        BasicTestCase.setUp(self)

        self.called = False
        self.action_queue.client = SillyClass()
        self.patch(self.main, 'restart', lambda: None)

        self.main.start()

    def fail_please(self, an_exception):
        """Raise the given exception."""
        def inner(*args, **kwargs):
            """A request to the server that fails."""
            self.called = True
            return defer.fail(an_exception)
        return inner

    def succeed_please(self, result):
        """Return the given result."""
        def inner(*args, **kwargs):
            """A request to the server that succeeds."""
            self.called = True
            return defer.succeed(result)
        return inner

    def mock_caps(self, accepted):
        """Reply to query caps with False."""
        def gset_caps(caps):
            """get/set caps helper."""
            req = SillyClass()
            req.caps = caps
            req.accepted = accepted
            return defer.succeed(req)
        return gset_caps

    def test_valid_event(self):
        """SYS_SERVER_ERROR is valid in EventQueue."""
        event = 'SYS_SERVER_ERROR'
        self.assertTrue(event in EVENTS)
        self.assertEqual(('error',), EVENTS[event])

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_no_error(self):
        """_send_request_and_handle_errors is correct when no error."""

        event = 'SYS_SPECIFIC_OK'
        EVENTS[event] = () # add event to the global valid events list
        self.addCleanup(lambda: EVENTS.pop(event))

        result = object()
        request = self.succeed_please(result)
        kwargs = dict(request=request, request_error=SpecificException,
                      event_error='YADDA_YADDA', event_ok=event,
                      args=(1, 2), kwargs={})
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        actual_result = yield d

        self.assertTrue(self.called, 'the request was called')
        self.assertEqual(actual_result, result)
        self.assertEqual((event, {}),
                         self.action_queue.event_queue.events[-1])

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__, 'OK'))

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_with_no_event_ok(self):
        """_send_request_and_handle_errors does not push event if is None."""
        original_events = self.action_queue.event_queue.events[:]

        result = object()
        request = self.succeed_please(result)
        kwargs = dict(request=request, request_error=SpecificException,
                      event_error='YADDA_YADDA', event_ok=None)
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        actual_result = yield d

        self.assertTrue(self.called, 'the request was called')
        self.assertEqual(actual_result, result)
        self.assertEqual(original_events,
                         self.action_queue.event_queue.events)

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__, 'OK'))

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_valid_error(self):
        """_send_request_and_handle_errors is correct when expected error."""

        event = 'SYS_SPECIFIC_ERROR'
        EVENTS[event] = ('error',) # add event to the global valid events list
        self.addCleanup(lambda: EVENTS.pop(event))

        exc = SpecificException('The request failed! please be happy.')
        request = self.fail_please(exc)
        kwargs = dict(request=request, request_error=SpecificException,
                      event_error=event, event_ok='YADDA_YADDA')
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        yield d

        self.assertTrue(self.called, 'the request was called')
        self.assertEqual((event, {'error': str(exc)}),
                         self.action_queue.event_queue.events[-1])

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__,
                                                event, str(exc)))

    @defer.inlineCallbacks
    def assert_send_request_and_handle_errors_on_server_error(self, serr):
        """_send_request_and_handle_errors is correct when server error."""
        # XXX: we need to replace this list with and exception list
        # once bug #557718 is resolved
        msg = protocol_pb2.Message()
        msg.type = protocol_pb2.Message.ERROR
        msg.error.type = serr
        msg.error.comment = 'Error message for %s.' % serr
        exc = errors.error_to_exception(serr)(request=None, message=msg)

        request = self.fail_please(exc)
        kwargs = dict(request=request, request_error=SpecificException,
                      event_error='BAR', event_ok='FOO')
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        yield d

        event = 'SYS_SERVER_ERROR'
        self.assertEqual((event, {'error': str(exc)}),
                         self.action_queue.event_queue.events[-1])

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__,
                                                event, str(exc)))

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_try_again(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.TRY_AGAIN
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_internal_error(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.INTERNAL_ERROR
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_protocol_error(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.PROTOCOL_ERROR
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_unsupported_version(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.UNSUPPORTED_VERSION
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_authetication_failed(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.AUTHENTICATION_FAILED
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_no_permission(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.NO_PERMISSION
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_already_exists(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.ALREADY_EXISTS
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_does_not_exist(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.DOES_NOT_EXIST
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_not_a_dir(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.NOT_A_DIRECTORY
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_not_empty(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.NOT_EMPTY
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_not_available(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.NOT_AVAILABLE
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_upload_in_progress(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.UPLOAD_IN_PROGRESS
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_upload_corrupt(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.UPLOAD_CORRUPT
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_upload_canceled(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.UPLOAD_CANCELED
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_conflict(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.CONFLICT
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_quota_exceeded(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.QUOTA_EXCEEDED
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_invalid_filename(self):
        """_send_request_and_handle_errors is correct when server error."""
        serr = protocol_pb2.Error.INVALID_FILENAME
        yield self.assert_send_request_and_handle_errors_on_server_error(serr)

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_unknown_error(self):
        """_send_request_and_handle_errors is correct when unknown error."""
        # XXX: we need to replace this list with and exception list
        # once bug #557718 is resolved
        serr = protocol_pb2.Error.AUTHENTICATION_REQUIRED
        msg = protocol_pb2.Message()
        msg.type = protocol_pb2.Message.ERROR
        msg.error.type = serr
        msg.error.comment = 'Error message for %s.' % serr
        exc = errors.error_to_exception(serr)(request=None, message=msg)

        request = self.fail_please(exc)
        kwargs = dict(request=request, request_error=SpecificException,
                  event_error='BAR', event_ok='FOO')
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        yield d

        event = 'SYS_UNKNOWN_ERROR'
        self.assertIn((event, {}), self.action_queue.event_queue.events)

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__,
                                                event, str(exc)))

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_no_protocol_error(self):
        """_send_request_and_handle_errors is ok when no-protocol error."""

        event = 'SYS_UNKNOWN_ERROR'
        error_msg = 'Error message for any Exception.'
        exc = Exception(error_msg)
        request = self.fail_please(exc)
        kwargs = dict(request=request, request_error=SpecificException,
                      event_error='BAR', event_ok='FOO')
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        yield d

        self.assertIn((event, {}), self.action_queue.event_queue.events)

        # assert over logging
        self.assertTrue(self.handler.check_info(request.__name__,
                                                event, str(exc)))

    @defer.inlineCallbacks
    def test_send_request_and_handle_errors_on_client_mismatch(self):
        """_send_request_and_handle_errors is correct when client mismatch."""

        def change_client(*args, **kwargs):
            """Change AQ's client while doing the request."""
            self.action_queue.client = object()

        self.action_queue.event_queue.events = [] # event cleanup
        kwargs = dict(request=change_client, request_error=SpecificException,
                      event_error='BAR', event_ok='FOO')
        d = self.action_queue._send_request_and_handle_errors(**kwargs)
        yield d

        self.assertEqual([], self.action_queue.event_queue.events)

        # assert over logging
        self.assertTrue(self.handler.check_warning(change_client.__name__,
                                                   'Client mismatch'))

    @defer.inlineCallbacks
    def test_check_version_when_unsupported_version_exception(self):
        """Test error handling after UnsupportedVersionError."""
        # raise a UnsupportedVersionError
        msg = protocol_pb2.Message()
        msg.type = protocol_pb2.Message.ERROR
        msg.error.type = protocol_pb2.Error.UNSUPPORTED_VERSION
        msg.error.comment = 'This is a funny comment.'
        exc = errors.UnsupportedVersionError(request=None, message=msg)

        self.action_queue.client.protocol_version = self.fail_please(exc)
        yield self.action_queue.check_version()
        event = ('SYS_PROTOCOL_VERSION_ERROR', {'error': str(exc)})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])

    @defer.inlineCallbacks
    def test_set_capabilities_when_query_caps_not_accepted(self):
        """Test error handling when the query caps are not accepeted."""

        # query_caps returns False
        self.action_queue.client.query_caps = self.mock_caps(accepted=False)

        yield self.action_queue.set_capabilities(caps=None)
        msg = "The server doesn't have the requested capabilities"
        event = ('SYS_SET_CAPABILITIES_ERROR', {'error': msg})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])
        self.assertNotIn(('SYS_SET_CAPABILITIES_OK', {}),
                          self.action_queue.event_queue.events)

    @defer.inlineCallbacks
    def test_set_capabilities_when_set_caps_not_accepted(self):
        """Test error handling when the query caps are not accepted."""

        # query_caps returns True and set_caps returns False
        self.action_queue.client.query_caps = self.mock_caps(accepted=True)
        self.action_queue.client.set_caps = self.mock_caps(accepted=False)

        caps = 'very difficult cap'
        yield self.action_queue.set_capabilities(caps=caps)
        msg = "The server denied setting '%s' capabilities" % caps
        event = ('SYS_SET_CAPABILITIES_ERROR', {'error': msg})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])
        self.assertNotIn(('SYS_SET_CAPABILITIES_OK', {}),
                          self.action_queue.event_queue.events)

    @defer.inlineCallbacks
    def test_set_capabilities_when_client_is_none(self):
        """Test error handling when the client is None."""

        self.action_queue.client = None

        yield self.action_queue.set_capabilities(caps=None)
        msg = "'NoneType' object has no attribute 'query_caps'"
        event = ('SYS_SET_CAPABILITIES_ERROR', {'error': msg})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])
        self.assertNotIn(('SYS_SET_CAPABILITIES_OK', {}),
                          self.action_queue.event_queue.events)

    @defer.inlineCallbacks
    def test_set_capabilities_when_set_caps_is_accepted(self):
        """Test error handling when the query caps are not accepeted."""

        # query_caps returns True and set_caps returns True
        self.action_queue.client.query_caps = self.mock_caps(accepted=True)
        self.action_queue.client.set_caps = self.mock_caps(accepted=True)

        yield self.action_queue.set_capabilities(caps=None)
        event = ('SYS_SET_CAPABILITIES_OK', {})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])

    @defer.inlineCallbacks
    def test_authenticate_when_authenticated(self):
        """Test error handling after authenticate with no error."""
        request = client.Authenticate(self.action_queue.client,
                                      {'dummy_token': 'credentials'})
        request.session_id = str(uuid.uuid4())
        self.action_queue.client.oauth_authenticate = \
            self.succeed_please(result=request)
        yield self.action_queue.authenticate()
        event = ('SYS_AUTH_OK', {})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])

    @defer.inlineCallbacks
    def test_authenticate_when_authentication_failed_exception(self):
        """Test error handling after AuthenticationFailedError."""
        # raise a AuthenticationFailedError
        msg = protocol_pb2.Message()
        msg.type = protocol_pb2.Message.ERROR
        msg.error.type = protocol_pb2.Error.AUTHENTICATION_FAILED
        msg.error.comment = 'This is a funny comment.'
        exc = errors.AuthenticationFailedError(request=None, message=msg)

        self.action_queue.client.oauth_authenticate = self.fail_please(exc)
        yield self.action_queue.authenticate()
        event = ('SYS_AUTH_ERROR', {'error': str(exc)})
        self.assertEqual(event, self.action_queue.event_queue.events[-1])


class GetDeltaTestCase(ConnectedBaseTestCase):
    """Test for GetDelta ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(GetDeltaTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_action_queue_get_delta(self):
        """Test AQ get delta."""
        res = self.action_queue.get_delta(VOLUME, 0)
        self.assertTrue(res is None) # this is what start returns

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        cmd = GetDelta(self.rq, VOLUME, 0)
        self.assertTrue(isinstance(cmd, ActionQueueCommand))

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        cmd = GetDelta(self.rq, VOLUME, 0)
        res = cmd._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's get delta is called."""
        called = []
        self.patch(self.action_queue.client, 'get_delta',
                   lambda *a: called.append(a))

        cmd = GetDelta(self.rq, VOLUME, 35)
        cmd._run()
        self.assertEqual(called[0], (VOLUME, 35))

    def test_handle_success_push_event(self):
        """Test AQ_DELTA_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.GetDelta(self.action_queue.client,
                                  share_id=VOLUME, from_generation=21)
        request.response = ['foo', 'bar']
        request.end_generation = 76
        request.full = True
        request.free_bytes = 1231234

        # create a command and trigger it success
        cmd = GetDelta(self.rq, VOLUME, 21)
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        delta_info = dict(volume_id=VOLUME, delta_content=['foo', 'bar'],
                          end_generation=76,
                          full=True, free_bytes=1231234)
        self.assertEqual(received, ('AQ_DELTA_OK', delta_info))
        self.assertTrue(isinstance(received[1]["delta_content"], DeltaList))

    def test_handle_generic_failure_push_event(self):
        """Test AQ_DELTA_ERROR is pushed on failure."""
        # create a failure
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))

        # create a command and trigger it success
        cmd = GetDelta(self.rq, VOLUME, 77)
        cmd.handle_failure(failure=failure)

        # check for event
        received = self.action_queue.event_queue.events[0]
        self.assertEqual(received, ('AQ_DELTA_ERROR',
                                    {'volume_id': VOLUME, 'error': msg}))

    def test_handle_notpossible_failure_push_event(self):
        """Test AQ_DELTA_NOT_POSSIBLE is pushed on that failure."""
        # create a failure
        msg = protocol_pb2.Message()
        msg.type = protocol_pb2.Message.ERROR
        msg.error.type = protocol_pb2.Error.CANNOT_PRODUCE_DELTA
        msg.error.comment = 'Something went wrong'
        failure = Failure(errors.CannotProduceDelta(self.rq, msg))

        # create a command and trigger it success
        cmd = GetDelta(self.rq, VOLUME, 2)
        cmd.handle_failure(failure=failure)

        # check for event
        received = self.action_queue.event_queue.events[0]
        self.assertEqual(received, ('AQ_DELTA_NOT_POSSIBLE',
                                    {'volume_id': VOLUME}))

    def test_queued_mixed_types(self):
        """Command gets queued if other command is waiting."""
        cmd1 = FakeCommand()
        self.rq.waiting.append(cmd1)
        cmd2 = GetDelta(self.rq, 'vol2', 0)
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])
        self.assertTrue(self.handler.check_debug("queueing"))

    def test_queued_two_different(self):
        """Two different queued commands is ok."""
        cmd1 = GetDelta(self.rq, 'vol1', 0)
        cmd1.queue()
        cmd2 = GetDelta(self.rq, 'vol2', 0)
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])

    def test_queued_two_equal_second_bigger(self):
        """When two equals, only survive the one with smaller gen, first."""
        cmd1 = GetDelta(self.rq, 'vol', 3)
        cmd1.queue()
        cmd2 = GetDelta(self.rq, 'vol', 5)
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1])
        self.assertTrue(self.handler.check_debug("not queueing self"))

    def test_queued_two_equal_second_smaller_first_not_running(self):
        """Two equals, survive the one with smaller gen, second, not run."""
        cmd1 = GetDelta(self.rq, 'vol', 5)
        cmd1.queue()
        cmd1.running = False
        cmd2 = GetDelta(self.rq, 'vol', 3)
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd2])
        self.assertTrue(self.handler.check_debug("removing previous command"))

    def test_queued_two_equal_second_smaller_first_running(self):
        """Two equals, survive the one with smaller gen, second, running."""
        cmd1 = GetDelta(self.rq, 'vol', 5)
        cmd1.queue()
        cmd2 = GetDelta(self.rq, 'vol', 3)
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])

    def test_queued_three_equal(self):
        """When several equals, only survive the one with smaller gen."""
        cmd1 = GetDelta(self.rq, 'vol', 5)
        cmd1.queue()
        cmd1.running = False
        cmd2 = GetDelta(self.rq, 'vol', 3)
        cmd2.queue()
        cmd3 = GetDelta(self.rq, 'vol', 7)
        cmd3.queue()
        self.assertEqual(list(self.rq.waiting), [cmd2])

    def test_uniqueness(self):
        """Info used for uniqueness."""
        cmd = GetDelta(self.rq, 'vol', 1)
        self.assertEqual(cmd.uniqueness, ('GetDelta', 'vol'))


class GetDeltaFromScratchTestCase(ConnectedBaseTestCase):
    """Test for GetDelta ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(GetDeltaFromScratchTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_action_queue_get_delta(self):
        """Test AQ get delta."""
        res = self.action_queue.rescan_from_scratch(VOLUME)
        self.assertTrue(res is None) # this is what start returns

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        cmd = GetDeltaFromScratch(self.rq, VOLUME)
        self.assertTrue(isinstance(cmd, ActionQueueCommand))

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        cmd = GetDelta(self.rq, VOLUME, 0)
        res = cmd._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')
        res.addErrback(self.silent_connection_lost)

    def test_run_calls_protocol(self):
        """Test protocol's get delta is called."""
        called = []
        self.patch(self.action_queue.client, 'get_delta',
                   lambda *a, **b: called.append((a, b)))

        cmd = GetDeltaFromScratch(self.rq, VOLUME)
        cmd._run()
        self.assertEqual(called[0], ((VOLUME,), {'from_scratch': True}))

    def test_handle_success_push_event(self):
        """Test AQ_DELTA_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.GetDelta(self.action_queue.client,
                                  share_id=VOLUME, from_scratch=True)
        request.response = ['foo', 'bar']
        request.end_generation = 76
        request.full = True
        request.free_bytes = 1231234

        # create a command and trigger it success
        cmd = GetDeltaFromScratch(self.rq, VOLUME)
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        delta_info = dict(volume_id=VOLUME, delta_content=['foo', 'bar'],
                          end_generation=76,
                          free_bytes=1231234)
        self.assertEqual(received, ('AQ_RESCAN_FROM_SCRATCH_OK', delta_info))
        self.assertTrue(isinstance(received[1]["delta_content"], DeltaList))

    def test_handle_generic_failure_push_event(self):
        """Test AQ_DELTA_ERROR is pushed on failure."""
        # create a failure
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))

        # create a command and trigger it success
        cmd = GetDeltaFromScratch(self.rq, VOLUME)
        cmd.handle_failure(failure=failure)

        # check for event
        received = self.action_queue.event_queue.events[0]
        self.assertEqual(received, ('AQ_RESCAN_FROM_SCRATCH_ERROR',
                                    {'volume_id': VOLUME, 'error': msg}))

    def test_queued_mixed_types(self):
        """Command gets queued if other command is waiting."""
        cmd1 = FakeCommand()
        self.rq.waiting.append(cmd1)
        cmd2 = GetDeltaFromScratch(self.rq, 'vol2')
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])
        self.assertTrue(self.handler.check_debug("queueing"))

    def test_queued_two_different(self):
        """Two different queued commands is ok."""
        cmd1 = GetDeltaFromScratch(self.rq, 'vol1')
        cmd1.queue()
        cmd2 = GetDeltaFromScratch(self.rq, 'vol2')
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1, cmd2])

    def test_queued_two_equal(self):
        """When two equals, only survive the first one."""
        cmd1 = GetDeltaFromScratch(self.rq, 'vol')
        cmd1.queue()
        cmd2 = GetDeltaFromScratch(self.rq, 'vol')
        cmd2.queue()
        self.assertEqual(list(self.rq.waiting), [cmd1])
        self.assertTrue(self.handler.check_debug("not queueing self"))

    def test_uniqueness(self):
        """Info used for uniqueness."""
        cmd = GetDeltaFromScratch(self.rq, 'vol')
        self.assertEqual(cmd.uniqueness, ('GetDeltaFromScratch', 'vol'))


class UnlinkTestCase(ConnectedBaseTestCase):
    """Test for Unlink ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(UnlinkTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_handle_success_push_event(self):
        """Test AQ_UNLINK_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.Unlink(self.action_queue.client, VOLUME, 'node_id')
        request.new_generation = 13

        # create a command and trigger it success
        cmd = Unlink(self.rq, VOLUME, 'parent_id', 'node_id', 'path')
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(share_id=VOLUME, parent_id='parent_id',
                    node_id='node_id', new_generation=13)
        self.assertEqual(received, ('AQ_UNLINK_OK', info))

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        cmd = Unlink(self.rq, VOLUME, 'parent_id', 'node_id', 'path')
        res = [getattr(cmd, x) for x in cmd.possible_markers]
        self.assertEqual(res, ['node_id', 'parent_id'])

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        cmd = Unlink(self.rq, VOLUME, 'parent_id', 'node_id', 'foo/bar')
        cmd._acquire_pathlock()
        self.assertEqual(t, [('foo', 'bar'), {'on_parent': True,
                                              'on_children': True,
                                              'logger': None}])


class MoveTestCase(ConnectedBaseTestCase):
    """Test for Move ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(MoveTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_handle_success_push_event(self):
        """Test AQ_MOVE_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.Move(self.action_queue.client, VOLUME, 'node',
                                'new_parent', 'new_name')
        request.new_generation = 13

        # create a command and trigger it success
        cmd = Move(self.rq, VOLUME, 'node', 'o_parent', 'n_parent', 'n_name',
                   'path_from', 'path_to')
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(share_id=VOLUME, node_id='node', new_generation=13)
        self.assertEqual(received, ('AQ_MOVE_OK', info))

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        cmd = Move(self.rq, VOLUME, 'node', 'o_parent', 'n_parent', 'n_name',
                   'path_from', 'path_to')
        res = [getattr(cmd, x) for x in cmd.possible_markers]
        self.assertEqual(res, ['node', 'o_parent', 'n_parent'])

    def test_uniqueness(self):
        """Info used for uniqueness."""
        cmd = Move(self.rq, VOLUME, 'node', 'o_parent', 'n_parent', 'n_name',
                   'path_from', 'path_to')
        self.assertEqual(cmd.uniqueness, ('Move', VOLUME, 'node'))

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        fake_acquire = lambda s, *a, **k: t.extend((a, k)) or defer.succeed(None)
        self.patch(PathLockingTree, 'acquire', fake_acquire)
        cmd = Move(self.rq, VOLUME, 'node', 'o_parent', 'n_parent', 'n_name',
                   '/path/from', '/path/to')
        cmd._acquire_pathlock()
        should = [
            ("", "path", "from"), {'on_parent': True, 'on_children': True,
                                                            'logger': None},
            ("", "path", "to"), {'on_parent': True, 'logger': None},
        ]
        self.assertEqual(t, should)

    def test_pathlock_mergepaths(self):
        """Merge both path lockings."""
        d1 = defer.Deferred()
        d2 = defer.Deferred()
        fake_defers = [d1, d2]
        self.patch(PathLockingTree, 'acquire',
                   lambda *a, **k: fake_defers.pop())
        cmd = Move(self.rq, VOLUME, 'node', 'o_p', 'n_p', 'n_n', 'p/f', 'p/t')

        # get the path lock, and add a callback to get the release function
        dl = cmd._acquire_pathlock()
        merge_release = []
        dl.addCallback(merge_release.append)

        # prepare marks to check both original releases are called
        release_called = []

        # dl is triggered only when d1 and d2
        self.assertFalse(dl.called)
        d1.callback(lambda: release_called.append(1))
        self.assertFalse(dl.called)
        d2.callback(lambda: release_called.append(2))
        self.assertTrue(dl.called)

        # release!
        self.assertFalse(release_called)
        merge_release[0]()
        self.assertEqual(sorted(release_called), [1, 2])


class MakeFileTestCase(ConnectedBaseTestCase):
    """Test for MakeFile ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(MakeFileTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_handle_success_push_event(self):
        """Test AQ_FILE_NEW_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.MakeFile(self.action_queue.client, VOLUME,
                                  'parent', 'name')
        request.new_id = 'new_id'
        request.new_generation = 13

        # create a command and trigger it success
        cmd = MakeFile(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(marker='marker', new_id='new_id', new_generation=13,
                    volume_id=VOLUME)
        self.assertEqual(received, ('AQ_FILE_NEW_OK', info))

    def test_handle_failure_push_event(self):
        """Test AQ_FILE_NEW_ERROR is pushed on error."""
        # create a request and fill it with succesful information
        request = client.MakeFile(self.action_queue.client, VOLUME,
                                  'parent', 'name')
        request.new_id = 'new_id'
        request.new_generation = 13

        # create a command and trigger it fail
        cmd = MakeFile(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        failure = Failure(Exception('foo'))
        cmd.handle_failure(failure)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(marker='marker', failure=failure)
        self.assertEqual(received, ('AQ_FILE_NEW_ERROR', info))

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        cmd = MakeFile(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        res = [getattr(cmd, x) for x in cmd.possible_markers]
        self.assertEqual(res, [ 'parent'])

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        cmd = MakeFile(self.rq, VOLUME, 'parent', 'name', 'marker', 'foo/bar')
        cmd._acquire_pathlock()
        self.assertEqual(t, [('foo', 'bar'), {'on_parent': True,
                                              'logger': None}])


class MakeDirTestCase(ConnectedBaseTestCase):
    """Test for MakeDir ActionQueueCommand."""

    def setUp(self):
        """Init."""
        d = super(MakeDirTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    def test_handle_success_push_event(self):
        """Test AQ_DIR_NEW_OK is pushed on success."""
        # create a request and fill it with succesful information
        request = client.MakeDir(self.action_queue.client, VOLUME,
                                 'parent', 'name')
        request.new_id = 'new_id'
        request.new_generation = 13

        # create a command and trigger it success
        cmd = MakeDir(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        cmd.handle_success(request)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(marker='marker', new_id='new_id', new_generation=13,
                    volume_id=VOLUME)
        self.assertEqual(received, ('AQ_DIR_NEW_OK', info))

    def test_handle_failure_push_event(self):
        """Test AQ_DIR_NEW_ERROR is pushed on error."""
        # create a request and fill it with succesful information
        request = client.MakeDir(self.action_queue.client, VOLUME,
                                 'parent', 'name')
        request.new_id = 'new_id'
        request.new_generation = 13

        # create a command and trigger it fail
        cmd = MakeDir(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        failure = Failure(Exception('foo'))
        cmd.handle_failure(failure)

        # check for successful event
        received = self.action_queue.event_queue.events[0]
        info = dict(marker='marker', failure=failure)
        self.assertEqual(received, ('AQ_DIR_NEW_ERROR', info))

    def test_possible_markers(self):
        """Test that it returns the correct values."""
        cmd = MakeDir(self.rq, VOLUME, 'parent', 'name', 'marker', 'path')
        res = [getattr(cmd, x) for x in cmd.possible_markers]
        self.assertEqual(res, ['parent'])

    def test_path_locking(self):
        """Test that it acquires correctly the path lock."""
        t = []
        self.patch(PathLockingTree, 'acquire',
                   lambda s, *a, **k: t.extend((a, k)))
        cmd = MakeDir(self.rq, VOLUME, 'parent', 'name', 'marker', 'foo/bar')
        cmd._acquire_pathlock()
        self.assertEqual(t, [('foo', 'bar'), {'on_parent': True,
                                              'logger': None}])


class TestDeltaList(unittest.TestCase):
    """Tests for DeltaList."""

    def test_is_list(self):
        """A DeltaList is a list."""
        l = [1, 2, 3]
        a = DeltaList(l)
        self.assertTrue(isinstance(a, list))

    def test_is_equal_list(self):
        """A DeltaList is equal to the list it represents."""
        l = [1, 2, 3]
        a = DeltaList(l)
        self.assertEqual(a, l)

    def test_repr(self):
        """A DeltaList has a short representation."""
        a = DeltaList(["a"*1000])
        self.assertTrue(len(repr(a)) < 100)
        self.assertTrue(len(str(a)) < 100)


class AuthenticateTestCase(ConnectedBaseTestCase):
    """Tests for authenticate."""

    def setUp(self):
        """Init."""
        d = super(AuthenticateTestCase, self).setUp()
        self.rq = RequestQueue(action_queue=self.action_queue)
        return d

    @defer.inlineCallbacks
    def test_session_id_is_logged(self):
        """Test that session_id is logged after auth ok."""
        request = client.Authenticate(self.action_queue.client,
                                      {'dummy_token': 'credentials'})
        request.session_id = str(uuid.uuid4())
        self.action_queue.client.oauth_authenticate = \
                lambda *args: defer.succeed(request)

        yield self.action_queue.authenticate()

        self.assertTrue(self.handler.check_note('Session ID: %r' %
                                                str(request.session_id)))


class ActionQueueProtocolTests(TwistedTestCase):
    """Test the ACQ class."""

    def setUp(self):
        """Set up."""
        # create an AQP and put a factory to it
        self.aqp = ActionQueueProtocol()
        obj = Mocker().mock()
        obj.event_queue.push('SYS_CONNECTION_MADE')
        self.aqp.factory = obj

        # set up the logger
        self.handler = MementoHandler()
        self.handler.setLevel(logging.DEBUG)
        self.aqp.log.addHandler(self.handler)

    def tearDown(self):
        """Tear down."""
        self.aqp.log.removeHandler(self.handler)
        task = self.aqp._looping_ping
        if task is not None and task.running:
            task.stop()

    def test_connection_made(self):
        """Connection is made."""
        mocker = Mocker()
        obj = mocker.mock()
        obj.event_queue.push('SYS_CONNECTION_MADE')
        self.aqp.factory = obj
        super_called = []
        self.patch(ThrottlingStorageClient, 'connectionMade',
                   lambda s: super_called.append(True))
        # test
        with mocker:
            self.aqp.connectionMade()
        self.assertTrue(self.handler.check_info('Connection made.'))
        self.assertFalse(self.aqp._looping_ping is None)
        self.assertTrue(super_called)

    def test_connection_lost(self):
        """Connection is lost."""
        super_called = []
        self.patch(ThrottlingStorageClient, 'connectionLost',
                   lambda s, r: super_called.append(True))
        self.aqp.connectionLost('foo')
        self.assertTrue(self.handler.check_info(
                        'Connection lost, reason: foo.'))
        self.assertTrue(self.aqp._looping_ping is None)
        self.assertTrue(super_called)

    def test_ping_connection_made_twice(self):
        """If connection made is called twice, don't create two tasks."""
        self.aqp.connectionMade()
        task1 = self.aqp._looping_ping
        self.aqp.connectionMade()
        task2 = self.aqp._looping_ping
        self.assertTrue(task1 is task2)

    def test_ping_connection_lost_twice(self):
        """If connection lost is called twice, don't stop None."""
        self.aqp.connectionMade()
        self.assertFalse(self.aqp._looping_ping is None)
        self.aqp.connectionLost('reason')
        self.assertTrue(self.aqp._looping_ping is None)
        self.aqp.connectionLost('reason')
        self.assertTrue(self.aqp._looping_ping is None)

    @defer.inlineCallbacks
    def test_ping_task_calls_ping(self):
        """The task will call the _do_ping method."""
        self.aqp._ping_delay = .1
        deferred = defer.Deferred()

        def fake_ping():
            """Stop the loop and trigger the deferred test."""
            self.aqp._looping_ping.stop()
            deferred.callback(True)

        self.aqp._do_ping = fake_ping
        self.aqp.connectionMade()
        yield deferred

    def test_ping_do_ping(self):
        """Ping and log."""
        # mock the request
        mocker = Mocker()
        req = mocker.mock()
        req.rtt
        mocker.result(1.123123)
        self.aqp.ping = lambda: defer.succeed(req)

        # ping will be called, and req accessed, otherwise mocker will complain
        with mocker:
            self.aqp._do_ping()

        # check also the log
        self.assertTrue(self.handler.check_debug('Ping! rtt: 1.123 segs'))


class CommandCycleTestCase(BasicTestCase):
    """Test the command behaviour on run, retry, stop, etc.

    These tests are not exactly unit tests, but more about integration
    between the queue and the command, and how the command life cycle is.
    """

    def setUp(self):
        """Set up."""
        res = super(CommandCycleTestCase, self).setUp()

        class MyCommand(ActionQueueCommand):
            """Monkeypatchable AQC."""

            def _acquire_pathlock(self):
                pathlock = PathLockingTree()
                return pathlock.acquire('foo')

        self.queue = RequestQueue(action_queue=self.action_queue)
        self.queue.run()
        self.cmd = MyCommand(self.queue)
        self.cmd.log = self.cmd.make_logger()

        return res

    def _check_finished_ok(self):
        """Check that command finished ok."""
        self.assertFalse(self.cmd.running)
        self.assertNotIn(self.cmd, self.queue.waiting)
        self.assertFalse(self.action_queue.pathlock.count)

    def test_simple_start_end_ok_queue_active(self):
        """Simple normal cycle, ending ok, queued while active."""
        # check _run was called, and returned ok
        called = []
        self.cmd._run = lambda: called.append(1) or defer.succeed(2)

        # check handled success
        self.cmd.handle_success = lambda a: called.append(a)

        # let the command go
        self.cmd.queue()

        # all check
        self.assertEqual(called, [1, 2])
        self._check_finished_ok()

    def test_simple_start_end_bad_queue_active(self):
        """Simple normal cycle, ending bad, queued while active.."""
        # check _run was called, and returned bad
        called = []
        failure = Failure(ValueError('foo'))
        self.cmd._run = lambda: called.append(1) or defer.fail(failure)

        # check handled failure
        self.cmd.handle_failure = lambda f: called.append(f)

        # let the command go
        self.cmd.queue()

        # all check
        self.assertEqual(called, [1, failure])
        self._check_finished_ok()

    def test_simple_start_end_ok_queue_notactive(self):
        """Simple normal cycle, ending ok, queued while not active."""
        # check _run was called, and returned ok
        called = []
        self.cmd._run = lambda: called.append(1) or defer.succeed(2)

        # check handled success
        self.cmd.handle_success = lambda a: called.append(a)

        # stop the queue and let the command go
        self.queue.stop()
        self.cmd.queue()

        # active the queue later
        self.queue.run()

        # all check
        self.assertEqual(called, [1, 2])
        self._check_finished_ok()

    def test_simple_start_end_bad_queue_notactive(self):
        """Simple normal cycle, ending bad, queued while not active.."""
        # check _run was called, and returned bad
        called = []
        failure = Failure(ValueError('foo'))
        self.cmd._run = lambda: called.append(1) or defer.fail(failure)

        # check handled failure
        self.cmd.handle_failure = lambda f: called.append(f)

        # stop the queue and let the command go
        self.queue.stop()
        self.cmd.queue()

        # active the queue later
        self.queue.run()

        # all check
        self.assertEqual(called, [1, failure])
        self._check_finished_ok()

    def test_condition_goes_ok(self):
        """Command not run initially, but yes when conditions are ok."""
        # check _run was called, and returned ok
        called = []
        self.cmd._run = lambda: called.append('run') or defer.succeed('finish')

        # check handled success
        self.cmd.handle_success = lambda a: called.append(a)

        # command not ready to run
        self.cmd.is_runnable = False

        # let the command go, it will not run
        self.cmd.queue()
        self.assertEqual(called, [])

        # fix conditions and check them
        self.cmd.is_runnable = True
        self.queue.check_conditions()

        # all check
        self.assertEqual(called, ['run', 'finish'])
        self._check_finished_ok()

    def test_check_conditions_while_running(self):
        """Check conditions while the command is running.

        Check conditions can be executed at any time, we need to avoid
        running the command twice.
        """
        # monkeypatch _run to flag called and test "while running"
        called = []
        d = defer.Deferred()
        self.cmd._run = lambda: called.append(1) or d

        # check handled success
        self.cmd.handle_success = lambda a: called.append(a)

        # let the command go
        self.cmd.queue()

        # before the command finishes, all conditions are checked
        self.queue.check_conditions()

        # command finished
        d.callback(2)

        # all check
        self.assertEqual(called, [1, 2])
        self._check_finished_ok()

    def test_disconnect_connect_running(self):
        """Simulate a disconnection and connection while command running."""
        # monkeypatch _run to flag called and test "while running"
        called = []
        d = defer.Deferred()
        self.cmd._run = lambda: called.append('run') or d

        # check handled success and cleanup
        self.cmd.handle_success = lambda a: called.append(a)
        self.cmd.cleanup = lambda: called.append('clean')

        # let the command go
        self.cmd.queue()

        # before the command finishes, we disconnect, and reconnect
        self.queue.stop()
        self.queue.run()

        # command finished
        d.callback('finish')

        # all check
        self.assertEqual(called, ['run', 'clean', 'run', 'finish'])
        self._check_finished_ok()

    def test_disconnect_connect_pathlocked(self):
        """Simulate a disconnection and connection while waiting pathlock."""
        # check it called run
        called = []
        self.cmd._run = lambda: called.append('run') or defer.succeed('finish')

        # monkeypatch to test "while waiting pathlock"
        d = defer.Deferred()
        self.cmd._acquire_pathlock = lambda: d

        # check handled success
        self.cmd.handle_success = lambda a: called.append(a)

        # let the command go
        self.cmd.queue()

        # before the pathlock is released, we disconnect, and reconnect
        self.queue.stop()
        self.queue.run()

        # release the pathlock
        d.callback(None)

        # all check
        self.assertEqual(called, ['run', 'finish'])
        self._check_finished_ok()

    @defer.inlineCallbacks
    def test_retry_immediate(self):
        """Retry the command immediately."""
        finished = defer.Deferred()
        called = []
        failure = Failure(ClientNoLongerThere('foo')) # retryable!
        run_deferreds = [defer.fail(failure), defer.succeed('finish')]
        self.cmd._run = lambda: called.append('run') or run_deferreds.pop(0)

        # check handle success (failure is never called because it's retried)
        self.cmd.handle_success = lambda a: called.append(a)

        # need to wait finish() called, to be sure all ended ok, because of
        # the callLater for the retry
        def fake_finish():
            ActionQueueCommand.finish(self.cmd)
            finished.callback(True)
        self.cmd.finish = fake_finish

        # let the command go
        self.cmd.queue()

        # need to wait the callLater
        yield finished

        # all check
        self.assertEqual(called, ['run', 'run', 'finish'])
        self._check_finished_ok()

    def test_retry_conditions_solved(self):
        """Retry the command because conditions solved later."""
        finished = defer.Deferred()
        called = []

        def f1():
            """Fail and make conditions not ok to run."""
            self.cmd.is_runnable = False
            failure = Failure(ClientNoLongerThere('foo')) # retryable!
            return defer.fail(failure)

        def f2():
            """Run ok."""
            return defer.succeed('finish')

        run_functions = [f1, f2]
        self.cmd._run = lambda: called.append('run') or run_functions.pop(0)()

        # check handle success (failure is never called because it's retried)
        self.cmd.handle_success = lambda a: called.append(a)

        # need to wait finish() called, to be sure all ended ok, because of
        # the callLater for the retry
        def fake_finish():
            ActionQueueCommand.finish(self.cmd)
            finished.callback(True)
        self.cmd.finish = fake_finish

        # let the command go, it will fail and wait for conditions
        self.cmd.queue()
        self.assertEqual(called, ['run'])

        # fix conditions
        self.cmd.is_runnable = True
        self.queue.check_conditions()

        # need to wait the callLater
        yield finished

        # all check
        self.assertEqual(called, ['run', 'run', 'finish'])
        self._check_finished_ok()

    def test_cancel_while_running(self):
        """Cancel the command while running."""
        # monkeypatch _run to flag called and test "while running"
        called = []
        d = defer.Deferred()
        self.cmd._run = lambda: called.append(1) or d
        self.cmd.markers_resolved_deferred = defer.succeed(True)

        # check cleanup
        self.cmd.cleanup = lambda: called.append(2)
        self.cmd.finish = lambda: called.append(3)

        # let the command go
        self.cmd.run()

        # before it finishes, cancel
        self.cmd.cancel()

        # all check
        self.assertTrue(self.cmd.cancelled)
        self.assertEqual(called, [1, 2, 3])
        self._check_finished_ok()

    def test_cancel_while_pathclocked(self):
        """Cancel the command while running."""
        # monkeypatch _run to flag called and test "while running"
        called = []
        self.cmd.run = lambda: called.append('should not')

        # monkeypatch to test "while waiting pathlock"
        d = defer.Deferred()
        self.cmd._acquire_pathlock = lambda: d

        # let the command go, and cancel in the middle
        self.cmd.queue()
        self.cmd.cancel()

        # unlock the pathlock
        d.callback(lambda: called.append(1))

        # all check
        self.assertEqual(called, [1])
        self._check_finished_ok()
