#
# Author: Facundo Batista <facundo@canonical.com>
#
# Author: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the Event Queue."""

import logging
import unittest

from ubuntuone.syncdaemon import (
    event_queue,
    filesystem_manager,
    tritcask,
)
from contrib.testing import testcase
from twisted.internet import defer
from ubuntuone.devtools.handlers import MementoHandler


class BaseEQTestCase(testcase.BaseTwistedTestCase):
    """ Setup an EQ for test. """

    def setUp(self):
        """Setup the test."""
        testcase.BaseTwistedTestCase.setUp(self)
        self.fsmdir = self.mktemp('fsmdir')
        self.partials_dir = self.mktemp('partials_dir')
        self.root_dir = self.mktemp('root_dir')
        self.home_dir = self.mktemp('home_dir')
        self.vm = testcase.FakeVolumeManager(self.root_dir)
        self.db = tritcask.Tritcask(self.mktemp('tritcask'))
        self.fs = filesystem_manager.FileSystemManager(self.fsmdir,
                                                       self.partials_dir,
                                                       self.vm, self.db)
        self.fs.create(path=self.root_dir,
                       share_id='', is_dir=True)
        self.fs.set_by_path(path=self.root_dir,
                              local_hash=None, server_hash=None)
        self.eq = event_queue.EventQueue(self.fs)
        self.fs.register_eq(self.eq)

        # add a Memento handler to the logger
        self.log_handler = MementoHandler()
        self.log_handler.setLevel(logging.DEBUG)
        self.eq.log.addHandler(self.log_handler)

    def tearDown(self):
        """Clean up the tests."""
        for listener in self.eq._listeners:
            self.eq.unsubscribe(listener)
        self.eq.shutdown()
        self.db.shutdown()
        self.rmtree(self.tmpdir)
        testcase.BaseTwistedTestCase.tearDown(self)


class SubscriptionTests(BaseEQTestCase):
    """Test to subscribe and unsubscribe to the EQ."""

    def test_subscription(self):
        """Test that subscription works."""
        # pylint: disable-msg=W0212
        # simple one
        o = object()
        self.eq.subscribe(o)
        self.assertTrue(o in self.eq._listeners)

        # not duplicates
        o = object()
        self.eq.subscribe(o)
        self.eq.subscribe(o)
        self.assertTrue(self.eq._listeners.count(o) == 1)

    def test_unsubscription(self):
        """Test that unsubscription works."""
        # pylint: disable-msg=W0212
        # simple one
        o = object()
        self.eq.subscribe(o)
        self.eq.unsubscribe(o)
        self.assertFalse(o in self.eq._listeners)

        # duplicate ok (two suscs)
        o = object()
        self.eq.subscribe(o)
        self.eq.subscribe(o)
        self.eq.unsubscribe(o)
        self.assertFalse(o in self.eq._listeners)

        # duplicate error (two unsuscs)
        o = object()
        self.eq.subscribe(o)
        self.eq.subscribe(o)
        self.eq.unsubscribe(o)
        self.assertRaises(ValueError, self.eq.unsubscribe, o)

        # indecisive
        o = object()
        self.eq.subscribe(o)
        self.eq.unsubscribe(o)
        self.eq.subscribe(o)
        self.eq.unsubscribe(o)
        self.assertFalse(o in self.eq._listeners)


class PushTests(BaseEQTestCase):
    """Test the event distribution machinery."""

    def test_push_simple(self):
        """Test that events can be pushed (not listening yet)."""
        # not even an event
        self.assertRaises(TypeError, self.eq.push)

        # incorrect args, only kwargs supported
        self.assertRaises(TypeError, self.eq.push, "FS_FILE_MOVE", 1)
        self.assertRaises(TypeError,
                      self.eq.push, "FS_FILE_MOVE", 1, path_to=2)

        # ok: just kwargs
        self.eq.push("FS_FILE_MOVE", path_from=1, path_to=2)

    def test_events_kwargs(self):
        """Test that all events are defined correctly with tuples or lists.

        This is to avoid a typical mistake of making it a "(param)", not
        a "(param,)".
        """
        for name, params in event_queue.EVENTS.iteritems():
            self.assertTrue(isinstance(params, (tuple, list)),
                            "%s event has params bad defined!" % name)

    def test_listened_pushs(self):
        """Push events and listem them."""

        # helper class, pylint: disable-msg=C0111
        class Create(object):
            def __init__(self):
                self.a = None
            def handle_FS_FILE_CREATE(self, path):
                self.a = path

        # it get passed!
        c = Create()
        self.eq.subscribe(c)
        self.eq.push("FS_FILE_CREATE", path=1)
        self.assertEqual(c.a, 1)
        self.eq.unsubscribe(c)

        # don't get what don't listen
        c = Create()
        self.eq.subscribe(c)
        self.eq.push("FS_FILE_DELETE", path=1)
        self.assertEqual(c.a, None)
        self.eq.unsubscribe(c)

    def test_signatures(self):
        """Check that the handle signatures are forced when passing."""

        # helper class, pylint: disable-msg=C0111
        class Create(object):
            def handle_FS_FILE_CREATE(self, notpath): # it should be path here
                pass

        # it get passed!
        c = Create()
        self.eq.subscribe(c)

        # the listener has a wrong signature
        # this is logged as an error/exception
        self.eq.push("FS_FILE_CREATE", path=1)
        self.assertTrue(self.log_handler.check_error('FS_FILE_CREATE',
                                                     'Create object'))

        self.eq.unsubscribe(c)

    def test_log_pushing_data(self):
        """Pushed event and info should be logged."""
        self.eq.push("AQ_QUERY_ERROR", item='item', error='err')
        self.assertTrue(self.log_handler.check_debug(
                        "push_event: AQ_QUERY_ERROR, kwargs: "
                        "{'item': 'item', 'error': 'err'}"))

    def test_log_delete_in_info(self):
        """Pushed any deletion event should be logged in info."""
        self.eq.push("FS_DIR_DELETE", path='path')
        self.assertTrue(self.log_handler.check_info(
                        "push_event: FS_DIR_DELETE"))

    def test_log_pushing_private_data(self):
        """SYS_USER_CONNECT event info must not be logged."""
        self.eq.push("SYS_USER_CONNECT", access_token='foo')
        self.assertTrue(self.log_handler.check_debug(
            "push_event: SYS_USER_CONNECT, kwargs: *"))


class PushTestsWithCallback(BaseEQTestCase):
    """Test the error handling in the event distribution machinery."""

    def test_keep_going(self):
        """ Check that if a listener raises an Exception or have a
        wrong signature, the next listeners are called.
        """
        d = defer.Deferred()
        # helper class, pylint: disable-msg=C0111
        class BadListener(object):
            def handle_FS_FILE_CREATE(self, notpath): # it should be path here
                d.callback(False)

        class GoodListener(object):
            def handle_FS_FILE_CREATE(self, path):
                d.callback(path)

        bl = BadListener()
        gl = GoodListener()
        self.eq.subscribe(bl)
        self.eq.subscribe(gl)

        def cleanup():
            """ unsubscribe the listeners """
            self.eq.unsubscribe(bl)
            self.eq.unsubscribe(gl)
        self.addCleanup(cleanup)

        # one listener has a wrong signature
        self.eq.push("FS_FILE_CREATE", path=1)
        def callback(result):
            """ asserts that GoodListener was called. """
            self.assertTrue(result)
            self.assertEquals(1, result)

        d.addCallback(callback)
        return d

    def test_default_handler(self):
        """ Check that handler_default is called. """
        d = defer.Deferred()
        # helper class, pylint: disable-msg=C0111
        class Listener(object):
            def handle_default(self, event, **kwargs):
                d.callback((event, kwargs))

        l = Listener()
        self.eq.subscribe(l)

        def cleanup():
            """ unsubscribe the listeners """
            self.eq.unsubscribe(l)
        self.addCleanup(cleanup)

        # push some event and expect it'll be handled by handle_default
        self.eq.push("FS_FILE_CREATE", path=1)
        def callback(result):
            """ asserts that GoodListener was called. """
            self.assertEquals(2, len(result))
            self.assertEquals('FS_FILE_CREATE', result[0])
            self.assertEquals({'path': 1}, result[1])

        d.addCallback(callback)
        return d

    def test_ordered_dispatch(self):
        """ Check that the events are pushed to all listeners in order. """
        d = defer.Deferred()
        # helper class, pylint: disable-msg=C0111
        class Listener(object):
            def __init__(self, eq):
                self.eq = eq
                self.events = []

            def handle_FS_FILE_CREATE(self, path):
                self.events.append('FS_FILE_CREATE')
                self.eq.push('FS_FILE_MOVE', path_from=path, path_to=2)

            def handle_FS_FILE_DELETE(self, path):
                self.events.append('FS_FILE_DELETE')

            def handle_FS_FILE_MOVE(self, path_from, path_to):
                self.events.append('FS_FILE_MOVE')
                d.callback(True)

        # create 10 listeners in order to create an event madness
        listeners = []
        # pylint: disable-msg=W0612
        for i in xrange(0, 10):
            l = Listener(self.eq)
            listeners.append(l)
            self.eq.subscribe(l)

        def cleanup():
            """ unsubscribe the listeners """
            for l in listeners:
                self.eq.unsubscribe(l)
        self.addCleanup(cleanup)

        # push some events to unleash the event madness
        self.eq.push("FS_FILE_CREATE", path=1)
        self.eq.push('FS_FILE_DELETE', path=2)

        def callback(result):
            """ asserts that Listener was called in the right order"""
            listeners_events = [listener.events for listener in listeners]
            for l_events in listeners_events:
                for other_l_events in listeners_events:
                    self.assertEqual(l_events, other_l_events)

        d.addCallback(callback)
        return d


def test_suite():
    # pylint: disable-msg=C0111
    return unittest.TestLoader().loadTestsFromName(__name__)

if __name__ == "__main__":
    unittest.main()
