# ubuntuone.status.aggregator
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Aggregate status events."""

import itertools
import operator
import os
import sys

from twisted.internet import reactor, defer

from ubuntuone.logger import basic_formatter, logging
from ubuntuone.platform.notification import Notification
from ubuntuone.platform.messaging import Messaging

LOG_LEVEL = logging.DEBUG
logger = logging.getLogger('ubuntuone.status')
logger.setLevel(LOG_LEVEL)

if os.environ.get('DEBUG_STATUS'):
    debug_handler = logging.StreamHandler(sys.stderr)
    debug_handler.setFormatter(basic_formatter)
    debug_handler.setLevel(LOG_LEVEL)
    logger.addHandler(debug_handler)

UBUNTUONE_TITLE = "Ubuntu One"
FINAL_COMPLETED = "File synchronization completed."
FINAL_UPLOADED = "%d files were uploaded to your cloud."
FINAL_DOWNLOADED = "%d files were downloaded to your computer."
FILES_UPLOADING = "%d files are uploading to your cloud."
FILES_DOWNLOADING = "%d files are downloading to your computer."
PROGRESS_UPLOADED = "Uploaded %d/%d files."
PROGRESS_DOWNLOADED = "Downloaded %d/%d files."
PROGRESS_COMPLETED = "%d%% completed."


class StatusEvent(object):
    """An event representing a status change."""

    MESSAGE_ONE = None  # to be defined in child classes
    MESSAGE_MANY = None  # to be defined in child classes
    WEIGHT = 99
    DO_NOT_INSTANCE = "Do not instance this class, only children."""

    def __init__(self, *args):
        """Initialize this instance."""
        assert type(self) != StatusEvent, self.DO_NOT_INSTANCE
        self.args = args

    def one(self):
        """A message if this is the only event of this type."""
        return self.MESSAGE_ONE

    def many(self, count):
        """A message if there are many events of this type."""
        return self.MESSAGE_MANY % count


class FilePublishingStatus(StatusEvent):
    """Files that are made public with a url."""

    MESSAGE_ONE = "A file was just made public at %s"
    MESSAGE_MANY = "%d files were just made public"
    WEIGHT = 50

    def one(self):
        """Show the url if only one event of this type."""
        return self.MESSAGE_ONE % self.args


class FileUnpublishingStatus(StatusEvent):
    """Files that have stopped being published."""

    MESSAGE_ONE = "A file is no longer published"
    MESSAGE_MANY = "%d files are no longer published"
    WEIGHT = 51


class FolderAvailableStatus(StatusEvent):
    """Folders available for subscription."""

    MESSAGE_MANY = "Found %d new cloud folders."""
    WEIGHT = 60


class ShareAvailableStatus(FolderAvailableStatus):
    """A Share is available for subscription."""

    MESSAGE_ONE = "New cloud folder available: <%s> shared by <%s>"""

    def one(self):
        """Show the folder information."""
        volume = self.args[0]
        params = (volume.name, volume.other_visible_name)
        return self.MESSAGE_ONE % params


class UDFAvailableStatus(FolderAvailableStatus):
    """An UDF is available for subscription."""

    MESSAGE_ONE = "New cloud folder available: %s"""

    def one(self):
        """Show the folder information."""
        volume = self.args[0]
        params = (volume.suggested_path,)
        return self.MESSAGE_ONE % params


class Timer(defer.Deferred):
    """A deferred that fires past a given delay."""

    def __init__(self, delay, clock=reactor):
        """Initialize this instance."""
        defer.Deferred.__init__(self)
        self.clock = clock
        self.delay = delay
        self.delay_call = self.clock.callLater(delay, self.callback)

    def cancel_if_active(self, call):
        """Cancel a call if it is active."""
        if call.active():
            call.cancel()

    def cleanup(self):
        """Cancel all active calls."""
        self.cancel_if_active(self.delay_call)

    def callback(self, result=None):
        """Make sure the timers are stopped when firing the callback."""
        self.cleanup()
        defer.Deferred.callback(self, result)

    def reset(self):
        """Reset the delay."""
        if not self.called:
            self.delay_call.reset(self.delay)


class DeadlineTimer(Timer):
    """A Timer with a deadline."""

    def __init__(self, delay, timeout=None, clock=reactor):
        """Initialize this instance."""
        Timer.__init__(self, delay, clock)
        self.timeout = timeout
        self.timeout_call = self.clock.callLater(timeout, self.callback)

    def cleanup(self):
        """Cancel all active calls."""
        Timer.cleanup(self)
        self.cancel_if_active(self.timeout_call)


class FileDiscoveryBaseState(object):
    """States for file discovery bubble."""

    def __init__(self, bubble):
        """Initialize this instance."""
        self.bubble = bubble
        self.clock = bubble.clock

    def new_file_found(self):
        """New files found."""

    def cleanup(self):
        """Cleanup this instance."""

class FileDiscoveryIdleState(FileDiscoveryBaseState):
    """Waiting for first file to appear."""

    def new_file_found(self):
        """New files found."""
        self.bubble._start()


class FileDiscoveryGatheringState(FileDiscoveryBaseState):
    """Files are gathered then a notification is shown."""

    initial_delay = 0.5
    initial_timeout = 3.0

    def __init__(self, *args):
        """Initialize this instance."""
        super(FileDiscoveryGatheringState, self).__init__(*args)
        self.timer = DeadlineTimer(self.initial_delay,
                                 self.initial_timeout,
                                 clock=self.clock)
        self.timer.addCallback(self._timeout)

    def _timeout(self, result):
        """Show the notification bubble."""
        self.cleanup()
        self.bubble._popup()

    def new_file_found(self):
        """New files found."""
        self.timer.reset()

    def cleanup(self):
        """Cleanup this instance."""
        self.timer.cleanup()


class FileDiscoveryUpdateState(FileDiscoveryBaseState):
    """The bubble is updated if more files are found."""

    updates_delay = 0.5
    updates_timeout = 10.0

    def __init__(self, *args):
        """Initialize this instance."""
        super(FileDiscoveryUpdateState, self).__init__(*args)
        self.main_timer = Timer(self.updates_timeout, clock=self.clock)
        self.main_timer.addCallback(self._timeout)
        self.updates_timer = None

    def _timeout(self, result):
        """No more updates on the notification bubble."""
        self.cleanup()
        self.bubble.start_sleeping()

    def _update(self, result):
        """The bubble should be updated."""
        self.bubble._update()

    def new_file_found(self):
        """New files found."""
        if self.updates_timer is None:
            self.updates_timer = Timer(self.updates_delay, clock=self.clock)
            self.updates_timer.addCallback(self._update)

    def cleanup(self):
        """Clean up the timers."""
        self.main_timer.cleanup()
        if self.updates_timer:
            self.updates_timer.cleanup()


class FileDiscoverySleepState(FileDiscoveryBaseState):
    """The bubble is not updated while sleeping."""

    sleep_delay = 300.0

    def __init__(self, *args):
        """Initialize this instance."""
        super(FileDiscoverySleepState, self).__init__(*args)
        self.main_timer = Timer(self.sleep_delay, clock=self.clock)
        self.main_timer.addCallback(self._timeout)

    def _timeout(self, result):
        """Move the notification to the idle state."""
        self.bubble._set_idle()

    def cleanup(self):
        """Clean up the timers."""
        self.main_timer.cleanup()


class FileDiscoveryBubble(object):
    """
    Show a notification for file discovery.

    Waits 3 seconds for the file count to coalesce, then pops up a
    notification. If new files are found the notification is updated,
    but twice per second at most, and for up to 10 seconds.
    Finally, sleeps for 10 minutes so it does not get annoying.
    """

    state = None
    last_notification_id = 0

    def __init__(self, status_aggregator, clock=reactor):
        """Initialize this instance."""
        self.clock = clock
        self.status_aggregator = status_aggregator
        self._set_idle()

    def _change_state(self, new_state_class):
        """Change to a new state."""
        if self.state:
            self.state.cleanup()
        self.state = new_state_class(self)

    def _set_idle(self):
        """Reset this bubble to the initial state."""
        self._change_state(FileDiscoveryIdleState)

    def _start(self):
        """The first file was found, so start gathering."""
        self.notification = Notification()
        self._change_state(FileDiscoveryGatheringState)

    def _popup(self):
        """Display the notification."""
        text = self.status_aggregator.get_discovery_message()
        self.last_notification_id = self.notification.send_notification(
                                                        UBUNTUONE_TITLE, text)
        self.status_aggregator.restart_progress_bubble()
        logger.debug("notification shown: %s", text)
        self._change_state(FileDiscoveryUpdateState)

    def _update(self):
        """Update the notification."""
        text = self.status_aggregator.get_discovery_message()
        logger.debug("notification updated: %s", text)
        self.status_aggregator.restart_progress_bubble()
        self.notification.update_notification(self.last_notification_id,
                                              UBUNTUONE_TITLE, text)

    def start_sleeping(self):
        """Wait for 10 minutes before annoying again."""
        self._change_state(FileDiscoverySleepState)

    def cleanup(self):
        """Cleanup this instance."""
        self.state.cleanup()

    def new_file_found(self):
        """New files found."""
        self.state.new_file_found()


class ProgressBubble(object):
    """Show a notification for transfer progress."""

    sleep_delay = 600
    notification = None
    timer = None

    def __init__(self, status_aggregator, clock=reactor):
        """Initialize this instance."""
        self.status_aggregator = status_aggregator
        self.clock = clock

    def restart(self):
        """Start running the timer."""
        self.cleanup()
        self.timer = Timer(self.sleep_delay, clock=self.clock)
        self.timer.addCallback(self._timeout)

    def _timeout(self, result):
        """Show the bubble."""
        self.notification = Notification()
        text = self.status_aggregator.get_progress_message()
        self.notification.send_notification(UBUNTUONE_TITLE, text)
        self.restart()

    def cleanup(self):
        """Cleanup this instance."""
        if self.timer:
            self.timer.cleanup()

    stop = cleanup


class FinalStatusBubble(object):
    """Final bubble that shows the status of transfers."""

    notification = None

    def __init__(self, status_aggregator):
        """Initialize this instance."""
        self.status_aggregator = status_aggregator

    def cleanup(self):
        """Clean up this instance."""

    def show(self):
        """Show the final status notification."""
        self.notification = Notification()
        text = self.status_aggregator.get_final_status_message()
        self.notification.send_notification(UBUNTUONE_TITLE, text)


def group_statuses(status_events):
    """Groups statuses by weight."""
    weight_getter = operator.attrgetter("WEIGHT")
    sorted_status_events = sorted(status_events, key=weight_getter)
    return itertools.groupby(sorted_status_events, weight_getter)


class StatusAggregator(object):
    """The status aggregator backend."""

    file_discovery_bubble = None
    progress_bubble = None
    final_status_bubble = None

    def __init__(self, clock=reactor):
        """Initialize this instance."""
        self.clock = clock
        self.reset()
        self.messaging = Messaging()

    def reset(self):
        """Reset all counters and notifications."""
        self.total_counter = 0
        self.done_counter = 0
        self.download_done = 0
        self.download_total = 0
        self.upload_total = 0
        self.upload_done = 0
        self.files_uploading = []
        self.files_downloading = []
        self.interesting_events = []

        if self.file_discovery_bubble:
            self.file_discovery_bubble.cleanup()
        self.file_discovery_bubble = FileDiscoveryBubble(self,
                                                         clock=self.clock)

        if self.progress_bubble:
            self.progress_bubble.cleanup()
        self.progress_bubble = ProgressBubble(self, clock=self.clock)

        if self.final_status_bubble:
            self.final_status_bubble.cleanup()
        self.final_status_bubble = FinalStatusBubble(self)

    def interesting_event(self, event):
        """A new interesting event was found."""
        self.interesting_events.append(event)

    def get_discovery_message(self):
        """Get the text for the discovery bubble."""
        lines = []

        if len(self.files_uploading) > 0:
            text = FILES_UPLOADING % len(self.files_uploading)
            lines.append(text)

        if len(self.files_downloading) > 0:
            text = FILES_DOWNLOADING % len(self.files_downloading)
            lines.append(text)

        return "\n".join(lines)

    def get_progress_message(self):
        """Get some lines describing the progress."""
        assert self.total_counter > 0
        parts = []
        if self.upload_total:
            parts.append(PROGRESS_UPLOADED % (self.upload_done,
                                              self.upload_total))
        if self.download_total:
            parts.append(PROGRESS_DOWNLOADED % (self.download_done,
                                                self.download_total))
        progress_percentage = 100.0 * self.done_counter / self.total_counter
        parts.append(PROGRESS_COMPLETED % int(progress_percentage))
        return " ".join(parts)

    def get_final_status_message(self):
        """Get some lines describing all we did."""
        parts = []
        parts.append(FINAL_COMPLETED)
        if self.upload_done:
            parts.append(FINAL_UPLOADED % self.upload_done)
        if self.download_done:
            parts.append(FINAL_DOWNLOADED % self.download_done)
        return "\n".join(parts)

    def restart_progress_bubble(self):
        """Restart the progress bubble."""
        self.progress_bubble.restart()

    def queue_done(self):
        """Show final bubble and reset counters."""
        self.final_status_bubble.show()
        self.reset()

    def no_callback(self, status_events):
        """Pushed a bunch of events by the buffer."""

        lines = []

        if self.show_progress:
            lines.append(self.get_progress_message())

        for weight, statuses in group_statuses(status_events):
            statuses = list(statuses)
            count = len(statuses)
            first_status = statuses[0]
            if count == 1:
                lines.append(first_status.one())
            else:
                lines.append(first_status.many(count))

        text = "\n".join(lines)
        self.notification.send_notification(UBUNTUONE_TITLE, text)
        logger.debug("notification shown: %s", text)
        self.show_progress = False

    def misc_command_queued(self, command):
        """A miscellaneous command was queued."""
        self.total_counter += 1
        logger.debug("queueing command (%d/%d): %s", self.done_counter,
                     self.total_counter, command.__class__.__name__)

    def misc_command_unqueued(self, command):
        """A miscellaneous command was unqueued."""
        self.done_counter += 1
        logger.debug("unqueueing command (%d/%d): %s", self.done_counter,
                     self.total_counter, command.__class__.__name__)

    def download_started(self, command):
        """A download just started."""
        self.files_downloading.append(command)
        self.download_total += 1
        self.misc_command_queued(command)
        self.file_discovery_bubble.new_file_found()

    def download_finished(self, command):
        """A download just finished."""
        if command in self.files_downloading:
            self.files_downloading.remove(command)
        self.download_done += 1
        self.misc_command_unqueued(command)

    def upload_started(self, command):
        """An upload just started."""
        self.files_uploading.append(command)
        self.upload_total += 1
        self.misc_command_queued(command)
        self.file_discovery_bubble.new_file_found()

    def upload_finished(self, command):
        """An upload just finished."""
        if command in self.files_uploading:
            self.files_uploading.remove(command)
        self.upload_done += 1
        self.misc_command_unqueued(command)

class StatusFrontend(object):
    """Frontend for the status aggregator, used by the StatusListener."""

    def __init__(self, clock=reactor):
        """Initialize this instance."""
        self.aggregator = StatusAggregator(clock=clock)

    def file_published(self, public_url):
        """A file was published."""
        self.aggregator.interesting_event(FilePublishingStatus(public_url))

    def file_unpublished(self, public_url):
        """A file was unpublished."""
        self.aggregator.interesting_event(FileUnpublishingStatus(public_url))

    def download_started(self, command):
        """A file was queued for download."""
        self.aggregator.download_started(command)

    def download_finished(self, command):
        """A file download was unqueued."""
        self.aggregator.download_finished(command)

    def upload_started(self, command):
        """A file was queued for upload."""
        self.aggregator.upload_started(command)

    def upload_finished(self, command):
        """A file upload was unqueued."""
        self.aggregator.upload_finished(command)

    def queue_added(self, command):
        """A command was added to the queue."""
        self.aggregator.misc_command_queued(command)

    def queue_removed(self, command):
        """A command was removed from the queue."""
        self.aggregator.misc_command_unqueued(command)

    def queue_done(self):
        """The queue is empty."""
        self.aggregator.queue_done()

    def new_share_available(self, share):
        """A new share is available for subscription."""
        self.aggregator.interesting_event(ShareAvailableStatus(share))

    def new_udf_available(self, udf):
        """A new udf is available for subscription."""
        self.aggregator.interesting_event(UDFAvailableStatus(udf))
